// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

extern bool vanilla_timing;
extern vprFPGA *FPGA;
extern double tdPlacementLambda;
extern bool maxCutCost;
extern bool aveCutCost;
extern bool placementRetiming;

//#define TIMING_DEBUG

// Set name of mcnc netlist to n.
void Netlist::set_name(string n){
  if(magicNum!=18111)
    ERROR("ERROR: trying to access a deleted node...");
  name = n;
}

// Add p_block to list of logic blocks.
void Netlist::addto_lbList(logicBlock* p_block){
  if(!p_block)
    ERROR("logic block cannot be NULL...");
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  lbList.push_back(p_block);

  if(p_block->get_blockType() == GBT_NO_TYPE){
    ERROR("Block should have type!");
  }
  else if(p_block->get_blockType() == INPUTPAD)
    numInputs++;
  else if(p_block->get_blockType() == OUTPUTPAD)
    numOutputs++;
  else
    numlogicBlocks++;
}

//Return the number of registers in the netlist right now
int Netlist::getNumPipeliningRegisters(){
  return(placementPipelineRegisterVector.size());
}

// // for each net in this netlist, create vector of sinks, and then populate
// void Netlist::create_netSinkVectors(){
//   if(magicNum!=18111)
//     ERROR("trying to access a deleted node...");
//   list<Net*>::iterator net_itr = netList.begin();

//   for(;net_itr!=netList.end();net_itr++)
//     (*net_itr)->create_sinkVector();

// }

//For each logic block in the netlist, figure out whether or not we can push register
// forward or backward through them.  If so, figure out how many.
void Netlist::calculateRegPushes(){
  logicBlock *lb;

  create_lbListItr();
  for(lb = getnext_lbList(); lb != 0; lb = getnext_lbList()){
    lb->calculateRegPushBackward();
    lb->calculateRegPushForward();
  }
  remove_lbListItr();
}


// for each net in this netlist, create the slackVector
void Netlist::create_slackVectors(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  list<Net*>::iterator net_itr = netList.begin();

  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->create_slackVector();
}

// for each net in this netlist, create the linkDelayVector
void Netlist::create_linkDelayVectors(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  list<Net*>::iterator net_itr = netList.begin();

  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->create_linkDelayVector();
}

//For all logic blocks in this netlist, initialize all terminal arrival/departure times to a large number
//and required times to a large negative number 
//For all nets in this netlist, initialize all source/sink slackVector values to a large negative number
void Netlist::initTimingGraph(bool pipeline_aware){
  list<logicBlock*>::iterator lbIter;
  list<Terminal*>::iterator termIter;
  list<Net*>::iterator netIter;
  vector<double>::iterator slackIter;

  list<PipeliningNet*>::iterator pNetIter;

  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    //Initialize all input terminals
    for(termIter = (*lbIter)->inputList.begin(); termIter != (*lbIter)->inputList.end(); termIter++){
      (*termIter)->arrivalTime = -1.0E9;
      //The departure time for an input terminal is meaningless
      (*termIter)->departureTime = -1.0E9;
      (*termIter)->requiredTime = 1.0E9;
    }

    //Initialize all output terminals
    for(termIter = (*lbIter)->outputList.begin(); termIter != (*lbIter)->outputList.end(); termIter++){
      //The arrival time for an output terminal is meaningless
      (*termIter)->arrivalTime = -1.0E9;
      (*termIter)->departureTime = -1.0E9;
      (*termIter)->requiredTime = 1.0E9;
    }
  }
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    if(pipeline_aware && (*netIter)->pipelined){
      //This net has some pipelined sinks
      //Since we won't be using it, we set the host net's slack value to a large negative
      for(slackIter = (*netIter)->slackVector.begin(); slackIter != (*netIter)->slackVector.end(); slackIter++){
	(*slackIter) = -1.0E9;
      }

      //If we are doing pipeline aware routing and this is a net with pipelined sinks,
      //we have generated a number of pipelining nets within this net
      //we have to reset the slackVector for all of these pipelining nets
      for(pNetIter = (*netIter)->pipelineNetList.begin(); pNetIter != (*netIter)->pipelineNetList.end(); pNetIter++){
	for(slackIter = (*pNetIter)->slackVector.begin(); slackIter != (*pNetIter)->slackVector.end(); slackIter++){
	  (*slackIter) = 1.0E9;
	}
      }

      //Set the appropriate arrival, departure and required time for all delay nodes      
      for(termIter = (*netIter)->pipelineRegisterList.begin(); termIter != (*netIter)->pipelineRegisterList.end(); termIter++){
	(*termIter)->arrivalTime = -1.0E9;
	(*termIter)->departureTime = -1.0E9;
	(*termIter)->requiredTime = 1.0E9;
      }
    }
    else{
      for(slackIter = (*netIter)->slackVector.begin(); slackIter != (*netIter)->slackVector.end(); slackIter++){
	(*slackIter) = 1.0E9;
      }
    }
  }
}

//Set the arrival/departure times for all net terminals and delay registers
void Netlist::calculateArrivalDepartureTimes(bool pipeline_aware){
  list<logicBlock*>::iterator lbIter;
  list<Terminal*>::iterator outTermListIter;
  list<Terminal*>::iterator inTermListIter;
  
  Net *net;
  PipeliningNet *pNet;

  queue<Terminal *> timingQueue;
  vector<Terminal*>::iterator sinkTermIter;
  int i;
  logicBlock *lBlock;
  bool termReady;
  Terminal* outputTerminal;

  list<Net*>::iterator netIter;
  
  //Build a queue of output terminals to process.
  //Take all registered logic blocks and set their output terminal departure times to 0.0
  //Then add output terminals to queue.
  //If we are calculating timing normally, this will include all input blocks and registers.
  //If we are calculating "vanilla" timing, this will include all blocks
  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if(!vanilla_timing){
      //!!!!!!!!!!!!!!!!!!!!!!!LOOK AT THIS LINE LATER!!!!!!!!!!!!!!!!!
      if((*lbIter)->blockType == INPUTPAD || (*lbIter)->blockType == REGISTER_BANK || (*lbIter)->blockType == RAM_BLOCK){
	for(outTermListIter = (*lbIter)->outputList.begin(); outTermListIter != (*lbIter)->outputList.end(); outTermListIter++){
	  (*outTermListIter)->departureTime = 0.0;
#ifdef TIMING_DEBUG
	  cout << "Set terminal " << (*outTermListIter)->name << " departure time to " << (*outTermListIter)->departureTime << endl;
#endif
	  timingQueue.push(*outTermListIter);
	}
      }
    }
    else{
      for(outTermListIter = (*lbIter)->outputList.begin(); outTermListIter != (*lbIter)->outputList.end(); outTermListIter++){
	(*outTermListIter)->departureTime = 0.0;
#ifdef TIMING_DEBUG
	cout << "Set terminal " << (*outTermListIter)->name << " departure time to " << (*outTermListIter)->departureTime << endl;
#endif
	timingQueue.push(*outTermListIter);
      }
    }
  }

  //If we are pipeline aware, go through each net and find all delay registers.
  //Set their departure time to 0.0 and add to the queue
  if(pipeline_aware){
    for(netIter = netList.begin(); netIter != netList.end(); netIter++){
      if((*netIter)->pipelined){
	for(outTermListIter = (*netIter)->pipelineRegisterList.begin(); outTermListIter != (*netIter)->pipelineRegisterList.end(); outTermListIter++){
	  (*outTermListIter)->departureTime = 0.0;
#ifdef TIMING_DEBUG
	  cout << "Set terminal " << (*outTermListIter)->name << " departure time to " << (*outTermListIter)->departureTime << endl;
#endif
	  timingQueue.push(*outTermListIter);
	}
      }
    }
  }


  while(!timingQueue.empty()){
    //Until the queue is empty, take an output terminal out of the queue, and run through the
    //net that this output terminal sources.  
    //Using the linkDelayVector, determine the arrival time for all of the sink input terminals
    
    //Take output terminal out of queue
    outputTerminal = timingQueue.front();
#ifdef TIMING_DEBUG
    cout << "Popped terminal " << outputTerminal->name << endl;
#endif
    timingQueue.pop();
    
    //If we are not pipeline aware, all of these terminals should be output terminals
    //If we are though, some of these terminals will be delay nodes
    if(!pipeline_aware && !outputTerminal->isSource){
	ERROR("Timing Queue has a non-source terminal in it!");
    }

    //Go though all of the sinks of this net and set their arrival time via the output terminal's departure time and
    //the linkDelayVector
    //If we are pipeline aware, we should check to see if this terminal drives a pipelined net
    if(pipeline_aware && outputTerminal->destPipeliningNet){
      //This terminal drives a pipelined net
      //That means that the sinks could be logic blocks or delay nodes
      //cout << "Pipelined Net!" << endl;
      
      net = 0;
      pNet = outputTerminal->destPipeliningNet;

      for(sinkTermIter = pNet->sinkVector.begin(), i = 0; sinkTermIter != pNet->sinkVector.end(); sinkTermIter++, i++){
	//Set this sink's arrival time
	//Since this is a pipelined net, all we have to worry about whether or not to divide by the latency of the net
	//all pipelined nets have a latency of 0.
	(*sinkTermIter)->arrivalTime = outputTerminal->departureTime + pNet->linkDelayVector.at(i);
#ifdef TIMING_DEBUG
	cout << "Set sink " << (*sinkTermIter)->name << " arrival time to " << (*sinkTermIter)->arrivalTime << endl;
#endif

	//If we are doing "vanilla" timing, all blocks will be registered and we don't have to propagate the arrival time to the 
	//output terminal
	if(!vanilla_timing){
	  //If this sink is a delay node or belongs to an output stream, data reg or ram, we do not have to propagate this
	  //arrival time to output departure times
	  if((*sinkTermIter)->rNode->type != DELAY){
	    //This sink is some kind of logic block so we might have to propagate this arrival time to the output's departure time
	    //Check to see if this sink belongs to an output block/data reg/RAM.
	    lBlock = (*sinkTermIter)->lBlock;
	    if(lBlock->blockType != OUTPUTPAD && lBlock->blockType != RAM_BLOCK && lBlock->blockType != REGISTER_BANK){
	      //If it is not, go through each of the output terminals that this input terminal/sink fans out to
	      //and attempt to set their departure time
	      for(outTermListIter = (*sinkTermIter)->fanOut->begin(); outTermListIter != (*sinkTermIter)->fanOut->end(); outTermListIter++){
		if((*outTermListIter)->departureTime < (*sinkTermIter)->arrivalTime){
		  (*outTermListIter)->departureTime = (*sinkTermIter)->arrivalTime;
		  //Send out the timing of this terminal even though all of the inputs might not have checked in yet
		  //If they check in and force a later departure time, this will send out a wave of updates
		  timingQueue.push(*outTermListIter);

#ifdef TIMING_DEBUG
	cout << "Pushed source " << (*outTermListIter)->name << " into queue with departure time " << (*outTermListIter)->departureTime << endl;
#endif
		}
	      }
	    }
	  }
	}
      }
    }
    else{
      //This terminal drives a normal net so we can handle this normally
      net = outputTerminal->net;
      pNet = 0;
      
      //cout << "Unpipelined net!" << endl;
      
      for(sinkTermIter = net->sinkVector.begin(), i = 0; sinkTermIter != net->sinkVector.end(); sinkTermIter++, i++){
	//For unpipelined nets we have to treat vanilla and non-vanilla nets differently
	if(vanilla_timing){
	  //If we are doing "vanilla" timing, we will consider the arrival time of sinks to be
	  //the real arrival time divided by the desired latency + 1.
	  //This is OK even if we are doing pipeline aware routing since any non-pipeline nets will always have
	  //a latency of 0.
	  (*sinkTermIter)->arrivalTime = outputTerminal->departureTime + (net->linkDelayVector.at(i) / (net->latencyVector.at(i) + 1));
#ifdef TIMING_DEBUG
	  cout << "Set sink " << (*sinkTermIter)->name << " arrival time to " << (*sinkTermIter)->arrivalTime << endl;
#endif
	}
	else{
	  //If we are doing non-vanilla timing, we will need to set the arrival time of the sinks
	  //correctly and also check to see if we should propagate this to the output pins of this logic block
	  (*sinkTermIter)->arrivalTime = outputTerminal->departureTime + net->linkDelayVector.at(i);
#ifdef TIMING_DEBUG
	  cout << "Set sink " << (*sinkTermIter)->name << " arrival time to " << (*sinkTermIter)->arrivalTime << endl;
#endif

	  //Check to see if this sink belongs to an output block/data reg/RAM/delay block.
	  lBlock = (*sinkTermIter)->lBlock;
	  if(lBlock->blockType != OUTPUTPAD && lBlock->blockType != RAM_BLOCK && lBlock->blockType != REGISTER_BANK){
	    //If it is not, go through each of the output terminals that this input terminal/sink fans out to
	    //and attempt to set their departure time
	    for(outTermListIter = (*sinkTermIter)->fanOut->begin(); outTermListIter != (*sinkTermIter)->fanOut->end(); outTermListIter++){
	      if((*outTermListIter)->departureTime < (*sinkTermIter)->arrivalTime){
		(*outTermListIter)->departureTime = (*sinkTermIter)->arrivalTime;
		//Send out the timing of this terminal even though all of the inputs might not have checked in yet
		//If they check in and force a later departure time, this will send out a wave of updates
		timingQueue.push(*outTermListIter);
	      }
	    }
	  }
	}
      }
    }
  }
  //Go through all of the logic blocks and make sure that all input terminals have arrival times
  //and all output terminals have departure times
  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    //Check input terminals
    for(inTermListIter = (*lbIter)->inputList.begin(); inTermListIter != (*lbIter)->inputList.end(); inTermListIter++){
      if((*inTermListIter)->arrivalTime == -1.0E9){
	ERROR("Input terminal does not have arrival time!");
      }
      //The departure time for an input terminal is meaningless
      else if((*inTermListIter)->departureTime != -1.0E9){
	ERROR("Input terminal has a departure time set!");
      }
#ifdef TIMING_DEBUG
      cout << (*inTermListIter)->name << " arrival time " << (*inTermListIter)->arrivalTime << endl;
#endif
    }

    //Check output terminals
    for(outTermListIter = (*lbIter)->outputList.begin(); outTermListIter != (*lbIter)->outputList.end(); outTermListIter++){
      //The arrival time for an output terminal is meaningless
      if((*outTermListIter)->arrivalTime != -1.0E9){
	ERROR("Output terminal has arrival time set!");
      }
      else if((*outTermListIter)->departureTime == -1.0E9){
	ERROR("Output terminal does not have departure time!");
      }
#ifdef TIMING_DEBUG
      cout << (*outTermListIter)->name << "departure time " << (*outTermListIter)->departureTime << endl;
#endif
    }
  }

  //Go through all of the net's and find the delay registers.
  //Make sure that all of the arrival times have been set
  if(pipeline_aware){
    for(netIter = netList.begin(); netIter != netList.end(); netIter++){
      if((*netIter)->pipelined){
	for(outTermListIter = (*netIter)->pipelineRegisterList.begin(); outTermListIter != (*netIter)->pipelineRegisterList.end(); outTermListIter++){
	  if((*outTermListIter)->arrivalTime == -1.0E9){
	    ERROR("Delay register does not have arrival time!");
	  }
	}
      }
    }
  }
}

//Now that we know the critical path for the system we can calculate required times
void Netlist::calculateRequiredTimes(double criticalPath, bool pipeline_aware){
  list<logicBlock*>::iterator lbIter;
  list<Terminal*>::iterator sinkTermListIter;
  vector<Terminal*>::iterator sinkTermVectorIter;
  vector<double>::iterator requiredTimeIter;

  list<Terminal*>::iterator inTermIter;
  list<Terminal*>::iterator outTermIter;

  logicBlock *lBlock;

  Net *net;
  PipeliningNet *pNet;

  int i, j;
  bool termReady;
  double requiredTime;
  queue<Terminal *> timingQueue;
  Terminal* sinkTerminal;
  Terminal* sourceTerminal;
  Terminal* inputTerminal;

  list<Net*>::iterator netIter;

  //Build a queue of input terminals to process
  //Take all registered logic blocks and set their input terminal required times to the critical path delay
  //then add input terminal to the queue.
  //If we are calculating timing normally, this will include all output blocks, data regs and RAM input terminals.
  //If we are calculating "vanilla" timing, this will include all blocks
  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if(!vanilla_timing){
      if((*lbIter)->blockType == OUTPUTPAD || (*lbIter)->blockType == RAM_BLOCK || (*lbIter)->blockType == REGISTER_BANK){
	for(sinkTermListIter = (*lbIter)->inputList.begin(); sinkTermListIter != (*lbIter)->inputList.end(); sinkTermListIter++){
	  (*sinkTermListIter)->requiredTime = criticalPath;
#ifdef TIMING_DEBUG
	  cout << "Set " << (*sinkTermListIter)->name << " required time to " << (*sinkTermListIter)->requiredTime << endl;
#endif
	  timingQueue.push(*sinkTermListIter);
	}
      }
    
      //In addition, there may be some alus or multipliers in the netlist that have an output attached to nothing
      //These outputs actually go to the control structure, but they are not used in the datapath
      //Thus, we assume these inputs to have a required time of the CP
      else if((*lbIter)->outputList.empty()){
	for(sinkTermListIter = (*lbIter)->inputList.begin(); sinkTermListIter != (*lbIter)->inputList.end(); sinkTermListIter++){
	  (*sinkTermListIter)->requiredTime = criticalPath;
#ifdef TIMING_DEBUG
	  cout << "Set " << (*sinkTermListIter)->name << " required time to " << (*sinkTermListIter)->requiredTime << endl;
#endif
	  timingQueue.push(*sinkTermListIter);
	}
      }
    }
    else{
      for(sinkTermListIter = (*lbIter)->inputList.begin(); sinkTermListIter != (*lbIter)->inputList.end(); sinkTermListIter++){
	(*sinkTermListIter)->requiredTime = criticalPath;
#ifdef TIMING_DEBUG
	cout << "Set " << (*sinkTermListIter)->name << " required time to " << (*sinkTermListIter)->requiredTime << endl;
#endif
	timingQueue.push(*sinkTermListIter); 
      }
    }
  }

  //If we are performing pipeline-aware routing we should also set the required time of delay registers
  //Go through each net and find all delay registers. Set their required time to the CP and add to the queue
  if(pipeline_aware){
    for(netIter = netList.begin(); netIter != netList.end(); netIter++){
      if((*netIter)->pipelined){
	for(sinkTermListIter = (*netIter)->pipelineRegisterList.begin(); sinkTermListIter != (*netIter)->pipelineRegisterList.end(); sinkTermListIter++){
	  (*sinkTermListIter)->requiredTime = criticalPath;
#ifdef TIMING_DEBUG
	  cout << "Set " << (*sinkTermListIter)->name << " required time to " << (*sinkTermListIter)->requiredTime << endl;
#endif
	  timingQueue.push(*sinkTermListIter);
	}
      }
    }
  }


  while(!timingQueue.empty()){
    //Until the queue is empty, take a sink out of the queue and attempt to set the 
    //required time of it's source using the required time of the sink and the linkDelayVector.

    //1) Take input terminal out of queue
    sinkTerminal = timingQueue.front();
#ifdef TIMING_DEBUG
    cout << "Popped terminal " << sinkTerminal->name << endl;
#endif
    timingQueue.pop();
    
    //If we are not pipeline aware, all of these terminals should be input terminals
    //If we are though, some of these terminals will be delay nodes;
    if(!pipeline_aware && !sinkTerminal->isSink){
      ERROR("Timing Queue has a non-sink terminal in it!");
    }

    //If we are pipeline aware we should check to see if this sink is part of a pipelined net
    if(pipeline_aware && sinkTerminal->sourcePipeliningNet){
      //This terminal is driven by a pipelined net
      //That means that the source could be a logic block or a delay node
      //2) Grab source output Terminal of the net
      net = 0;
      pNet = sinkTerminal->sourcePipeliningNet;
      sourceTerminal = pNet->source;

      //If this source terminal is a delay node we should continue to the next node in the queue.
      //The required time should remain at CP
      if(sourceTerminal->rNode->type == DELAY){
	continue;
      }

      //Otherwise we should try to update the required time of the source node
      //Figure out where the sink terminal is in the sinkVector
      for(sinkTermVectorIter = pNet->sinkVector.begin(), i = 0; sinkTermVectorIter != pNet->sinkVector.end(); sinkTermVectorIter++, i++){
	if((*sinkTermVectorIter) == sinkTerminal){
	  break;
	}
      }
      
      //Make sure that we actually found the right terminal
      if(pNet->sinkVector.at(i) != sinkTerminal){
	ERROR("Could not find sink while calulating required times!");
      }
      
      //3) Based upon the required time of the sink terminal and the link delay, calculate the required time
      //   of the source terminal.
      //   Since this is a pipelined net, we don't have to worry about the
      //   difference between calculating the vanilla or non-vanilla required time
      requiredTime = sinkTerminal->requiredTime - pNet->linkDelayVector.at(i);
      
      //See if this source/sink pair defines an earlier required time than previous sinks
      if(sourceTerminal->requiredTime > requiredTime){
	//If so, reset the required time for the source
	sourceTerminal->requiredTime = requiredTime;
#ifdef TIMING_DEBUG
	cout << "Set source " << sourceTerminal->name << " required time to " << requiredTime << endl;
#endif

	//If we are performing vanilla timing analysis, we can stop here
	if(!vanilla_timing){
	  //Since we updated the source terminal's required time, we may have to propagate this value to the
	  //  inputs of the logic block
	  //See if this source terminal belongs to an input block, data reg or RAM
	  lBlock = sourceTerminal->lBlock;
	  if(lBlock->blockType != INPUTPAD && lBlock->blockType != RAM_BLOCK && lBlock->blockType != REGISTER_BANK){
	    //This is not an input block,etc., so go through all of the input terminals that depend on this source/output terminal
	    //and update their required times
	    for(inTermIter = sourceTerminal->dependency->begin(); inTermIter != sourceTerminal->dependency->end(); inTermIter++){
	      if((*inTermIter)->requiredTime > requiredTime){
#ifdef TIMING_DEBUG	
		cout << "Set sink " << (*inTermIter)->name << " required time to " << requiredTime << endl;
#endif
		
		(*inTermIter)->requiredTime = requiredTime;
		timingQueue.push(*inTermIter);
	      }
	    }
	  }
	}
      }
    }
    else{
      //This sink is driven by a normal net so we can handle this normally
      //2) Grab source output Terminal of the net
      net = sinkTerminal->net;
      pNet = 0;
      sourceTerminal = net->source;
      
      //Figure out where the sink terminal is in the sinkVector
      for(sinkTermVectorIter = net->sinkVector.begin(), i = 0; sinkTermVectorIter != net->sinkVector.end(); sinkTermVectorIter++, i++){
	if((*sinkTermVectorIter) == sinkTerminal){
	  break;
	}
      }
      
      //Make sure that we actually found the right terminal
      if(net->sinkVector.at(i) != sinkTerminal){
	ERROR("Could not find sink while calulating required times!");
      }
      
      //3) Based upon the required time of the sink terminal and the link delay, calculate the required time
      //   of the source terminal.
      //   If we are performing vanilla timing analysis, we have to divide the delay by the desired latency + 1
      if(vanilla_timing){
	requiredTime = sinkTerminal->requiredTime - (net->linkDelayVector.at(i) / (net->latencyVector.at(i) + 1));
#ifdef TIMING_DEBUG
	if(requiredTime < 0){
	  cout << "STOP, required time is negative " << requiredTime << endl;
	  cout << "Sink Terminal " << sinkTerminal->name << " required time " << sinkTerminal->requiredTime << endl;
	  cout << "Net " << net->name << " link delay = " << net->linkDelayVector.at(i);
	  cout << " latency " << net->latencyVector.at(i) << endl;
	}
#endif
      }
      else{
	requiredTime = sinkTerminal->requiredTime - net->linkDelayVector.at(i);
      }
      
      //See if this source/sink pair defines a lower required time than previous sinks
      if(sourceTerminal->requiredTime > requiredTime){
	//If so, reset the required time for the source
	sourceTerminal->requiredTime = requiredTime;
#ifdef TIMING_DEBUG
	cout << "Set source " << sourceTerminal->name << " required time to " << requiredTime << endl;
#endif
	
	//After we set the required time, we can just move on if we are performing vanilla timing anaylsis
	if(!vanilla_timing){
	  //Since we updated the source terminal's required time, we may have to propagate this value to the
	  //  inputs of the logic block
	  //See if this source terminal belongs to an input block, data reg or RAM
	  lBlock = sourceTerminal->lBlock;
	  if(lBlock->blockType != INPUTPAD && lBlock->blockType != RAM_BLOCK && lBlock->blockType != REGISTER_BANK){
	    //If this is not an input block,etc., go through all of the input terminals that depend on this source/output terminal
	    //and update their required times
	    for(inTermIter = sourceTerminal->dependency->begin(); inTermIter != sourceTerminal->dependency->end(); inTermIter++){
	      if((*inTermIter)->requiredTime > requiredTime){
#ifdef TIMING_DEBUG	
		cout << "Set sink " << (*inTermIter)->name << " required time to " << requiredTime << endl;
#endif
		
		(*inTermIter)->requiredTime = requiredTime;
		timingQueue.push(*inTermIter);
	      }
	    }
	  }
	}
      }
    }
  }
    
  //Go through all of the logic blocks and make sure that all terminals have required times!
  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    //Check input terminals
    for(inTermIter = (*lbIter)->inputList.begin(); inTermIter != (*lbIter)->inputList.end(); inTermIter++){
      if((*inTermIter)->requiredTime == 1.0E9){
	ERROR("Input terminal does not have required time!");
      }
#ifdef TIMING_DEBUG
      cout << (*inTermIter)->name << " required time "<< (*inTermIter)->requiredTime << endl;
#endif
    }
    
    //Check output terminals
    for(outTermIter = (*lbIter)->outputList.begin(); outTermIter != (*lbIter)->outputList.end(); outTermIter++){
      //The logic block output terminal for a self-feeding register will still have a non-set
      // required time. Don't worry about it, since the register itself will be fine.
      if((*outTermIter)->rNode->type == DELAY){
	continue;
      }

      //The arrival time for an output terminal is meaningless
      if((*outTermIter)->requiredTime == 1.0E9){
	cout << (*outTermIter)->name << endl;
	ERROR("Output terminal does not have required time!");
      }
#ifdef TIMING_DEBUG
      cout << (*outTermIter)->name << " required time "<< (*outTermIter)->requiredTime << endl;
#endif
    }
  }
  //Go though all of the data regs and make sure that their required times are still at the CP
  if(pipeline_aware){
    for(netIter = netList.begin(); netIter != netList.end(); netIter++){
      if((*netIter)->pipelined){
	for(sinkTermListIter = (*netIter)->pipelineRegisterList.begin(); sinkTermListIter != (*netIter)->pipelineRegisterList.end(); sinkTermListIter++){
	  if((*sinkTermListIter)->requiredTime != criticalPath){
	    ERROR("Delay register does not have correct required time!");
	  }
	}
      }
    }
  }
}

//Based upon arrival, required and link delays we can calculate slack for all links
void Netlist::calculateSlack(double criticalPathDelay, bool pipeline_aware){
  list<Net*>::iterator netIter;
  vector<Terminal*>::iterator sinkIter;

  int i;

  list<PipeliningNet*>::iterator pNetIter;

  //cout << "Critical Path = " << criticalPathDelay << endl;

  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
#ifdef TIMING_DEBUG
    cout << "Net " << (*netIter)->name << endl;
#endif
    //If a net is pipelined, we have to go through the pipelining nets instead, un
    if((*netIter)->pipelined && pipeline_aware){
      for(pNetIter = (*netIter)->pipelineNetList.begin(); pNetIter != (*netIter)->pipelineNetList.end(); pNetIter++){
	for(sinkIter = (*pNetIter)->sinkVector.begin(), i=0; sinkIter != (*pNetIter)->sinkVector.end(); sinkIter++, i++){
	  //Check to see if sink required time is still 1.0E9.
	  if((*sinkIter)->requiredTime == 1.0E9){
	    ERROR("Required time is not set!");
	  }
	  
	  else{
	    //If we are dealing with a pipelined net, we don't have to worry about whether or not we are doing 
	    //vanilla timing analysis.
	    (*pNetIter)->slackVector.at(i) = (*sinkIter)->requiredTime - (*pNetIter)->source->departureTime - (*pNetIter)->linkDelayVector.at(i);
#ifdef TIMING_DEBUG
	    cout << "Connection from " << (*pNetIter)->source->name << " to " << (*sinkIter)->name << " has slack " << (*pNetIter)->slackVector.at(i);
	    cout << " RT = " << (*sinkIter)->requiredTime << " DT = " << (*pNetIter)->source->departureTime << " delay = " <<  (*pNetIter)->linkDelayVector.at(i) << endl;
#endif
	  }
	  
	  if((*pNetIter)->slackVector.at(i) < 0.0){
#ifdef TIMING_DEBUG
	    cout << endl<< "Net " << (*pNetIter)->name << " sink " << (*pNetIter)->sinkVector.at(i)->name << " ( " << i << " of " << (*pNetIter)->sinkVector.size() << ") has delay "<< (*pNetIter)->linkDelayVector.at(i) << " and slack " << (*pNetIter)->slackVector.at(i) << endl;
#endif
	    //Due to round-off error, we can get slightly negative values that should be zero
	    if((*pNetIter)->slackVector.at(i) > -1.0E-20){
	    }
	    else{
	      cout << "Slack = " << (*pNetIter)->slackVector.at(i) <<endl;
	      ERROR("Very negative slack!");
	    }
	  }
	}
      }

      //Now that we have determined the slack for each sink in each pipelined net we can determine what the minimum slack is for each sink in the overall net.
      (*netIter)->calculatePipelinedSlacks();
    }
    else{
      for(sinkIter = (*netIter)->sinkVector.begin(), i=0; sinkIter != (*netIter)->sinkVector.end(); sinkIter++, i++){
	//Check to see if sink required time is still 1.0E9.
	if((*sinkIter)->requiredTime == 1.0E9){
	  ERROR("Required time is not set!");
	}
	
	else{
	  if(!vanilla_timing){
	    (*netIter)->slackVector.at(i) = (*sinkIter)->requiredTime - (*netIter)->source->departureTime - (*netIter)->linkDelayVector.at(i);
	  }
	  else{
	    (*netIter)->slackVector.at(i) = (*sinkIter)->requiredTime - (*netIter)->source->departureTime - 
	      ((*netIter)->linkDelayVector.at(i) / ((*netIter)->latencyVector.at(i) + 1));
	  }
#ifdef TIMING_DEBUG
	  cout << "Connection from " << (*netIter)->source->name << " to " << (*sinkIter)->name << " has slack " << (*netIter)->slackVector.at(i);
	  cout << " RT = " << (*sinkIter)->requiredTime << " DT = " << (*netIter)->source->departureTime << " delay = " <<  (*netIter)->linkDelayVector.at(i) << endl;
#endif
	}
	
	if((*netIter)->slackVector.at(i) < 0.0){
#ifdef TIMING_DEBUG
	  cout << endl<< "Net " << (*netIter)->name << " sink " << (*netIter)->sinkVector.at(i)->name << " ( " << i << " of " << (*netIter)->sinkVector.size() << ") has delay "<< (*netIter)->linkDelayVector.at(i) << " and slack " << (*netIter)->slackVector.at(i) << endl;
#endif
	  if((*netIter)->slackVector.at(i) > -1.0E-20){
	  }
	  else{
	    cout << "Slack = " << (*pNetIter)->slackVector.at(i) <<endl;
	    ERROR("Very negative slack!");
	  }
	}
      }
    }
  }
}

//Go through all nets and sort sinks in increasing slack
void Netlist::sort_netSinksTiming(){
  list<Net*>::iterator netIter;

  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->sort_sinkVectorTiming();
  }
}

//Go through the netlist and show what nets are critical
void Netlist::printCriticalPath(bool pipeline_aware){
  list<Net*>::iterator netIter;

  cout << endl << "List of critial or very nearly critical paths!" << endl;
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->printCriticalPath(pipeline_aware);
  }
}

void Netlist::clearLinkDelayAndSlackVectors(){
  for(list<Net*>::iterator netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->clearLinkDelayAndSlackVectors();
  }
}

// Goes through all of output blocks, data regs and RAMs to determine the latest arrival time on all of their input terminals
double Netlist::returnCriticalPathDelay(bool pipeline_aware){
  double criticalPathDelay;
  list<logicBlock*>::iterator lbIter;
  list<Terminal*>::iterator inTermIter;

  list<Net*>::iterator netIter;

  criticalPathDelay = 0.0;
  for(lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if(!vanilla_timing){
      //Find the arrival time of all registered blocks
      if((*lbIter)->blockType == OUTPUTPAD || (*lbIter)->blockType == RAM_BLOCK || (*lbIter)->blockType == REGISTER_BANK){
#ifdef TIMING_DEBUG
	cout << "Examining block " << (*lbIter)->name << " of type " << (*lbIter)->blockType << endl;
#endif

	for(inTermIter = (*lbIter)->inputList.begin(); inTermIter != (*lbIter)->inputList.end(); inTermIter++){
	  if((*inTermIter)->arrivalTime > criticalPathDelay){
	    criticalPathDelay = (*inTermIter)->arrivalTime;
#ifdef TIMING_DEBUG
	    cout << "\tNew Critical Path = " << criticalPathDelay << endl;
#endif
	  }
	}
      }
      //Sometimes, there will be an alu or multiplier that has no output.
      //This is because the output goes to the control logic only.
      //If this is the case, we should also check the arrival time of blocks that do not have any fan-out
      else if((*lbIter)->outputList.empty()){
#ifdef TIMING_DEBUG
	cout << "Examining block " << (*lbIter)->name << " of type " << (*lbIter)->blockType << " because it has no fan-out" << endl;
#endif
	for(inTermIter = (*lbIter)->inputList.begin(); inTermIter != (*lbIter)->inputList.end(); inTermIter++){
	  if((*inTermIter)->arrivalTime > criticalPathDelay){
	    criticalPathDelay = (*inTermIter)->arrivalTime;
#ifdef TIMING_DEBUG
	    cout << "\tNew Critical Path = " << criticalPathDelay << endl;
#endif
	  }
	}
      }
    }
    else{
#ifdef TIMING_DEBUG
      cout << "Examining block " << (*lbIter)->name << " of type " << (*lbIter)->blockType << endl;
#endif
	  
      for(inTermIter = (*lbIter)->inputList.begin(); inTermIter != (*lbIter)->inputList.end(); inTermIter++){
	if((*inTermIter)->arrivalTime > criticalPathDelay){
	  criticalPathDelay = (*inTermIter)->arrivalTime;
#ifdef TIMING_DEBUG
	  cout << "\tNew Critical Path = " << criticalPathDelay << endl;
#endif
	}
      }
    }
  }
  
  //If we are performing pipeline-aware routing we should also check the arrival time of all
  //of the delay registers
  if(pipeline_aware){
    for(netIter = netList.begin(); netIter != netList.end(); netIter++){
      if((*netIter)->pipelined){
	for(inTermIter = (*netIter)->pipelineRegisterList.begin(); inTermIter != (*netIter)->pipelineRegisterList.end(); inTermIter++){
	  if((*inTermIter)->arrivalTime > criticalPathDelay){
	    criticalPathDelay = (*inTermIter)->arrivalTime;
#ifdef TIMING_DEBUG
	    cout << "\tNew Critical Path = " << criticalPathDelay << endl;
#endif
	  }
	}
      }
    }    
  }

  return (criticalPathDelay);
}


// for each net in this netlist, sort the sinks in increasing order of separation from source  
void Netlist::sort_netSinksDistance(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  list<Net*>::iterator net_itr = netList.begin();
  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->sort_sinkVectorDistance();
}

// Create vector of pointers to logic blocks that comprise this netlist. 
// Note that this function can be called only AFTER lbList has been completely populated.
void Netlist::create_lbVector(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  int num_elements = numlogicBlocks + numInputs + numOutputs;
  lbVector=new vector<logicBlock*>(num_elements);
  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(int i=0;lb_itr!=lbList.end();lb_itr++,i++)
    lbVector->at(i)=*lb_itr;
}

// Return logic block at lbVector[i]
logicBlock* Netlist::get_logicBlock(int i){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(i >= lbVector->size())
    ERROR("lbVector overflow here...");
  return lbVector->at(i);
}

// Add p_net to list of nets and increment num_nets.
void Netlist::addto_netList(Net* p_net){
  if(!p_net)
    ERROR("net cannot be NULL...");
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  netList.push_back(p_net);
  numNets++;
}

// Get total num logic blocks in this netlist.
int Netlist::get_numlogicBlocks(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numlogicBlocks;
}

int Netlist::countLUT(){
  int count = 0;

  create_lbListItr();
  for(logicBlock* lb=getnext_lbList();lb!=0;lb=getnext_lbList()){
    if(lb->get_blockType() == LUT)
      count++;
  }
  remove_lbListItr();
  return (count);
}

int Netlist::countMult(){
  int count = 0;

  create_lbListItr();
  for(logicBlock* lb=getnext_lbList();lb!=0;lb=getnext_lbList()){
    if(lb->get_blockType() == MULTIPLIER)
      count++;
  }
  remove_lbListItr();
  return (count);
}

int Netlist::countFixedReg(){
  int count = 0;

  create_lbListItr();
  for(logicBlock* lb=getnext_lbList();lb!=0;lb=getnext_lbList()){
    if(lb->get_blockType() == REGISTER_BANK)
      count++;
  }
  remove_lbListItr();
  return (count);
}

int Netlist::countRAM(){
 int count = 0;

  create_lbListItr();
  for(logicBlock* lb=getnext_lbList();lb!=0;lb=getnext_lbList()){
    if(lb->get_blockType() == RAM_BLOCK)
      count++;
  }
  remove_lbListItr();
  return (count);
}

// Get total num input-only IO blocks in this netlist.
int Netlist::get_numInputs(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get total num output-only IO blocks in this netlist.
int Netlist::get_numOutputs(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Get total num nets in this netlist.
int Netlist::get_numNets(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return numNets;
}

// Get the name of this netlist
string Netlist::get_name(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Print details of each logic block in lbList.
void Netlist::print_lbList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT LOGIC BLOCK LIST";
  cout<<endl<<"***************************************************************";

  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(;lb_itr!=lbList.end();lb_itr++)
    (*lb_itr)->print_Info();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT LOGIC BLOCK LIST";
  cout<<endl<<"***************************************************************";

}

// Print details of each net in netList.
void Netlist::print_netList(bool pipeline_aware){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT NET LIST";
  cout<<endl<<"***************************************************************";

  list<Net*>::iterator net_itr = netList.begin();
  for(;net_itr!=netList.end();net_itr++)
    (*net_itr)->print_Info(pipeline_aware);

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT NET LIST";
  cout<<endl<<"***************************************************************";

}

// Print the name of this netlist.
void Netlist::print_name(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  cout<<endl<<endl<<"NAME OF NETLIST IS "<<name;
}

// Print the details of logic block.
void Netlist::print_logicBlock(logicBlock* p_block){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!p_block)
    ERROR("logic block cannot be NULL...");
  p_block->print_Info();
}

// Print the details of net pointed to by p_net.
void Netlist::print_Net(Net* p_net, bool pipeline_aware){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!p_net)
    ERROR("net cannot be NULL...");
  p_net->print_Info(pipeline_aware);
}

// Look for logic block with block_name in lbList. If found, return the logic block.
logicBlock* Netlist::found_logicBlock(char* block_name){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(!block_name)
    ERROR("logic block name cannot be NULL...");
  list<logicBlock*>::iterator lb_itr = lbList.begin();
  for(;lb_itr!=lbList.end();lb_itr++){
    if(strcmp((*lb_itr)->get_name().c_str(),block_name)==0)
      return *lb_itr;
  }
  return 0;
}

// Look for net with net_name in netList. If found, return the net.
Net* Netlist::found_Net(string net_name){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  list<Net*>::iterator net_itr = netList.begin();
  for(;net_itr!=netList.end();net_itr++){
    if(strcmp((*net_itr)->get_name().c_str(),net_name.c_str())==0)
      return *net_itr;
  }
  return 0;
}

// Create a new iterator for lbList
void Netlist::create_lbListItr(){
  //cout << "Creating LB list iter" << endl;
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(lb_itr)
    ERROR("logic block list iterator already exists...");
  lb_itr=new list<logicBlock*>::iterator;
  *lb_itr = lbList.begin();
}

// Remove an already existing lbList iterator 
void Netlist::remove_lbListItr(){
  //cout << "Removing LB list iter" << endl;
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(lb_itr==0)
    ERROR("trying to delete NULL logic block list iterator...");
  delete lb_itr;
  lb_itr=0;
}

// Create a new iterator for netList
void Netlist::create_netListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(net_itr)
    ERROR("net list iterator already exists...");
  net_itr=new list<Net*>::iterator;
  *net_itr = netList.begin();
}

// Remove an already existing netList iterator 
void Netlist::remove_netListItr(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");
  if(net_itr==0)
    ERROR("trying to delete NULL net iterator...");
  delete net_itr;
  net_itr=0;
}

// Get next element from lbList
logicBlock* Netlist::getnext_lbList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  if(*lb_itr == lbList.end())
    return 0;
  else{
    logicBlock* ret_ptr = **lb_itr;
    (*lb_itr)++;
    return ret_ptr;
  }

}

// Get next element from netList
Net* Netlist::getnext_netList(){
  if(magicNum!=18111)
    ERROR("trying to access a deleted node...");

  if(*net_itr == netList.end())
    return 0;
  else{
    Net* ret_ptr = **net_itr;
    (*net_itr)++;
    return ret_ptr;
  }

}

//Go through all of the nets in the netlist and count the latency at each sink
int Netlist::countDelayReg(){
  list<Net*>::iterator netIter;
  vector<int>::iterator latencyIter;
  int maxLatency;
  
  int count = 0;
  
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    maxLatency = 0;
    for(latencyIter = (*netIter)->latencyVector.begin(); latencyIter != (*netIter)->latencyVector.end(); latencyIter++){
      if(*latencyIter > maxLatency){
	maxLatency = *latencyIter;
      }
    }

    (*netIter)->netMaxLatency = maxLatency;
    count += maxLatency;
  }

  return (count);
}

//Go through all of the nets in the netlist and sort the sinks
//based on non-decreasing pipeline latency.  How we handle equal latency signals
//depends on the sortByTiming variable.  If it is false, we use the normal
//A* distance to tie-break.  If it is true, we use timing criticality to break the tie.
//Criticality is obvious for non-pipelined signals, and the
//criticality of a pipelined signal is determined by the minimum slack of
//all its links
void Netlist::sort_netPipelineSinks(bool sortByTiming){
  list<Net*>::iterator netIter;
  Terminal *tempSink;
  int low, high;
  int tempInt;
  double tempDouble;
  int rowLow, colLow, rowHigh, colHigh;
  int sourceRow, sourceCol;
  bool swap;
  
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    //cout << "Net " << (*netIter)->name << endl;
    for(low = 0; low < (*netIter)->sinkVector.size(); low++){
      for(high = low + 1; high < (*netIter)->sinkVector.size(); high++){
	swap = false;
	//cout << "   Considering net " << (*netIter)->name << " sink " << low << " slack = " << (*netIter)->slackVector.at(low);
	//cout << " and sink " << high << " slack = " << (*netIter)->slackVector.at(high)<< endl;
	if((*netIter)->latencyVector.at(low) > (*netIter)->latencyVector.at(high)){
	  swap = true;
	}
	else if((*netIter)->latencyVector.at(low) == (*netIter)->latencyVector.at(high)){
	  //The low and high sinks have the same latency so we have to tie-break
	  if(sortByTiming){
	    //If we are sorting these based on timing, we determine which signal that has the
	    //larger slack.  If the sink is a pipelined sink we place the minimum slack of all links
	    //into the slackVector.  If it is not a pipelined sink, we use the slack vector as usual
	    if((*netIter)->slackVector.at(low) > (*netIter)->slackVector.at(high)){
	      swap = true;
	    }
	  }
	  else{
	    //If we are not timing-driven, we determine which signal is closer to the source
	    sourceRow = (*netIter)->source->rNode->row_low;
	    sourceCol = (*netIter)->source->rNode->col_low;
	    rowLow = (*netIter)->sinkVector.at(low)->rNode->row_low;
	    colLow = (*netIter)->sinkVector.at(low)->rNode->col_low;
	    rowHigh = (*netIter)->sinkVector.at(high)->rNode->row_low;
	    colHigh = (*netIter)->sinkVector.at(high)->rNode->col_low;
	    //Swap these if the sink at low is farther away from the source than the sink at high
	    if((abs(sourceRow-rowLow) + abs(sourceCol-colLow)) > (abs(sourceRow-rowHigh) + abs(sourceCol-colHigh))){
	      swap = true;
	    }
	  }
	}

	if(swap){
	  //cout << "     Swapping net " << (*netIter)->name << " sink " << low << " slack = " << (*netIter)->slackVector.at(low);
	  //cout << " and sink " << high << " slack = " << (*netIter)->slackVector.at(high)<< endl;
	  //We have decided to swap these sinks so we have to swap all of their information at the same time
	  //1) Change positions in sinkVector 
	  tempSink = (*netIter)->sinkVector.at(high);
	  (*netIter)->sinkVector.at(high) = (*netIter)->sinkVector.at(low);
	  (*netIter)->sinkVector.at(low) = tempSink;
	  //2) Swap values in linkDelayVector
	  tempDouble = (*netIter)->linkDelayVector.at(high);
	  (*netIter)->linkDelayVector.at(high) = (*netIter)->linkDelayVector.at(low);
	  (*netIter)->linkDelayVector.at(low) = tempDouble;
	  //3) Swap values slackVector
	  tempDouble = (*netIter)->slackVector.at(high);
	  (*netIter)->slackVector.at(high) = (*netIter)->slackVector.at(low);
	  (*netIter)->slackVector.at(low) = tempDouble;
	  //4) Swap values in latencyVector
	  tempInt = (*netIter)->latencyVector.at(high);
	  (*netIter)->latencyVector.at(high) = (*netIter)->latencyVector.at(low);
	  (*netIter)->latencyVector.at(low) = tempInt;
	}
      }
    }
  }
}

//Go through all of the nets and pipelined nets in the netlist.
//If we are doing congestion-driven routing, sort based on Manhattan distance.
//If we are doing timing-driven routing, sort based on the slack
void Netlist::sort_netPipelineSinks_LockedRegs(bool sortByTiming){
  list<Net*>::iterator netIter;
  Terminal *tempSink;
  int low, high;
  int tempInt;
  double tempDouble;
  int rowLow, colLow, rowHigh, colHigh;
  int sourceRow, sourceCol;
  bool swap;
  
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    //cout << "Net " << (*netIter)->name << endl;
    if((*netIter)->pipelined){
      //Go through all of the pipelined nets and sort them
      (*netIter)->sortLockedRegs(sortByTiming);
    }
    else{
      if(sortByTiming){
	(*netIter)->sort_sinkVectorTiming();
      }
      else{
	(*netIter)->sort_sinkVectorDistance();
      }
    }
  }
}

//Go through all nets and increment all nodes in routing trees
void Netlist::reCalculateOccupancy(bool pipeline_aware){
  list<Net*>::iterator netIter;

  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->reCalculateOccupancy(pipeline_aware);
  }
}

//This attempt is very simplistic, but we are trying to determine what the
//lower bound on the pipelined frequency of the circuit might be.
//
double Netlist::calculateFakePipelinedDelay(){
  


}

// Calculate the placement cost for the netlist from scratch
void Netlist::calculatePlacementCost(bool timing_driven, double criticalPathDelay){
  list<Net*>::iterator netIter;

  wireCost = 0.0;
  timingCost = 0.0;
  if(maxCutCost){
    maxCut = 0.0;
    aveCut = 0.0;

    //Go through the cut cost vector and zero out everything
    FPGA->clearCutCostVector();
  }
  
  //Calculate the wire and timing cost for all nets
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->calculatePlacementCost(timing_driven, criticalPathDelay);
    wireCost += (*netIter)->wireCost;
    timingCost += (*netIter)->timingCost;

    //printf("Net %s has Wire Cost = %f\n", (*netIter)->name, (*netIter)->wireCost);
    //printf("Net %s has Timing Cost = %e\n", (*netIter)->name, (*netIter)->timingCost);
  }

  if(maxCutCost){
    //Go through the cut cost vector to determine the largest value
    for(vector <double>::iterator iter = FPGA->cutCostVector.begin(); iter != FPGA->cutCostVector.end(); iter++){
      //printf("Cut Cost = %d\n", (int) *iter);
      if(aveCutCost){
	aveCut += *iter;
      }

      if(maxCut < *iter){
	maxCut = *iter;
      }
    }
    if(aveCutCost){
      aveCut /= FPGA->cutCostVector.size();
    }
  }
}

// //This will calculate the delta cost for all of the nets
// double Netlist::determineDeltaCostAllNets(bool timing_driven, double criticalPathDelay){
//   list<Net*>::iterator netIter;

//   double deltaWire = 0.0;
//   double deltaTiming = 0.0;
//   double deltaCut = 0.0;
//   double deltaCost = 0.0;
  
//   for(netIter = netList.begin(); netIter != netList.end(); netIter++){
//     //Go through all of the nets and update everyone's wire and timing costs
//     (*netIter)->reCalculatePlacementCost(timing_driven, criticalPathDelay);
//   }

//   for(netIter = netList.begin(); netIter != netList.end(); netIter++){
//     // Now calculate the delta cost for all of the nets
//     deltaWire += (*netIter)->getDeltaWireCost();
//     (*netIter)->overWriteOldWireCost();
//     deltaTiming += (*netIter)->getDeltaTimingCost();
//     (*netIter)->overWriteOldTimingCost();
//   }

//   if(timing_driven){
//     //If we are timing-driven, we have to consider both the timing impact and wirelength impact of this move
//     deltaCost = tdPlacementLambda * (deltaTiming / getTimingCost()) + ((1.0 - tdPlacementLambda) * (deltaWire / getWireCost()));
//     setWireCost(deltaWire + getWireCost());
//     setTimingCost(deltaTiming + getTimingCost());
//     //cout << "Delta Wire Cost = " << deltaWire << endl;
//     //cout << "Wire Cost = " << getWireCost() << endl;
//     //cout << "Delta Timing Cost = " << deltaTiming << endl;
//     //cout << "Timing Cost = " << getTimingCost() << endl << endl <<endl;

//     //cout << "DeltaCost = " << deltaCost << endl;
//     return(deltaCost);
//   }
//   else{
//     //If we are not timing-driven, only consider the wirelength impact
//     setWireCost(deltaWire + getWireCost());
//     //cout << "DeltaCost = " << deltaWire << endl;
//     return deltaWire;
//   }
// }

//Go through the netlist and figure out what pipeline nets are eligable to be unzipped.
//We can unzip a net if 
// void Netlist::fillUnZipVector(vector<unZipWidget> *unZipVector){
// }

// Remove the placement for this netlist from the architecture
void Netlist::removePlacement(){
  create_lbListItr();
  for(logicBlock* lb = getnext_lbList(); lb != 0; lb = getnext_lbList()){  
    if(lb->get_blockType() == GBT_NO_TYPE){
      ERROR("Block should have type!");
    }
    else if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      unPlaceBlock(lb, (void *) FPGA->getIOBAt(lb->row, lb->col), IOPIN, lb->ptc_index);
    }
    else if(lb->get_blockType() == REGISTER_BANK){      
      ERROR("Register Banks not supported yet!");
    }
    else if(lb->get_blockType() == LUT){
      unPlaceBlock(lb, (void *) FPGA->getCLBAt(lb->row, lb->col), LB, lb->ptc_index);
    }
    else{
      ERROR("Unknown block type!");
    }
  }
  remove_lbListItr();
}

void Netlist::updatePlacementTiming(){
  Net *net;

  //cout << "Updating timing based on placement!" <<endl;
  //Go through all of the nets and update their timing based on this placement
  create_netListItr();
  for(net = getnext_netList(); net != 0; net = getnext_netList()){
    net->updatePlacementTiming();
  }
  remove_netListItr();
}

//This not accounts for not only the normal logic blocks and monolithic register banks
// that are in the normal nets, this also accounts for the pipelining registers
// created by pipelined nets.
int Netlist::calculateNumBlocksAndPipeliningRegisters(){
  int blockCount = getNumPipeliningRegisters();

  blockCount = blockCount + numlogicBlocks + numInputs + numOutputs;
  
  return (blockCount);
}

int Netlist::calculateNumNetsAndPNets(){
  int netCount = 0;

  for(list<Net*>::iterator iter = netList.begin(); iter != netList.end(); iter++){
    netCount ++;
    netCount += (*iter)->getNumPipelinedNets();
  }
  return netCount;
}

//This goes through the netlist and pulls out a random pipelining register from one of the pipelined nets
Terminal* Netlist::getRandomPipeliningRegister(){
  return(placementPipelineRegisterVector.at(lrand48() % placementPipelineRegisterVector.size()));
}

//Go through this net and add and pipelining registers to the overall netlist list
void Netlist::addPipelinedRegistersToPlacementList(Net *net){
  for(list<Terminal *>::iterator iter = net->pipelineRegisterList.begin(); iter != net->pipelineRegisterList.end(); iter++){
    placementPipelineRegisterVector.push_back(*iter);
  }
}

//Go through all of the nets and zero out their slack vectors
void Netlist::zeroSlackVectors(){
  for(list<Net*>::iterator iter = netList.begin(); iter != netList.end(); iter++){
    (*iter)->zeroSlackVectors();
  }
}

void Netlist::printPlacement(){
  // Go through each block and print out it's location
  for(list<logicBlock*>::iterator iter = lbList.begin(); iter != lbList.end(); iter++){
    if((*iter)->get_blockType() == INPUTPAD){
      cout << "Input pin " << (*iter)->name << " is placed in location (" << (*iter)->col << ", ";
      cout << (*iter)->row << ", " << (*iter)->ptc_index << ")" <<endl;
    }
    else if((*iter)->get_blockType() == OUTPUTPAD){
      cout << "Output pin " << (*iter)->name << " is placed in location (" << (*iter)->col << ", ";
      cout << (*iter)->row << ", " << (*iter)->ptc_index << ")" <<endl;
    }
    else if((*iter)->get_blockType() == LUT || (*iter)->get_blockType() == RAM_BLOCK || (*iter)->get_blockType() == MULTIPLIER){
      cout << "Logic block " << (*iter)->name << " is placed in location (" << (*iter)->col << ", " << (*iter)->row <<")" <<endl;
    }
    else if((*iter)->get_blockType() == REGISTER_BANK){
      cout << "Forced explict register " << (*iter)->name << " is placed in a register bank at location (" << (*iter)->col << ", ";
      cout << (*iter)->row << ") which is inside delay block at (" << (*iter)->rb->getHostCLB()->col << ", ";
      cout << (*iter)->rb->getHostCLB()->row << ") bank " << (*iter)->rb->getHostBankNum() << endl;
    }
    else{
      ERROR("Unknown block type!");
    }
  }

  for( vector <Terminal *>::iterator iter = placementPipelineRegisterVector.begin(); iter !=  placementPipelineRegisterVector.end(); iter++){
    cout << "Delay Register " << (*iter)->name << " is placed in delay block location (" 
	 << (*iter)->getRegisterBank()->getHostCLB()->col << ", ";
    cout << (*iter)->getRegisterBank()->getHostCLB()->row <<") bank " << (*iter)->rb->getHostBankNum() << endl;
  }
}

void Netlist::removePipelineRegisterFromPlacementPRVector(Terminal *pipeReg){
  vector<Terminal*>::iterator rvIter;
  
  for(rvIter = placementPipelineRegisterVector.begin(); 
      rvIter != placementPipelineRegisterVector.end(); rvIter++){
    if(*rvIter == pipeReg){
      break;
    }
  }
  
  if(rvIter == placementPipelineRegisterVector.end()){
    ERROR("Register not found!");
  }
  else{
    placementPipelineRegisterVector.erase(rvIter);
  } 
}

void Netlist::checkNets(){
  //Go through all of the nets and make sure that all pipelining nets have the correct parent net
  for(list<Net*>::iterator netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->checkNet();
  }

  //Go through all of the logic blocks and make sure that the number of registers we can push forwards
  // and backwards is correct.
  if(placementRetiming){
    for(list<logicBlock*>::iterator iter = lbList.begin(); iter != lbList.end(); iter++){
      (*iter)->checkPushForwardsAndBackwards();
    }
  }
}

//Print out the new netlist assuming each block is in a different CLB;
void Netlist::printNewNetlist(int randomSeed, char* rrGraphName, char* arguments){
  char filename [256];
  FILE *fp;
  int i;

  sprintf(filename,"%s_%s_%s_%ld.postplacementNetlist",get_name().c_str(), rrGraphName, arguments, randomSeed);
  fp=fopen(filename,"w");
  
  //Print out all of the input pads
  for(list<logicBlock*>::iterator lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() == INPUTPAD){
      fprintf(fp, ".input %s\n", (*lbIter)->get_name().c_str());
      for(list<Terminal*>::iterator outPinIter = (*lbIter)->outputList.begin(); outPinIter != (*lbIter)->outputList.end(); outPinIter++){
	if((*outPinIter)->destPipeliningNet){
	  fprintf(fp, "pinlist: %s\n\n", (*outPinIter)->destPipeliningNet->get_name().c_str());
	}
	else{
	  fprintf(fp, "pinlist: %s\n\n", (*outPinIter)->net->get_name().c_str());
	}
      }

    }
  }
  //Print out all of the output pads
  for(list<logicBlock*>::iterator lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() == OUTPUTPAD){
      fprintf(fp, ".output %s\n", (*lbIter)->get_name().c_str());
      for(list<Terminal*>::iterator inPinIter = (*lbIter)->inputList.begin(); inPinIter != (*lbIter)->inputList.end(); inPinIter++){
	if((*inPinIter)->sourcePipeliningNet){
	  fprintf(fp, "pinlist: %s\n", (*inPinIter)->sourcePipeliningNet->get_name().c_str());
	}
	else{
	  fprintf(fp, "pinlist: %s\n", (*inPinIter)->net->get_name().c_str());
	}
	fprintf(fp, "delay: 0\n\n");
      }
    }
  }

 //Print out all of the logic blocks
  for(list<logicBlock*>::iterator lbIter = lbList.begin(); lbIter != lbList.end(); lbIter++){
    if((*lbIter)->get_blockType() != INPUTPAD && (*lbIter)->get_blockType() != OUTPUTPAD){
      fprintf(fp, ".clb %s\n", (*lbIter)->get_name().c_str());
      if((*lbIter)->get_blockType() == LUT){
	fprintf(fp, ".type CLB\n");
	fprintf(fp, ".subblock %s\n", (*lbIter)->get_name().c_str());
	fprintf(fp, "pinlist: ");

	//Dump all of the inputs
	if((*lbIter)->inputList.size() < 1){
	  ERROR("SOMETHING IS WRONG!");
	    }	
	i = 0;
	for(list<Terminal*>::iterator inPinIter = (*lbIter)->inputList.begin(); inPinIter != (*lbIter)->inputList.end(); inPinIter++){
	  if((*inPinIter)->sourcePipeliningNet){
	    fprintf(fp, "%s ", (*inPinIter)->sourcePipeliningNet->get_name().c_str());
	  }
	  else{
	    fprintf(fp, "%s ", (*inPinIter)->net->get_name().c_str());
	  }
	  i++;
	}
	//There might be some open inputs
	for(; i < 4; i++){
	  fprintf(fp, "open ");
	}

	//Dump all of the outputs
	if((*lbIter)->outputList.size() != 1){
	  ERROR("SOMETHING IS WRONG!");
	}
	for(list<Terminal*>::iterator outPinIter = (*lbIter)->outputList.begin(); outPinIter != (*lbIter)->outputList.end(); outPinIter++){
	  if((*outPinIter)->destPipeliningNet){
	    fprintf(fp, "%s\n", (*outPinIter)->destPipeliningNet->get_name().c_str());
	  }
	  else{
	    fprintf(fp, "%s\n", (*outPinIter)->net->get_name().c_str());
	  }
	  i++;
	}
	fprintf(fp, "delay: 0 0 0 0\n\n");
      }
      else if((*lbIter)->get_blockType() == REGISTER_BANK){
	fprintf(fp, ".type CLB\n");
	fprintf(fp, ".subblockREG %s\n", (*lbIter)->get_name().c_str());
	fprintf(fp, "pinlist: ");

	//Dump all of the inputs
	if((*lbIter)->inputList.size() != 1){
	  ERROR("SOMETHING IS WRONG!");
	}
	for(list<Terminal*>::iterator inPinIter = (*lbIter)->inputList.begin(); inPinIter != (*lbIter)->inputList.end(); inPinIter++){
	  if((*inPinIter)->sourcePipeliningNet){
	    fprintf(fp, "%s ", (*inPinIter)->sourcePipeliningNet->get_name().c_str());
	  }
	  else{
	    fprintf(fp, "%s ", (*inPinIter)->net->get_name().c_str());
	  }
	}

	//Dump all of the outputs
	if((*lbIter)->outputList.size() != 1){
	  ERROR("SOMETHING IS WRONG!");
	}
	for(list<Terminal*>::iterator outPinIter = (*lbIter)->outputList.begin(); outPinIter != (*lbIter)->outputList.end(); outPinIter++){
	  if((*outPinIter)->destPipeliningNet){
	    fprintf(fp, "%s\n", (*outPinIter)->destPipeliningNet->get_name().c_str());
	  }
	  else{
	    fprintf(fp, "%s\n", (*outPinIter)->net->get_name().c_str());
	  }
	}
	fprintf(fp, "delay: 0\n\n");
      }
      //else if((*lbIter)->get_blockType() == RAM_BLOCK){
      //	fprintf(fp, ".type RAM\n");
      //}
      //else if((*lbIter)->get_blockType() == MULTIPLIER){
      //	fprintf(fp, ".type RAM\n");
      //}
      else{
	ERROR("Unknown block type!");
      }
    }
  }

  //Dump all of the pipelining registers
  for(vector <Terminal *>::iterator prIter = placementPipelineRegisterVector.begin(); 
      prIter != placementPipelineRegisterVector.end(); prIter++){
      fprintf(fp, ".clb %s\n", (*prIter)->get_name().c_str());

      fprintf(fp, ".type CLB\n");
      fprintf(fp, ".subblockREG %s\n", (*prIter)->get_name().c_str());
      fprintf(fp, "pinlist: ");

      //Dump all of the inputs
      if((*prIter)->sourcePipeliningNet){
	fprintf(fp, "%s ", (*prIter)->sourcePipeliningNet->get_name().c_str());
      }
      else{
	ERROR("Pipelining register has no source net!");
      }
      if((*prIter)->destPipeliningNet){
	fprintf(fp, "%s\n", (*prIter)->destPipeliningNet->get_name().c_str());
      }
      else{
	ERROR("Pipelining register has no destination net!");
      }
      
      fprintf(fp, "delay: 0\n\n");
  }
}

void Netlist::printNetWireCost(bool pipeline_aware){
  list<Net*>::iterator netIter;  
  list<PipeliningNet*>::iterator pNetIter;

  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    if(pipeline_aware && (*netIter)->pipelined){
      for(pNetIter = (*netIter)->pipelineNetList.begin(); pNetIter != (*netIter)->pipelineNetList.end(); pNetIter++){
	cout << "Net " << (*pNetIter)->name;
	cout << " wire cost = " << (*pNetIter)->wireCost << endl;
      }
    }
    else{
      cout << "Net " << (*netIter)->name;
      cout << " wire cost = " << (*netIter)->wireCost << endl;
    }
  }
}

//Calculate the average, max and minimum critialities of all pin-to-pin connections
void Netlist::calculateAvgMaxMinCrits(double criticalPathDelay){

}

//Go through the netlist and see if we have any registers
bool Netlist::needClock(){
  list<Net*>::iterator netIter;
  
  //Look for explicit registers in the netlist
  create_lbListItr();
  for(logicBlock* lb = getnext_lbList(); lb!=0; lb = getnext_lbList()){
    if(lb->get_blockType() == REGISTER_BANK){
      remove_lbListItr();
      return true;
    }
  }
  remove_lbListItr();
  
  //Look for pipelining registers
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    if((*netIter)->pipelineRegisterList.size() > 0){
      return true;
    }
  }
  return false;
}

//Go through all of the logic blocks and make sure that the mapping
// it thinks it has matches that of the physical array block
void Netlist::checkLogicBlockMappings(){
  int i;
  
  create_lbListItr();
  for(logicBlock* lb = getnext_lbList(); lb!=0; lb = getnext_lbList()){
    if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
      if(FPGA->getIOBAt(lb->row, lb->col)->getfrom_currBlockArray(lb->ptc_index) != lb){
	ERROR("Mismatching IOB placement!");
      }
    }
    else if(lb->get_blockType() == REGISTER_BANK){
      if(FPGA->getCLBAt(lb->row, lb->col)->getRegBank(lb->ptc_index)->get_currBlock() != lb){
	ERROR("Mismatching register placement!");
      }
    }
    else if(lb->get_blockType() == LUT){
      if(FPGA->getCLBAt(lb->row, lb->col)->getLUT(lb->ptc_index)->get_currBlock() != lb){
	ERROR("Mismatching LUT placement!");
      }
    }
    else{
      ERROR("Unknown logic block type");
    }
  }
  remove_lbListItr();  

  //Go through the list of pipeline registers and make sure they are all valid
  for(vector <Terminal *>::iterator prIter = placementPipelineRegisterVector.begin();
      prIter != placementPipelineRegisterVector.end(); prIter++){
    if((*prIter)->magicNum!=16111){
      ERROR("pipelining register has been deleted!");
    }
  }

//   for(list<Net*>::iterator netIter = netList.begin(); netIter != netList.end(); netIter++){
//     for(list<Terminal *>::iterator prIter = (*netIter)->pipelineRegisterList.begin(), i = 0;
// 	prIter != (*netIter)->pipelineRegisterList.end(); prIter++, i++){
//       if((*prIter)->getRegisterBank()->routingNodeVector( != (*prIter)->rNode)
//     }
//   }
}

//Go through this netlist and figure out what the max input and output
// criticality is for each of the logic blocks and pipelining registers in this netlist
void Netlist::calculateMaxInputOutputCriticalities(double criticalPathDelay){
  // Go through each block and print out it's location
  for(list<logicBlock*>::iterator iter = lbList.begin(); iter != lbList.end(); iter++){
    (*iter)->calculateMaxInputOutputCriticality(criticalPathDelay);
  }

  for(vector <Terminal *>::iterator iter = placementPipelineRegisterVector.begin(); iter !=  placementPipelineRegisterVector.end(); iter++){
    (*iter)->calculateMaxInputOutputCriticality(criticalPathDelay);
  }
}

//Go through all of the nets and print out the criticality of each connection
void Netlist::printAllSinkCriticalities(double criticalPathDelay, bool pipeline_aware){
  list<Net*>::iterator netIter;

  cout << endl << "Criticalities of all sinks" << endl;
  for(netIter = netList.begin(); netIter != netList.end(); netIter++){
    (*netIter)->printAllSinkCriticalities(criticalPathDelay, pipeline_aware);
  }
}

