// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#ifndef CONSTANTS_H
#define CONSTANTS_H

#define NDEBUG 1
//#define MAX_STR_LEN 256  //set the max length of a string
#define BIGINTEGER 10000000
//#define HUGE_NUMBER 1000000.0  //the value of a HUGE floating point initializer
/* #define PARSER_DEBUG 1 */  //switch on parser debugging
// #define PLACER_DEBUG 1 //switch on placer debugging
//#define PARANOID_CHECK_PIPEPLACE
//#define PIPEPLACEDEBUG
//#define PIPEPLACEDEBUGHEAVY
//#define BREAKONANNEALNUMBER
//#define ROUTER_DEBUG 1  //switch on router debugging
//#define PARANOID_CHECK_ROUTE
//#define PARANOID_CHECK_ROUTE_PIPELINE
//#define TIMING_DEBUG

//These are the types that routingNode can be.  routingNodes make up the vprArch
enum e_routingNodeType{RN_NO_TYPE, SOURCE, SINK, BUS_OR_WIRE, INPUT_PIN, OUTPUT_PIN, DELAY};

//These are the types of routing edges used (We don't actually use this information right now)
enum e_routingEdgeType{RE_NO_TYPE, BUFFERED, UNBUFFERED};

//These are the types of blocks that can be in the generic netlist.  Thus, they are also
// the types of blocks routing nodes can be a part of (source and sink of a LUT, etc)
//REGISTER_BANKs in the netlist are forced explicit registers.
//vprLUTs can be of type LUT, RAM_BLOCK or MULTIPLER
//I will evenutally change this to a vprCLB
enum e_genericBlockType{GBT_NO_TYPE, INPUTPAD, OUTPUTPAD, REGISTER_BANK, LUT, RAM_BLOCK, MULTIPLIER};

//These are the types of blocks that can be in a mcnc-specific netlist
enum e_mcncBlockType{MBT_NO_TYPE, MCNC_INPUT, MCNC_OUTPUT, MCNC_CLB, MCNC_LUT, MCNC_REG};

//These are the types of terminals we can have - notice we are not talking about
// the Terminal type, but the terminals of a generic block type (we also use the Terminal type
// for registers)
enum e_terminalType{T_NO_TYPE, INPUT_TERMINAL, OUTPUT_TERMINAL};


/* /\* define types of routing nodes - these are the allowable types in the rr_graph.echo file *\/ */
/* #define SOURCE 29 */
/* #define SINK 30 */
/* #define BUS_OR_WIRE 31 */
/* #define INPUT_PIN 32 */
/* #define OUTPUT_PIN 33 */
/* #define DELAY 34 */
/* //#define HELPER 35 */

/* // Define block types - these are the allowable types in the generic netlist */
/* #define NO_TYPE -1  //default value for logic block functionality type */
/* #define INPUTPAD 0 */
/* #define OUTPUTPAD 1 */
/* #define LUT 2 */
/* #define REGISTER_BANK 3  //These are fixed FF's - either because we are doing CLB-level placement or because they are */
/*                          //forced explict registers. */
/* #define RAM_BLOCK 4 */
/* #define MULTIPLIER 5 */

/* // These are the allowable types in the mcnc-specific netlist */
/* #define MCNC_INPUT 0 */
/* #define MCNC_OUTPUT 1 */
/* #define MCNC_CLB 2 */
/* #define MCNC_LUT 3 */
/* #define MCNC_REG 4 */

/* //These are types of terminals */
/* #define INPUT_TERMINAL 0 */
/* #define OUTPUT_TERMINAL 1 */

/* //#define MCNC_CLB_TYPE 0  //functionality type of an mcnc logic block that consists of a  */
/* //                          //4-LUT and a latch */
/* //#define MCNC_IO_TYPE 1  //functionality type of an mcnc IO pad */
/* //#define MCNC_OUTPUT_TERMINAL_TYPE 2  //mcnc output terminal type */
/* //#define MCNC_INPUT_TERMINAL_TYPE 3  //mcnc input terminal type */
/* //#define MCNC_DELAY_TYPE 4 */

/* define the direction factor ala VPR */
#define ALPHA 0.0
/* define the sharing factor i.e. the importance of congestion wrt wirelength */
#define SHARE_FAC 2.0 

#define NORTH 97
#define SOUTH 98
#define EAST 99
#define WEST 101

#define NUMASSUMEDCRIT 10

//Types of moves we can make
#define MOVENULL 997

//Basic swapping moves we can perform on a logic block
#define MOVELOGICBLOCK 998
#define MOVELOGICBLOCKSELF 999
#define MOVELOGICBLOCKALLLUTS 1000
#define MOVELOGICBLOCKCLB 1001

//Basic swapping moves we can perform on a pipeline register
#define MOVEPIPEREGSOMETHING 1002
#define MOVEPIPEREGSWAPSELF 1003
#define MOVEPIPEREGSWAPCLB 1004

//Pipeline register moves I'm not using right now
#define MOVEZIPPNET 1005
#define MOVEUNZIPPNET 1006
#define MOVEREDOUNZIPPNET 1007
#define MOVEREGEMPTYFIRST 1008
#define MOVEREGEMPTYLAST 1009
#define MOVEREGPARENT 1010
#define MOVEREGCHILD 1011

//Retime registers through a logic block
#define MOVEREGRETIMEBACK 1013
#define MOVEREGRETIMEFORWARD 1014
#define MOVEREGUNDOREGRETIMEBACK 1015
#define MOVEREGUNDOREGRETIMEFORWARD 1016
#define MOVEREGRETIMEBACKABORT 1017
#define MOVEREGRETIMEFORWARDABORT 1018

#define NUMVECTORBITS (8 * sizeof(unsigned int))

// These are the types of different blocks we need to worry about to calculate the timing
// tables for timing-driven placement
enum e_timingMoveBlockType{LB, IOPIN, REGISTERBANK};

#endif
