// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#ifndef GLOBAL_H
#define GLOBAL_H

#include "Netlist.h"
#include "mcncNetlist.h"
class mcncClb;
class vprFPGA;

/* curr_tok holds the current token type that is being parsed */
int curr_tok=0;

/* pinlist_tok holds the token type for which the pinlist is currently being read */
int pinlist_tok=0;

/* stringList is a list of strings (generally net names, 'open' and digits) */
list<char*> stringList;
list<char*>::iterator stringListItr;

/* curr_clb and curr_block store the objects whose pinlist are being processed */
mcncBlock* curr_clb=0;
mcncBlock* curr_block=0;

/* declare an mcnc_netlist */
mcncNetlist *mcnc_netlist;

// declare a generic netlist
Netlist* netlist;

// declare a vpr challenge FPGA
vprFPGA* FPGA;

map<string,routingNode*> rNodeMap;

/* declare the random seed for this run */
long int randomSeed=0;

/* this is the variable that will hold the current cost of the placement. the cost of a */
/* placement will be defined in terms of sharingCost and wireCost. */ 
double placementCost=0;

/* declare the variable that will hold the total sharing summed across all routing nodes */
/* in the FPGA. this is a cumulative measure of the extent to which the routing nodes are */
/* currently being shared */
long int sharingCost=0;
long int actual_sharingCost=0;

/* declare the variable that will hold the sum of the routing nodes used by the current */
/* route tree of each net in the netlist. this is a measure of the total wirelength of */
/* the current placement. */
//long int wireCost=0;
//long int actual_wireCost=0;

/* The routing node that is currently being read in from rr_graph.echo */
routingNode* G_currNode=0;
int G_currPTC = -1;

int MAXLATENCY = 0;
int MAXID = 0;
unsigned int *VECTORMASK;

vector <float> switches;

bool updateSlack = true;
bool deltaDelayCost = false;
bool pipeline_aware_placement = false;

bool lut_level_placement = false;
bool clb_level_placement = true;
double ff_level_placement_activation = -1.0;
double ff_level_placement_crit = 100.0;
double ff_level_placement_probability = 0.0;
double ff_level_placement_homing = 0.0;
bool carefulHoming = false;
bool lutStickiness = false;
bool tvpack_initial_placement = false;
bool detectOrphanRegisters = false;

bool placementRetiming = false;
double retime_placement_activation = -1.0;
double retime_placement_crit = 100.0;
double retime_placement_probability = 0.0;

bool splitDelayBlocks = false;
double splitDelayBlocks_activation = -1.0;
double splitDelayBlocks_crit = 100.0;
double splitDelayBlocks_probability = 0.0;

bool unzipNets = false;
double unzipNets_activation = -1.0;
double unzipNets_crit = 100.0;
double unzipNets_probability = 0.0;

int shiftRegisterMinPacking = 1000;
int shiftRegisterMaxPacking = 1000;

double tdPlacementLambda = 0.5;
double tdPlacementUpdateInterval = 1.0;
double tdPlacementCritExponentFinal = 8.0; //Exponent starts at 1.0, then goes to the final exponent value gradually.
double tdPlacementCritExponent = 1.0;
bool noWindowing = false;
double placeInnerLoop = 10.0;

bool pipeline_aware_routing = false;
bool lockRegPlacement = false;

bool cut_occ_reg = false;
bool maxCutCost = false;
bool aveCutCost = false;
bool critical_path_headroom = false;
double critical_path_headroom_fac = 0.0;
int pipelined_cost_model = 1;
bool vanilla_timing = false;
bool lowTempAnneal = false;
bool zeroCLBWireCost = false;
bool noIOAnneal = true;

bool testInitialPlacementOnly = false;

bool lotsTimingInfo = false;

#ifdef ROUTER_DEBUG
long routerPops = 0;
long routerPushes = 0;
#endif

#ifdef PARANOID_CHECK_PIPEPLACE
int numRegistersEstimatedFreed = 0;
int numRegistersEstimatedNeeded = 0;
#endif

int routeBoundingBox = 3;

double averageManhattanDistance = 0.0;

#endif
