// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

extern vprFPGA *FPGA;
extern bool tvpack_initial_placement;

logicBlock::~logicBlock(){
  if(magicNum!=15111)
    ERROR("trying to double delete a node...");
  else
    magicNum=0;
  
  //Delete the terminals of this logic block
  for(list<Terminal*>::iterator inputIter = inputList.begin(); inputIter != inputList.end(); inputIter++){
    delete *inputIter;
  }
  for(list<Terminal*>::iterator outputIter = outputList.begin(); outputIter != outputList.end(); outputIter++){
    delete *outputIter;
  }
}

// Create a new iterator for inputList
void logicBlock::create_inputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(interm_itr)
    ERROR("input terminal iterator already exists...");
  interm_itr=new list<Terminal*>::iterator;
  *interm_itr = inputList.begin();
}

// Create a new iterator for outputList
void logicBlock::create_outputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(outterm_itr){
    //cout << "Problem with block " << name << endl;
    ERROR("output terminal iterator already exists...");
  }
  outterm_itr=new list<Terminal*>::iterator;
  *outterm_itr = outputList.begin();

  //cout << "Creating outputListIter for block " << name << endl;
}

// Remove an already existing inputList iterator 
void logicBlock::remove_inputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(interm_itr==0)
    ERROR("trying to delete NULL input terminal iterator...");
  delete interm_itr;
  interm_itr=0;
}

// Remove an already existing outputList iterator 
void logicBlock::remove_outputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(outterm_itr==0)
    ERROR("trying to delete NULL output terminal iterator...");
  delete outterm_itr;
  outterm_itr=0;

  //cout << "Removing outputListIter for block " << name << endl;
}

// Get next element from inputList
Terminal* logicBlock::getnext_inputList(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  if(*interm_itr == inputList.end())
    return 0;
  else{
    Terminal* ret_ptr = **interm_itr;
    (*interm_itr)++;
    return ret_ptr;
  }

}

// Get next element from outputList
Terminal* logicBlock::getnext_outputList(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  if(*outterm_itr == outputList.end())
    return 0;
  else{
    Terminal* ret_ptr = **outterm_itr;
    (*outterm_itr)++;
    return ret_ptr;
  }

}

void logicBlock::set_blockType(enum e_genericBlockType t){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  blockType = t;
}

enum e_genericBlockType logicBlock::get_blockType(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return (blockType);
}

// Give this block a name
void logicBlock::set_name(string n){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  //strcpy(name,n);
  name = n;
}

void logicBlock::set_ptc_index(int n){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  ptc_index = n;
}

// Add a terminal to the input terminal list
void logicBlock::addto_inputList(Terminal* t){
  if(!t)
    ERROR("trying to add NULL terminal...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  inputList.push_back(t);
  numInputs++;
}

// Add a terminal to the output terminal list
void logicBlock::addto_outputList(Terminal* t){
  if(!t)
    ERROR("trying to add NULL terminal...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  outputList.push_back(t);
  numOutputs++;
}

// Get name of this block
string logicBlock::get_name(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get num input terminals
int logicBlock::get_numInputs(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get num output terminals
int logicBlock::get_numOutputs(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Print details about this logic block
void logicBlock::print_Info(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR CLB "<<name;

  //print IO or logic
  if(blockType == OUTPUTPAD)
    cout<<endl<<"OUTPUT BLOCK";
  else if(blockType == INPUTPAD)
    cout<<endl<<"INPUT BLOCK";
  else if(blockType == LUT)
    cout<<endl<<"LUT";
  else if(blockType == REGISTER_BANK)
    cout<<endl<<"REGISTER BANK";

  //print input list
  cout<<endl<<"INPUT TERMINALS -->";
  list<Terminal*>::iterator input_itr = inputList.begin();
  for(;input_itr!=inputList.end();input_itr++)
    (*input_itr)->print_Info();

  //print output list
  cout<<endl<<"OUTPUT TERMINALS -->";
  list<Terminal*>::iterator output_itr = outputList.begin();
  for(;output_itr!=outputList.end();output_itr++)
    (*output_itr)->print_Info();

  cout << endl << "Row = " << row;
  cout << endl << "Col = " << col << endl;

  if(blockType == LUT || blockType == REGISTER_BANK){
    cout << endl << "Parent CLB = " << parentCLB->name << endl;
  }

  cout << "Num reg can push backwards = " <<numRegCanPushBackward <<endl;
  cout << "Num reg can push forwards = " <<numRegCanPushForward <<endl;
}

//void logicBlock::initOutputDependency(){
//   outputDependency.resize(outputList.size());
// }

// void logicBlock::initInputFanOut(){
//   inputFanOut.resize(inputList.size());
// }

//Sets all outputs to depend upon all inputs
void logicBlock::setOutputDependency(){
  list<Terminal*>::iterator inputIter;
  list<Terminal*>::iterator outputIter;

  for(outputIter = outputList.begin(); outputIter != outputList.end(); outputIter++){
    (*outputIter)->dependency = new list <Terminal*>;
    for(inputIter = inputList.begin(); inputIter != inputList.end(); inputIter++){
      (*outputIter)->dependency->push_back(*inputIter);
    }
  }
}

//Sets all inputs to fan out to all outputs
void logicBlock::setInputFanOut(){
  list<Terminal*>::iterator inputIter;
  list<Terminal*>::iterator outputIter;

  for(inputIter = inputList.begin(); inputIter != inputList.end(); inputIter++){
    (*inputIter)->fanOut = new list <Terminal*>;
    for(outputIter = outputList.begin(); outputIter != outputList.end(); outputIter++){
      (*inputIter)->fanOut->push_back(*outputIter);
    }
  }
}

void logicBlock::set_registerBank(registerBank *regBank){
  rb = regBank;
  row = rb->getHostCLB()->row;
  col = rb->getHostCLB()->col;
}

void logicBlock::unset_registerBank(registerBank *regBank){
  //make sure this is the block that is actually mapped to this logic block first
  if(rb != regBank){
    ERROR("Block mis-match!");
  }

  rb = NULL; 
  row = -1;
  col = -1;
}


// //Sets output i to depend upon input j
// void logicBlock::setOutputDependency(int outputNum, int inputNum){
//   list<Terminal*>::iterator inputIter;
//   list<Terminal*>::iterator input;
//   int i;

//   if(outputNum < 0 || outputNum >= outputList.size() || inputNum < 0 || inputNum >= inputList.size()){
//     ERROR("Problem with setting output dependency!");
//   }
  
//   //Find the appropriate input
//   input = inputList.begin();
//   for(i = 0; i < inputNum && input != inputList.end(); i++){
//     input++;
//   }

// //   //See if output i is already dependent upon input j
// //   for(inputIter = outputDependency.at(outputNum).begin(); inputIter != outputDependency.at(outputNum).end(); inputIter++){
// //     if(*inputIter == *input){
// //       ERROR("Output already dependent upon input!");
// //     }
// //   }

// //   //Push the input into the appropriate output's dependency list
// //   outputDependency.at(outputNum).push_back(*input);
// }

// //Sets input i to fan out to output j
// void logicBlock::setInputFanOut(int inputNum, int outputNum){
//   list<Terminal*>::iterator outputIter;
//   list<Terminal*>::iterator output;
//   int i;

//   if(outputNum < 0 || outputNum >= outputList.size() || inputNum < 0 || inputNum >= inputList.size()){
//     ERROR("Problem with setting output dependency!");
//   }
  
//   //Find the appropriate output
//   output = outputList.begin();
//   for(i = 0; i < outputNum && output != outputList.end(); i++){
//     output++;
//   }

// //   //See if input i already fans out to output j
// //   for(outputIter = inputFanOut.at(inputNum).begin(); outputIter != inputFanOut.at(inputNum).end(); outputIter++){
// //     if(*outputIter == *output){
// //       ERROR("Input already fans out to input!");
// //     }
// //   }

// //   //Push the output into the appropriate input's dependency list
// //   inputFanOut.at(inputNum).push_back(*output);
// }

void logicBlock::updateSinkDelaysAndSlacks(){
  list<Terminal*>::iterator iter;
  
  // Update the sink timing for the nets attached to this block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net
    (*iter)->updateSinkDelayAndSlack();
  }
  
  // Recalculate the sink timing for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net
    (*iter)->updateSourceDelaysAndSlacks();
  }
}

void logicBlock::updateSinkDelaysAndSlacksForAllCLBLUTs(){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this logic block's parent CLB
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }
}

void logicBlock::updateSinkDelaysAndSlacksForEntireCLB(){
  vector<logicBlock *>::iterator iter;
  
  //Go through all of the LUTs associated with this logic block's parent CLB
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }

  for(iter = parentCLB->registerVector.begin(); iter != parentCLB->registerVector.end(); iter++){
    (*iter)->updateSinkDelaysAndSlacks();
  }  
}

//We just created lots of pipelining nets upstream of this block's inputs.
//Don't forget to update those guys.
void logicBlock::updateSinkDelaysAndSlacksBackwards(){
  list<Terminal*>::iterator iter;
  
  // Update the sink timing for the nets attached to this block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net and those nerfed ones upstream.
    (*iter)->updateSinkDelayAndSlackBackwards();
  }
  
  // Recalculate the sink timing for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net
    (*iter)->updateSourceDelaysAndSlacks();
  }
}

//We just created lots of pipelining nets downstream of this block's outputs
//Don't forget to update those guys.
void logicBlock::updateSinkDelaysAndSlacksForwards(){
  list<Terminal*>::iterator iter;
  // Update the sink timing for the nets attached to this block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the sink timing for this net.
    (*iter)->updateSinkDelayAndSlack();
  }
  
  // Recalculate the sink timing for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    //This terminal is the source of a net.
    //Update the sink timings for all the sinks of this net and the nerfed ones downstream
    (*iter)->updateSourceDelaysAndSlacksForwards();
  }
}

//We have updated the sink timings for the nets associate with this block, so go through these nets
// and recalculate their wire and timing costs
void logicBlock::updateWireAndTimingCosts(bool timing_driven, double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    (*iter)->updateSinkWireAndTimingCost(timing_driven, criticalPathDelay);
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    (*iter)->updateSourceWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
}

//ATTENTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//I think that this function might have to change to accomodate multiple LUTs inside a CLB
void logicBlock::updateWireAndTimingCostsForEntireCLB(bool timing_driven, double criticalPathDelay){
  vector<logicBlock *>::iterator iter;

  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end(); iter++){
    (*iter)->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  }

  if(parentCLB->registerVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->registerVector.begin(); iter != parentCLB->registerVector.end(); iter++){
    (*iter)->updateWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
}

void logicBlock::updateCostsAndDeltas(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);

  *deltaWire += determineDeltaWireCost();

  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();  
  }
}

void logicBlock::updateCostsAndDeltasCostDeltaDelay(bool timing_driven, double criticalPathDelay, double *deltaWire, double *deltaTiming, double *deltaTimingPrime){
  updateWireAndTimingCosts(timing_driven, criticalPathDelay);

  *deltaWire += determineDeltaWireCost();

  if(timing_driven){
    *deltaTiming += determineDeltaTimingCost();
    *deltaTimingPrime += determineDeltaTimingCostPrime();
  }
}

//We have updated the sink timings for the nets associate with this block, so go through these nets
// and recalculate their wire and timing costs.  However, we have also added a number of new
// pipelining nets to the inputs of this logic block.  Update all nets attached to nerfed upstream
// registers.
void logicBlock::updateWireAndTimingCostsBackwards(bool timing_driven, double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    (*iter)->updateSinkWireAndTimingCostBackwards(timing_driven, criticalPathDelay);
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    (*iter)->updateSourceWireAndTimingCosts(timing_driven, criticalPathDelay);
  }
}

//We have updated the sink timings for the nets associate with this block, so go through these nets
// and recalculate their wire and timing costs.  However, we have also added a number of new
// pipelining nets to the outputs of this logic block.  Update all nets attached to nerfed downstream
// registers.
void logicBlock::updateWireAndTimingCostsForwards(bool timing_driven, double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    (*iter)->updateSinkWireAndTimingCost(timing_driven, criticalPathDelay);
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    (*iter)->updateSourceWireAndTimingCostsForwards(timing_driven, criticalPathDelay);
  }
}

double logicBlock::determineDeltaWireCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in wirelength.
  list<Terminal*>::iterator iter;
  double deltaWire = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaWire += (*iter)->sinkDeltaWireCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaWire += (*iter)->sourceDeltaWireCost();
  }
  return deltaWire;
}

//ATTENTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//I think that this function might have to change to accomodate multiple LUTs inside a CLB
double logicBlock::determineDeltaWireCostForEntireCLB(){
  vector<logicBlock *>::iterator iter;
  double deltaWire = 0.0;
  
  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end(); iter++){
    deltaWire += (*iter)->determineDeltaWireCost();
  }
  
  if(parentCLB->registerVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->registerVector.begin(); iter != parentCLB->registerVector.end(); iter++){
    deltaWire += (*iter)->determineDeltaWireCost();
  }
  return deltaWire;
}

//There are brand new pipelining nets upstream of this logic block, so don't forget those
double logicBlock::determineDeltaWireCostBackwards(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in wirelength.
  list<Terminal*>::iterator iter;
  double deltaWire = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaWire += (*iter)->sinkDeltaWireCostBackwards();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaWire += (*iter)->sourceDeltaWireCost();
  }
  return deltaWire;
}

//There are brand new pipelining nets downstream of this logic block, so don't forget those
double logicBlock::determineDeltaWireCostForwards(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in wirelength.
  list<Terminal*>::iterator iter;
  double deltaWire = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaWire += (*iter)->sinkDeltaWireCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaWire += (*iter)->sourceDeltaWireCostForwards();
  }
  return deltaWire;
}

double logicBlock::determineDeltaTimingCost(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTiming = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaTiming += (*iter)->sinkDeltaTimingCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaTiming += (*iter)->sourceDeltaTimingCost();
  }
  return deltaTiming;
}

double logicBlock::determineDeltaTimingCostPrime(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTimingPrime = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaTimingPrime += (*iter)->sinkDeltaTimingCostPrime();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaTimingPrime += (*iter)->sourceDeltaTimingCostPrime();
  }
  return deltaTimingPrime;
}

//ATTENTION!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//I think that this function might have to change to accomodate multiple LUTs inside a CLB
double logicBlock::determineDeltaTimingCostForEntireCLB(){
  vector<logicBlock *>::iterator iter;
  double deltaTiming = 0.0;
  
  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end(); iter++){
    deltaTiming += (*iter)->determineDeltaTimingCost();
  }
  
  if(parentCLB->registerVector.size() > 1){
    ERROR("We can only handle 1 LUT/1 FF CLBs!");
  }
  for(iter = parentCLB->registerVector.begin(); iter != parentCLB->registerVector.end(); iter++){
    deltaTiming += (*iter)->determineDeltaTimingCost();
  }
  return deltaTiming;
}

//There are brand new pipelining nets upstream of this logic block, so don't forget those
double logicBlock::determineDeltaTimingCostBackwards(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTiming = 0.0;
  double tempDouble;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaTiming += (*iter)->sinkDeltaTimingCostBackwards();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    tempDouble = (*iter)->sourceDeltaTimingCost();
    //cout << "source timing cost for terminal " << (*iter)->name << " = " << tempDouble <<endl;
    deltaTiming += tempDouble;
  }
  return deltaTiming;
}

//There are brand new pipelining nets downstream of this logic block, so don't forget those
double logicBlock::determineDeltaTimingCostForwards(){
  //Go through the all of the nets this block is attached to and figure out the total increase
  // or decrease in timing cost.
  list<Terminal*>::iterator iter;
  double deltaTiming = 0.0;
  
  // Recalculate the placement cost for the net attached to the sinks of this logic block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    deltaTiming += (*iter)->sinkDeltaTimingCost();
  }
  
  // Recalculate the placement cost for the nets attached to the sources of this logic block
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    deltaTiming += (*iter)->sourceDeltaTimingCostForwards();
  }
  return deltaTiming;
}

//Go though the output nets of this block and figure out how many registers we push back through
void logicBlock::calculateRegPushBackward(){
  Terminal *term;
  vector<int>::iterator intIter;
  int numReg;

  //We should be able to push register backwards from everything but an I/O pad
  if(blockType != OUTPUTPAD && blockType != INPUTPAD){
    //If this block has no input, we should not be able to push any blocks backwards.
    if(inputList.size() != 0){
      //Go through the output nets of this block and figure out what the minimum
      // pipeline latency of any of the sinks is.  This is the maximum number of registers
      // that can be pushed backward through this logic block
      //I say output net*s* because some blocks have the multiple output lines.
      //We have to pull one register off of each output line to the inputs when we 
      // move a register backwards.
      //Also, what do we do about blocks that have no output in the datapath?  We will assume those
      // blocks are connected to the control and thus have zero registers on them
      numReg = BIGINTEGER;
      create_outputListItr();
      for(term = getnext_outputList(); term != 0; term = getnext_outputList()){
	for(intIter = term->net->latencyVector.begin(); intIter != term->net->latencyVector.end(); intIter++){
	  if(*intIter < numReg){
	    numReg = *intIter;
	  }
	}
      }
      remove_outputListItr();
      
      //This logic block might not have had any output so we need to set the numReg to zero
      if(numReg == BIGINTEGER){
	numReg = 0;
      }

      numRegCanPushBackward = numReg;
    }
    else{
      numRegCanPushBackward = 0;
    }
  }
  //cout << "Logic block " << name << " can push " << numRegCanPushBackward << " register backward" << endl;
}

//Go through the input nets of this block and figure out how many registers we can push forwards
void logicBlock::calculateRegPushForward(){
  Terminal *term;
  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator intIter;
  int i;
  int numReg;

  //We should be able to push register forwards from everything but an I/O pad
  if(blockType != OUTPUTPAD && blockType != INPUTPAD){
    //If this logic block has no output, we should not be able to push any blocks forwards.
    if(outputList.size() != 0){
      //Go through the input nets of this block and figure out what the minimum pipeline latency of any of
      // the inputs are.  If this block does not have any inputs we'll assume that the block is control
      // limited and we cannot pull any register forward through this block.
      numReg = BIGINTEGER;
      create_inputListItr();
      for(term = getnext_inputList(); term != 0; term = getnext_inputList()){
	//Find this sink in the host net
	for(sinkIter = term->net->sinkVector.begin(), i = 0; sinkIter != term->net->sinkVector.end(); sinkIter++, i++){
	  if(*sinkIter == term){
	    break;
	  }
	}
	if(sinkIter == term->net->sinkVector.end()){
	  ERROR("Did not find sink");
	}
	if(term->net->latencyVector.at(i) < numReg){
	  numReg = term->net->latencyVector.at(i);
	}
      }
      remove_inputListItr();
      
      //This logic block might not have had any input so we need to set the numReg to zero
      if(numReg == BIGINTEGER){
	numReg = 0;
      }    
      numRegCanPushForward = numReg;
    }
    else{
      numRegCanPushForward = 0;
    }
    //cout << "Logic block " << name << " can push " << numRegCanPushForward << " register forward" << endl;  
  }
}

//Figure out how critical the most critial input to this logic block is.
void logicBlock::calculateMaxInputCriticality(double criticalPathDelay){
  double terminalSlack;
  double minInputSlack = 10000.0;
  for(list<Terminal*>::iterator iter = inputList.begin(); iter != inputList.end(); iter++){
    terminalSlack = (*iter)->returnInputSlack();
    if(terminalSlack < minInputSlack){
      minInputSlack = terminalSlack;
    }
  }
  maxInputCriticality = 1.0 - (minInputSlack / criticalPathDelay);
}

//Figure out what the most critical sink of this logic block is
void logicBlock::calculateMaxOutputCriticality(double criticalPathDelay){
  double terminalSlack;
  double minOutputSlack = 10000.0;
  for(list<Terminal*>::iterator iter = outputList.begin(); iter != outputList.end(); iter++){
    terminalSlack = (*iter)->returnMinOutputSlack();
    if(terminalSlack < minOutputSlack){
      minOutputSlack = terminalSlack;
    }
  }
  maxOutputCriticality = 1.0 - (minOutputSlack / criticalPathDelay);
}

void logicBlock::calculateMaxInputOutputCriticality(double criticalPathDelay){
  calculateMaxInputCriticality(criticalPathDelay);
  calculateMaxOutputCriticality(criticalPathDelay);
}

void logicBlock::calculateMaxInputOutputCriticalitiesForwards(double criticalPathDelay){
  list<Terminal*>::iterator iter;

  //Calcalate the max input and output criticalities for this logic block
  calculateMaxInputOutputCriticality(criticalPathDelay);

  //Now, calculate the max input and output criticalities for the registers just downstream of this
  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    //This terminal is the source of a net.
    //Update the timings for all the sinks of this net and the nerfed ones downstream
    (*iter)->calculateMaxInputOutputCriticalityForwards(criticalPathDelay);
  }
}

void logicBlock::calculateMaxInputOutputCriticalitiesBackwards(double criticalPathDelay){
  list<Terminal*>::iterator iter;
  
 //Calcalate the max input and output criticalities for this logic block
  calculateMaxInputOutputCriticality(criticalPathDelay);
  
  // Update the sink timing for the nets attached to this block
  for(iter = inputList.begin(); iter != inputList.end(); iter++){
    //This terminal is the sink of a net.
    //Update the timing for this net and those nerfed ones upstream.
    (*iter)->calculateMaxInputOutputCriticalityBackwards(criticalPathDelay);
  }
}

//Push N registers from each of the outputs of this block to each of the inputs.
//We already know that there is sufficient room in the architecture to perform this retiming move
bool logicBlock::retimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;

  //Go through each of the outputs of this block, free numMove registers from each sink
  // and connect them directly to the appropriate output pin.
  //However, we may want to undo this retiming move, so although the register banks will need to
  // be updated, don't actually destroy the terminals or pipelining nets that represent the original
  // arrangement.  We want to be able to put things back in the correct place.
  if(!retimeUndoList->empty()){
    ERROR("Undo list has stuff in it!");
  }
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    retimeUndoList->push_back(tempTerm->detachAndReattachHereForward(netlist, numMove));
  }
  remove_outputListItr();

  //Now, add new registers to all of the inputs
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    //This addition could fail at any time.  If so, we have to be able to undo the partially
    // completed retiming move.
    if(!tempTerm->addNRegistersBackward(netlist, numMove, criticalPathDelay)){
      remove_inputListItr();
      return false;
    }
  }
  remove_inputListItr();  

  //We succeeded in removing and allocating all registers - update the retiming
  // capabilities of all related blocks.
  //First reduce this logic block's capability of pushing registers backwards and increase it's capability of
  // pushing registers forwards
  numRegCanPushBackward -= numMove;
  numRegCanPushForward += numMove;

  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    //We just took a bunch of registers off of this output.  That may change the
    // retiming capabilities of the downstream logic block.  We should re-evaluated its
    // retiming capacity.
    tempTerm->checkForwardsCapacity();
  }
  remove_outputListItr();

  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    //Update the capacity of upstream logic blocks to push registers backwards.
    //Since we increased the latency of this input, we might have changed the overall
    // capacity of the immediate upstream logic block.
    tempTerm->checkBackwardsCapacity();
  }
  remove_inputListItr(); 
  return true;
}

bool logicBlock::retimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;

  //Go through each of the inputs of this block, free numMove registers from each source
  // and connect them directly to the appropriate input pin.
  //However, we may want to undo this retiming move, so although the register banks will need to
  // be updated, don't actually destroy the terminals or pipelining nets that represent the original
  // arrangement.  We want to be able to put things back in the correct place.
  if(!retimeUndoList->empty()){
    ERROR("Undo list has stuff in it!");
  }
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    retimeUndoList->push_back(tempTerm->detachAndReattachHereBackward(netlist, numMove));
  }
  remove_inputListItr();

  //Now, add new registers to all of the outputs
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    //This addition could fail at any time.  If so, we have to be able to undo the partially
    // completed retiming move.
    if(!tempTerm->addNRegistersForward(netlist, numMove, criticalPathDelay)){
      remove_outputListItr(); 
      return false;
    }
  }
  remove_outputListItr();  

  //Reduce this logic block's capability of pushing registers forwards and increase it's capability of
  // pushing registers backwards
  numRegCanPushForward -= numMove;
  numRegCanPushBackward += numMove;

  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    //We just reduced the number of available registers on this input net.  That may change
    // the retiming capabilities of the upstream logic block.  We should re-evaluated its
    // retiming capacity.
    tempTerm->checkBackwardsCapacity();
  }
  remove_inputListItr();

  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    //Update the capacity of downstream logic blocks to push registers forwards.
    //Since we increased the latency of this output, we might have changed the overall
    // capacity of the logic block.
    tempTerm->checkForwardsCapacity();
  }
  remove_outputListItr();  
  return true;
}

//We had pushed N registers from each of the outputs of this block to each of the inputs.
//We would like to undo this move.  The undo list provides the list of pipelining nets that
// were formerly driven by each of the output terminals of this logic block.
void logicBlock::undoRetimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;
  PipeliningNet *tempPNet;
  list <PipeliningNet *> newRetimeUndoList;
  
  //First, go through and remove numMove registers from each of the inputs
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    newRetimeUndoList.push_back(tempTerm->detachAndReattachHereBackward(netlist, numMove));
  }
  remove_inputListItr();

  //Now, go through and re-attach the nets provided by the re-time undoList.
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->reattachNetForwards(retimeUndoList->front(), netlist, numMove);
    retimeUndoList->pop_front();
  }
  remove_outputListItr();

  //Reduce this logic block's capability of pushing registers forwards and increase it's capability of
  // pushing registers backwards
  numRegCanPushForward -= numMove;
  numRegCanPushBackward += numMove;

  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->checkBackwardsCapacity();
  }
  remove_inputListItr();

  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->checkForwardsCapacity(); 
  }
  remove_outputListItr();

  //Make sure that the old retimeUndoList is empty, then fill it with the new undo list
  if(!retimeUndoList->empty()){
    ERROR("Incomplete undo!");
  }
  else{
    while(!newRetimeUndoList.empty()){
      retimeUndoList->push_back(newRetimeUndoList.front());
      newRetimeUndoList.pop_front();
    }
  }
}

//We had pushed N registers from each of the inputs of this block to each of the outputs.
//We would like to undo this move now.  The undo list provides a list of pipelining nets
//that formerly drove each of the input terminals of this logic block.
void logicBlock::undoRetimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;
  PipeliningNet *tempPNet;
  list <PipeliningNet *> newRetimeUndoList;

  //First, go through and remove numMove registers from each of the outputs
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    newRetimeUndoList.push_back(tempTerm->detachAndReattachHereForward(netlist, numMove));
  }
  remove_outputListItr();

  //Now, go through and re-attach the nets provided by the re-time undoList.
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->reattachNetBackwards(retimeUndoList->front(), netlist, numMove);
    retimeUndoList->pop_front();
  }
  remove_inputListItr();

  //Reduce this logic block's capability of pushing registers backwards and increase it's capability of
  // pushing registers forwards
  numRegCanPushForward += numMove;
  numRegCanPushBackward -= numMove;

  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->checkForwardsCapacity();   
  }
  remove_outputListItr();
  
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->checkBackwardsCapacity(); 
  }
  remove_inputListItr();
 
  //Make sure that the old retimeUndoList is empty, then fill it with the new undo list
  if(!retimeUndoList->empty()){
    ERROR("Incomplete undo!");
  }
  else{
    while(!newRetimeUndoList.empty()){
      retimeUndoList->push_back(newRetimeUndoList.front());
      newRetimeUndoList.pop_front();
    }
  }
}

//We tried to retime registers backwards through this logic block, but we ran out of empty register banks.
//This function is almost the same as doing an undo of a completed retiming 
// backwards step, but we have to keep track of things a bit differently.
//For example, we didn't update any registers' push forward/backward capabilities.
//We are also going to figure out the delta costs differently since we never actually
// called "determineDeltaCosts" the first time.
void logicBlock::abortRetimeRegistersBackwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;
  PipeliningNet *tempPNet;
  
  //First, go through and remove nerfed registers from each of the inputs
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->detachAndReattachHereBackwardsAborted(netlist);
  }
  remove_inputListItr();

  //Now, go through and re-attach the nets provided by the re-time undoList.
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->reattachNetForwards(retimeUndoList->front(), netlist, numMove);
    retimeUndoList->pop_front();
  }
  remove_outputListItr();
}

//We tried to retime registers forwards through this logic block, but we ran out of empty register banks
//This function is almost the same as doing an undo of a completed retiming 
// forwards step, but we have to keep track of things a bit differently
void logicBlock::abortRetimeRegistersForwards(Netlist *netlist, int numMove, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList){
  Terminal* tempTerm;
  PipeliningNet *tempPNet;
  list <PipeliningNet *> newRetimeUndoList;

  //First, go through and remove nerfed registers from each of the outputs
  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->detachAndReattachHereForwardsAborted(netlist);
  }
  remove_outputListItr();

  //Now, go through and re-attach the nets provided by the re-time undoList.
  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->reattachNetBackwards(retimeUndoList->front(), netlist, numMove);
    retimeUndoList->pop_front();
  }
  remove_inputListItr();
}

//There are nerfed registers upstream of this logic block.
//Un-nerf them.
void logicBlock::clearNerfedBlocksBackward(){
  Terminal* tempTerm;

  create_inputListItr();
  for(tempTerm = getnext_inputList(); tempTerm != 0; tempTerm = getnext_inputList()){
    tempTerm->clearNerfedBlocksBackward();
  }
  remove_inputListItr();
}

//There are nerfed registers downstream of this logic block.
//Un-nerf them.
void logicBlock::clearNerfedBlocksForward(){
  Terminal* tempTerm;

  create_outputListItr();
  for(tempTerm = getnext_outputList(); tempTerm != 0; tempTerm = getnext_outputList()){
    tempTerm->clearNerfedBlocksForward();
  }
  remove_outputListItr();
}

//Check to make sure that the number of registers we can push forwards and backwards
// through this logic block are correct.
void logicBlock::checkPushForwardsAndBackwards(){
  Terminal *term;
  vector<Terminal*>::iterator sinkIter;
  vector<int>::iterator intIter;
  int numReg;
  int i;

  //We should be able to push a block past everything but an I/O pad
  if(blockType != OUTPUTPAD && blockType != INPUTPAD){
   //If this logic block has no input, we should not be able to push any blocks backwards.
    if(inputList.size() != 0){
      //Go through the output nets of this block and figure out what the minimum
      // pipeline latency of any of the sinks is.  This is the maximum number of registers
      // that can be pushed backward through this logic block
      //I say output net*s* because some blocks have the multiple output lines.
      //We have to pull one register off of each output line to the inputs when we 
      // move a register backwards.
      //Also, what do we do about blocks that have no output in the datapath?  We will assume those
      // blocks are connected to the control and thus have zero registers on them
      numReg = BIGINTEGER;
      create_outputListItr();
      for(term = getnext_outputList(); term != 0; term = getnext_outputList()){
	for(intIter = term->net->latencyVector.begin(); intIter != term->net->latencyVector.end(); intIter++){
	  if(*intIter < numReg){
	    numReg = *intIter;
	  }
	}
      }
      remove_outputListItr();
      
      //This logic block might not have had any output so we need to set the numReg to zero
      if(numReg == BIGINTEGER){
	numReg = 0;
      }
    }
    else{
      numReg = 0;
    }
    if(numRegCanPushBackward != numReg){
      ERROR("Number of registers we can push backwards is not the same!");
    }

    //If this logic block has no output, we should not be able to push any blocks forwards.
    if(outputList.size() != 0){
      //Go through the input nets of this block and figure out what the minimum pipeline latency of any of
      // the inputs are.  If this block does not have any inputs we'll assume that the block is control
      // limited and we cannot pull any register forward through this block.
      numReg = BIGINTEGER;
      create_inputListItr();
      for(term = getnext_inputList(); term != 0; term = getnext_inputList()){
	//Find this sink in the host net
	for(sinkIter = term->net->sinkVector.begin(), i = 0; sinkIter != term->net->sinkVector.end(); sinkIter++, i++){
	  if(*sinkIter == term){
	    break;
	  }
	}
	if(sinkIter == term->net->sinkVector.end()){
	  ERROR("Did not find sink");
	}
	if(term->net->latencyVector.at(i) < numReg){
	  numReg = term->net->latencyVector.at(i);
	}
      }
      remove_inputListItr();
      
      //This logic block might not have had any input so we need to set the numReg to zero
      if(numReg == BIGINTEGER){
	numReg = 0;
      }    
    }
    else{
      numReg = 0;
    }
    if(numRegCanPushForward != numReg){
      ERROR("Number of registers we can push forwards is not the same!");
    }
  }
}

void logicBlock::placeEntireCLB(registerBank *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->getHostCLB()->row;
  int targetCol = block->getHostCLB()->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

  //Update the parent CLB of the new position
  updateCLBLocation(block);  

  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > targetCLB->getNumEmptyLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }

  //Figure out what the first empty lut is!
  i = targetCLB->getNumLUTs() - targetCLB->getNumEmptyLUTs();
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end() && i < targetCLB->getNumLUTs(); iter++, i++){
    //for(iter = parentCLB->lutVector.begin(), i = 0; iter != parentCLB->lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
   else{
      ERROR("LUT has current placement!");
    }
  }
  if(parentCLB->registerVector.size() > targetCLB->getNumEmptyRegBank(1)){
    ERROR("We cannot handle that many FFs in this CLB!");
  }

  for(iter = parentCLB->registerVector.begin(); 
      iter != parentCLB->registerVector.end(); iter++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      i = targetCLB->firstEmptyRegisterIndex();
      targetRB = targetCLB->getRegBank(i);
      targetRB->placeLogicBlock(*iter);
    }
    else{
      ERROR("FF has current placement!");
    }
  }
}

void logicBlock::placeEntireCLB(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  registerBank* targetRB;
  int i;

  //Update the parent CLB of the new position
  updateCLBLocation(block);

  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > targetCLB->getNumEmptyLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }

  //Figure out what the first empty lut is!
  i = targetCLB->getNumLUTs() - targetCLB->getNumEmptyLUTs();
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end() && i < targetCLB->getNumLUTs(); iter++, i++){
  //for(iter = parentCLB->lutVector.begin(), i = 0; iter != parentCLB->lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
   else{
      ERROR("LUT has current placement!");
    }
  }
  if(parentCLB->registerVector.size() > targetCLB->getNumRegBanks()){
    ERROR("We cannot handle that many FFs in this CLB!");
  }

  for(iter = parentCLB->registerVector.begin(); 
      iter != parentCLB->registerVector.end(); iter++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      i = targetCLB->firstEmptyRegisterIndex();
      targetRB = targetCLB->getRegBank(i);
      targetRB->placeLogicBlock(*iter);
    }
    else{
      ERROR("FF has current placement!");
    }
  }
//   for(iter = parentCLB->registerVector.begin(), i = 0; iter != parentCLB->registerVector.end(); iter++, i++){
//     if((*iter)->row == -1 && (*iter)->col == -1){
//       targetRB = targetCLB->getRegBank(i);
//       targetRB->placeLogicBlock(*iter);
//     }
//     else{
//       ERROR("FF has current placement!");
//     }
//   }
}

//Place all of the LUTs in the host logical CLB into the
// LUTs in the physical CLB
void logicBlock::placeAllCLBLUTs(vprLUT *block){
  vector<logicBlock *>::iterator iter;
  int targetRow = block->row;
  int targetCol = block->col;
  vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);
  int i;

  //Update the parent CLB of the new position
  updateCLBLocation(block);  

  //Go through all of the LUTs associated with this logic block's parent CLB
  if(parentCLB->lutVector.size() > targetCLB->getNumEmptyLUTs()){
    ERROR("We cannot handle that many LUTs in this CLB!");
  }

  //Figure out what the first empty lut is!
  i = targetCLB->getNumLUTs() - targetCLB->getNumEmptyLUTs();
  for(iter = parentCLB->lutVector.begin(); iter != parentCLB->lutVector.end() && i < targetCLB->getNumLUTs(); iter++, i++){
  //for(iter = parentCLB->lutVector.begin(), i = 0; iter != parentCLB->lutVector.end(); iter++, i++){
    if((*iter)->row == -1 && (*iter)->col == -1){
      placeBlock((*iter), (void *) targetCLB, LB, i);
    }
   else{
      ERROR("LUT has current placement!");
    }
  }


//   //Go through all of the LUTs associated with this logic block's parent CLB
//   if(parentCLB->lutVector.size() > targetCLB->getNumLUTs()){
//     ERROR("We cannot handle that many LUTs in this CLB!");
//   }
//   for(iter = parentCLB->lutVector.begin(), i = 0; iter != parentCLB->lutVector.end(); iter++, i++){
//     if((*iter)->row == -1 && (*iter)->col == -1){
//       placeBlock((*iter), (void *) targetCLB, LB, i);
//     }
//   }
}

//Remove all of the LUTs in the host logical CLB from the
// LUTs in the physical CLB
// void logicBlock::unPlaceAllCLBLUTs(vprLUT *block){
//   vector<logicBlock *>::iterator iter;
//   int targetRow = block->row;
//   int targetCol = block->col;
//   int i;
//   vprCLB* targetCLB = FPGA->getCLBAt(targetRow, targetCol);

//   //Update the parent CLB of the new position
//   clearCLBLocation(block);  

//   //Go through all of the LUTs associated with this logic block's parent CLB
//   if(parentCLB->lutVector.size() > targetCLB->getNumLUTs()){
//     ERROR("We cannot handle that many LUTs in this CLB!");
//   }
//   for(iter = parentCLB->lutVector.begin(), i = 0; iter != parentCLB->lutVector.end(); iter++, i++){
//     if((*iter)->row != targetRow || (*iter)->col != targetCol){
//       ERROR("LUT does not have correct current placement!");
//     }
//     else{
//       unPlaceBlock((*iter), (void *) targetCLB, LB, (*iter)->ptc_index);
//     }
//   }
// }

//If this block is a register, try to return the source logic block.
//However, if this is not driven by a logic block 
// (instead driven by a pipelining register), return NULL
logicBlock* logicBlock::getRegisterSource(){
  if(get_blockType() != REGISTER_BANK){
    ERROR("Should not be called on non-register blocks");
  }
  if(inputList.size() > 1){
    ERROR("Register has more than 1 input!");
  }

  if(inputList.front()->sourcePipeliningNet){
    //This logic block is driven by a pipelining net. 
    if(!inputList.front()->sourcePipeliningNet->source->lBlock){
      return NULL;
    }
    else{
      return(inputList.front()->sourcePipeliningNet->source->lBlock);
    }
  }
  return(inputList.front()->net->source->lBlock);
}



//Return the type of block that drives this register bank.
//Could be a LUT, another register bank, an I/O pin, RAM, MULTIPLIER or a pipelining register.
enum e_genericBlockType logicBlock::returnDriverType(){
  if(get_blockType() != REGISTER_BANK){
    ERROR("Should not be called on non-register blocks");
  }
  if(inputList.size() > 1){
    ERROR("Register has more than 1 input!");
  }
  
  if(inputList.front()->sourcePipeliningNet){
    //This logic block is driven by a pipelining net. 
    if(!inputList.front()->sourcePipeliningNet->source->lBlock){
      //If this pipelining net is not driven by a logic block,
      // we know that this logic block is driven by a pipelining register
      return GBT_NO_TYPE;
    }
    else{
      return (inputList.front()->sourcePipeliningNet->source->lBlock->get_blockType());
    }
  }
  else{
    //This is driven by a regular net.
    //Thus it must be driven by another logic block
    return (inputList.front()->net->source->lBlock->get_blockType());
  }
}

//Put this register bank in the same CLB as it's driver, if there is room.
//Otherwise, just put it in an empty register bank.
void logicBlock::placeWithParent(){
  Terminal *sourcePR;
  logicBlock *sourceLB;
  vprCLB *sourceCLB;
  registerBank *emptyRB;
  bool done;
  
  if(get_blockType() != REGISTER_BANK){
    ERROR("Should not be called on non-register blocks");
  }
  if(inputList.size() > 1){
    ERROR("Register has more than 1 input!");
  }

  //Figure out where there source is at.
  if(inputList.front()->sourcePipeliningNet){
    //This logic block is driven by a pipelining net. 
    if(!inputList.front()->sourcePipeliningNet->source->lBlock){
      //If this pipelining net is not driven by a logic block,
      // we know that this logic block is driven by a pipelining register.
      sourcePR = inputList.front()->sourcePipeliningNet->source;
	
      //Make sure this pipelining register is actually placed
      if(sourcePR->getLogicBlockRow() == -1 || sourcePR->getLogicBlockCol() == -1){
	ERROR("Source pipelining register is not placed!");
      }
	
      //Notice, we don't try to put this register bank in the same register bank as
      // the pipelining register, even if there is room.
      //This is because this is a monolithic shift register.
      sourceCLB = FPGA->getCLBAt(sourcePR->getLogicBlockRow(), sourcePR->getLogicBlockCol());
    }
    else{
      //This logic block is driven by another logic block
      sourceLB = inputList.front()->sourcePipeliningNet->source->lBlock;
	
      if(sourceLB->get_blockType() == INPUTPAD){
	if(tvpack_initial_placement){
	  //We want to place this register some free location.
	  //If we were being totally faithful to what T-VPack does, we should
	  // go through all of the other logic blocks it is related to and pack this
	  // register with the "closest" other block that has room in it's CLB.
	  //~~~~~~~~~~~~~~~~~~~~~~~~~THIS IS A BIG ASSUMPTION! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	  //However, since right now we are targeting unclustered architectures, "closeness" doesn't matter and
	  // we can just put it into any completely free CLB.
	  //On clustered architectures we could enter an infinite loop with this implementation...
	  done = false;
	  while(!done){
	    emptyRB = FPGA->getRandomFreeSingleRegBank();
	    if(emptyRB->emptyCLB()){
	      emptyRB->placeLogicBlock(this);
	      done = true;
	    }
	  }
	  return;
	}
	else{
	  emptyRB = FPGA->getRandomFreeSingleRegBank();
	  emptyRB->placeLogicBlock(this);
	  return;
	}
      }
	
      //Make sure this logic block is actually placed
      if(sourceLB->row == -1 || sourceLB->col == -1){
	ERROR("Logic block is not placed!");
      }
	
      sourceCLB = FPGA->getCLBAt(sourceLB->row, sourceLB->col);
    }
  }
  else{
    //This is driven by a regular net.
    //Thus it must be driven by another logic block
    sourceLB = inputList.front()->net->source->lBlock;
      
    if(sourceLB->get_blockType() == INPUTPAD){
      if(tvpack_initial_placement){
	//We want to place this register some free location.
	//If we were being totally faithful to what T-VPack does, we should
	// go through all of the other logic blocks it is related to and pack this
	// register with the "closest" other block that has room in it's CLB.
	//~~~~~~~~~~~~~~~~~~~~~~~~~THIS IS A BIG ASSUMPTION! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//However, since right now we are targeting unclustered architectures, "closeness" doesn't matter and
	// we can just put it into any completely free CLB.
	//On clustered architectures we could enter an infinite loop with this implementation...
	done = false;
	while(!done){
	  emptyRB = FPGA->getRandomFreeSingleRegBank();
	  if(emptyRB->emptyCLB()){
	    emptyRB->placeLogicBlock(this);
	    done = true;
	  }
	}
	return;
      }
      else{
	emptyRB = FPGA->getRandomFreeSingleRegBank();
	emptyRB->placeLogicBlock(this);
	return;
      }
    }
      
    //Make sure this logic block is actually placed
    if(sourceLB->row == -1 || sourceLB->col == -1){
      ERROR("Logic block is not placed!");
    }
      
    sourceCLB = FPGA->getCLBAt(sourceLB->row, sourceLB->col);
  }
 
  //Find out if there is a free register bank in the same CLB as the source
  emptyRB = sourceCLB->getFirstEmptySingleOrShiftRegisterBank();
  
  if(emptyRB){
    emptyRB->placeLogicBlock(this);
  }
  else{
    if(tvpack_initial_placement){
      //We want to place this register in the same CLB as it's parent, but there isn't room.
      //What should we do?
      //Well, if we were being totally faithful to what T-VPack does, we should
      // go through all of the other logic blocks it is related to and pack this
      // register with the "closest" other block that has room in it's CLB.
      //~~~~~~~~~~~~~~~~~~~~~~~~~THIS IS A BIG ASSUMPTION! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      //However, since right now we are targeting unclustered architectures, "closeness" doesn't matter and
      // we can just put it into any completely free CLB.
      //On clustered architectures we could enter an infinite loop with this implementation...
      done = false;
      while(!done){
	emptyRB = FPGA->getRandomFreeSingleRegBank();
	if(emptyRB->emptyCLB()){
	  emptyRB->placeLogicBlock(this);
	  done = true;
	}
      }
    }
    else{ 
      emptyRB = FPGA->getRandomFreeSingleRegBank();
      emptyRB->placeLogicBlock(this);
    }
  }
}

//See if the driver of this logic block has been placed
bool logicBlock::parentPlaced(){
  Terminal *sourcePR;
  logicBlock *sourceLB;

  if(get_blockType() != REGISTER_BANK){
    ERROR("Should not be called on non-register blocks");
  }
  if(inputList.size() > 1){
    ERROR("Register has more than 1 input!");
  }
  if(inputList.front()->sourcePipeliningNet){
    //This logic block is driven by a pipelining net. 
    if(!inputList.front()->sourcePipeliningNet->source->lBlock){
      //If this pipelining net is not driven by a logic block,
      // we know that this logic block is driven by a pipelining register.
      sourcePR = inputList.front()->sourcePipeliningNet->source;

      //See if this pipelining register is actually placed
      if(sourcePR->getLogicBlockRow() == -1 || sourcePR->getLogicBlockCol() == -1){
	return false;
      }
    }
    else{
      //This logic block is driven by another logic block
      sourceLB = inputList.front()->sourcePipeliningNet->source->lBlock;

      //Make sure this logic block is actually placed
      if(sourceLB->row == -1 || sourceLB->col == -1){
	return false;
      }
    }
  }
  else{
    //This is driven by a regular net.
    //Thus it must be driven by another logic block
    sourceLB = inputList.front()->net->source->lBlock;
    
    //Make sure this logic block is actually placed
    if(sourceLB->row == -1 || sourceLB->col == -1){
      return false;
    }
  }
  return true;
}

//Is this register bank it's own driver?
bool logicBlock::isSelfFeeding(){
  logicBlock *sourceLB;

  if(get_blockType() != REGISTER_BANK){
    ERROR("Should not be called on non-register blocks");
  }

  if(inputList.size() > 1){
    ERROR("Register has more than 1 input!");
  }

  if(inputList.front()->sourcePipeliningNet){
    //This logic block is driven by a pipelining net. 
    if(!inputList.front()->sourcePipeliningNet->source->lBlock){
      //If this pipelining net is not driven by a logic block,
      // we know that this logic block is driven by a pipelining register.
      //Thus, it can't be self-feeding.
      return false;
    }
    else{
      //This logic block is driven by another logic block
      sourceLB = inputList.front()->sourcePipeliningNet->source->lBlock;

      //See if this is the same logic block
      if(sourceLB == this){
	return true;
      }
    }
  }
  else{
    //This is driven by a regular net.
    //Thus it must be driven by another logic block
    sourceLB = inputList.front()->net->source->lBlock;
    
    if(sourceLB == this){
      return true;
    }
  }
  return false;
}

//Go through all sinks of this block and print out their Manhattan distance.
void logicBlock::printDistanceToAllSinks(){
  list<Terminal*>::iterator iter;

  for(iter = outputList.begin(); iter != outputList.end(); iter++){
    (*iter)->printDistanceToAllSinks();
  }  
}

//Return the Manhattan distance to the closest sink
int logicBlock::minDistanceToAllSinks(){
  if(outputList.size() > 1){
    ERROR("This shouldn't be called!");
  }

  return (outputList.front()->minDistanceToAllSinks());
}

//Return true if this logic block drives something in the CLB at row, col
bool logicBlock::drivesSomethingIn(int row, int col){
  for(list<Terminal*>::iterator iter = outputList.begin(); iter != outputList.end(); iter++){
    if((*iter)->drivesSomethingIn(row, col)){
      return true;
    }
  }
  return false;
}
