// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of a block as found in an mcnc netlist. For our
purposes, each block could be a I/O pin, a CLB, or it's component
parts (LUT or FF);
****************************************************************/ 
#ifndef MCNC_BLOCK_H
#define MCNC_BLOCK_H

class mcncNet;

//This block could be a I/O pin, a CLB, a LUT or a register
class mcncBlock{
 public:
  //char name [MAX_STR_LEN];  //name of block
  string name;  //name of block
  list<mcncNet*> inputList;  //list of block inputs
  list<mcncNet*> outputList;  //list of block outputs
  list<mcncNet*>::iterator* input_itr;
  list<mcncNet*>::iterator* output_itr;
  vector <int> inputLatencyVector;

  //Is this block a IO pin, a CLB, a LUT or a FF?
  enum e_mcncBlockType blockType;

  int numInputNets;  //num nets that fanin to this clb
  int numOutputNets;  //num nets that fanin to this clb
  int magicNum;

  //If this block is a CLB, we need to keep track of the LUTs and FF inside
  vector <mcncBlock*> luts;
  vector <mcncBlock*> registers;
  
  mcncBlock(){
    numInputNets=numOutputNets=0;
    input_itr=0;output_itr=0;
    magicNum=12118;
  }

  ~mcncBlock(){
    if(magicNum!=12118)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;

    //Free all of the LUTs and registers
    for(vector <mcncBlock*>::iterator lutIter = luts.begin(); lutIter != luts.end(); lutIter++){
      delete(*lutIter);
    }
    for(vector <mcncBlock*>::iterator regIter = registers.begin(); regIter != registers.end(); regIter++){
      delete(*regIter);
    }
  }

  void set_name(char* n);  //set name of block
  void addto_inputList(mcncNet* p_net);  //add a net to inputList
  void addto_outputList(mcncNet* p_net);  //add a net to outputList
  void set_blockType(enum e_mcncBlockType t);

  void create_inputListItr();  //create a new iterator for inputList
  void remove_inputListItr();  //remove an already existing inputList iterator 
  void create_outputListItr();  //create a new iterator for outputList
  void remove_outputListItr();  //remove an already existing outputList iterator 

  string get_name();  //get name of block
  mcncNet* getnext_inputList();  //get next element from inputList
  mcncNet* getnext_outputList();  //get next element from outputList
  enum e_mcncBlockType get_blockType();

  int get_numInputNets();  //get num block input nets
  int get_numOutputNets();  //get num block output nets
  void print_Info();  //print info about this block

  void pushBackInputLatencyVector(int latency);

  void addLUT(mcncBlock *lut){
    luts.push_back(lut);
  }

  void addRegister(mcncBlock *reg){
    registers.push_back(reg);
  }
};

#endif
