// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

%{
#include "include.h"
#include "global.h"
  %}

//%error-verbose

%union{
  char* word;
  double double_num;
}

%token <double_num> DOUBLENUM_TOK

// Stuff for arch file
%token <word> PERIOD_TOK
%token <word> TO_TOK
%token <word> NODE_TOK
%token <word> BLOCKNAME_TOK
%token <word> BLOCKTYPE_TOK
%token <word> WIRENAME_TOK
%token <word> CHANY_TOK
%token <word> CHANX_TOK
%token <word> SOURCE_TOK
%token <word> SINK_TOK
%token <word> IPIN_TOK
%token <word> OPIN_TOK
//%token <word> HELPER_TOK
%token <word> REGISTER_TOK
%token <word> REGISTERCHAIN_TOK
%token <word> REGINDEX_TOK
%token <word> LUTASSOCIATION_TOK
%token <word> PTC_TOK
%token <word> EDGES_TOK
%token <word> SWITCHTYPE_TOK
%token <word> OCC_TOK
%token <word> CAPACITY_TOK
%token <word> R_TOK
%token <word> C_TOK
%token <word> COSTINDEX_TOK
%token <word> INTRINSICDELAY_TOK

//Stuff for netlist file
%token <word> NETLISTFILE_TOK
%token <word> ARCHFILE_TOK
%token <word> PLACEMENTFILE_TOK
%token <word> INPUT_TOK 
%token <word> OUTPUT_TOK 
%token <word> CLB_TOK 
%token <word> TYPE_TOK 
%token <word> PINLIST_TOK 
%token <word> SUBBLOCK_TOK 
%token <word> SUBBLOCKREG_TOK
%token <word> DELAY_TOK
%token <word> GLOBAL_TOK
%token <word> ANNOTATION_TOK
%token <word> NET_TOK

%%
statement_list: statement
|    statement_list statement
;

statement: circuit_info{
  processCircuitInfo(0);
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
|
rr_graph_info{
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
|
placement_info{
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
;

rr_graph_info:
NODE_TOK NET_TOK PERIOD_TOK { 
  //  cout << "Processed node " << $2 << endl;
  G_currNode=find_rnode($2);
  free($1);free($2);free($3);
}
| BLOCKNAME_TOK NET_TOK { 
  process_blockName($2);
  free($1);free($2);
}
| BLOCKTYPE_TOK NET_TOK { 
  process_blockType($2);
  free($1);free($2);
}
//| WIRENAME_TOK NET_TOK { 
//  process_wireName($2);
//  free($1);free($2);
//}
| SOURCE_TOK NET_TOK NET_TOK PTC_TOK NET_TOK { 
  process_sourceNode(atoi($2), atoi($3), atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| SINK_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_sinkNode(atoi($2), atoi($3), atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| IPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_ipinNode(atoi($2), atoi($3), atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| OPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_opinNode(atoi($2), atoi($3), atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6));
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
| CHANY_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANY_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6));
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
//| HELPER_TOK NET_TOK NET_TOK PTC_TOK NET_TOK{
//  process_helperNode(atoi($2), atoi($3));
//  free($1);free($2);free($3);free($4);free($5);
//}
| REGISTER_TOK NET_TOK NET_TOK PTC_TOK NET_TOK{
  process_Register(atoi($2), atoi($3), atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| REGISTERCHAIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK REGINDEX_TOK NET_TOK LUTASSOCIATION_TOK NET_TOK{
  process_RegisterChain(atoi($2), atoi($3), atoi($5), atoi($7), atoi($9));
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);free($9);
}
| NET_TOK EDGES_TOK nets {
  process_Edges(atoi($1));
  free($1);free($2);
}
| SWITCHTYPE_TOK nets {
  process_switchTypes();
  free($1);
}
| OCC_TOK NET_TOK CAPACITY_TOK NET_TOK PERIOD_TOK {
  G_currNode->set_capacity(atoi($4));
  free($1);free($2);free($3);free($4);free($5);
}  
| R_TOK DOUBLENUM_TOK C_TOK DOUBLENUM_TOK { 
  G_currNode->R=$2; G_currNode->C=$4;
  free($1); free($3);;
}
|  R_TOK NET_TOK C_TOK NET_TOK { 
  G_currNode->R=atof($2); G_currNode->C=atof($4);
  free($1);free($2);free($3);free($4); 
}
|  R_TOK DOUBLENUM_TOK C_TOK NET_TOK { 
  G_currNode->R=$2; G_currNode->C=atof($4);
  free($1);free($3);free($4); 
}
|  R_TOK NET_TOK C_TOK DOUBLENUM_TOK { 
  G_currNode->R=atof($2); G_currNode->C=$4;
  free($1);free($2);free($3); 
}
|  COSTINDEX_TOK NET_TOK { 
  if(atof($2) < 1.0){
    G_currNode->baseCost = 0.1;
  }
  else{
    G_currNode->baseCost = atof($2);
  }
  free($1);free($2); 
}
|  COSTINDEX_TOK DOUBLENUM_TOK {
  if($2 < 1.0){
    G_currNode->baseCost = 0.1;
  }
  else{
    G_currNode->baseCost = $2;
  }
  free($1);
}
| INTRINSICDELAY_TOK DOUBLENUM_TOK {
  //G_currNode->baseCost = $2;
  G_currNode->intrinsicDelay = $2;
  //cout << "IntrinsicDelay = " << G_currNode->intrinsicDelay << endl;
  free($1);
}
;

circuit_info: keyword nets   { ; } 
;

keyword: INPUT_TOK  { /* cout<<endl<<$1; */ curr_tok = INPUT_TOK; free($1); }
|    OUTPUT_TOK     { /* cout<<endl<<$1; */ curr_tok = OUTPUT_TOK;free($1);  }
|    CLB_TOK        { /* cout<<endl<<$1; */ curr_tok = CLB_TOK; free($1);  }
|    TYPE_TOK       { /* cout<<endl<<$1; */ curr_tok = TYPE_TOK; free($1);  }
|    SUBBLOCK_TOK   { /* cout<<endl<<$1; */ curr_tok = SUBBLOCK_TOK; free($1); }
|    SUBBLOCKREG_TOK   { /* cout<<endl<<$1; */ curr_tok = SUBBLOCKREG_TOK; free($1); }
|    DELAY_TOK      { /* cout<<endl<<$1; */ curr_tok = DELAY_TOK; free($1); }
|    PINLIST_TOK    { /* cout<<endl<<$1; */ curr_tok = PINLIST_TOK; free($1); }
|    GLOBAL_TOK     { /* cout<<endl<<$1; */ curr_tok = GLOBAL_TOK; free($1); }
|    ANNOTATION_TOK     { /* cout<<endl<<$1; */ curr_tok = ANNOTATION_TOK; free($1); }
;

placement_info:
/* NETLISTFILE_TOK NET_TOK PERIOD_TOK NET_TOK{ */
/*   process_netlistFile($2); */
/*   free($1); free($2); */
/* } */
/* | ARCHFILE_TOK NET_TOK PERIOD_TOK NET_TOK{ */
/*   process_archFile($2); */
/*   free($1); free($2); */
/* } */
/* |  */
PLACEMENTFILE_TOK NET_TOK NET_TOK NET_TOK NET_TOK{
  process_placementFile($2, atoi($3), atoi($4), atoi($5));
  free($1); free($2); free($3); free($4); free($5);
};

nets:
|    NET_TOK nets   { /* cout<<" "<<$1; */ stringList.push_front($1); }
;
%%

/* input file that is to be parsed */
extern FILE* yyin;

extern vector <float> switches;

extern bool pipeline_aware_placement;

extern bool lut_level_placement;
extern bool clb_level_placement;
extern bool lut_level_placement;
extern bool clb_level_placement;
extern double ff_level_placement_activation;
extern double ff_level_placement_crit;
extern double ff_level_placement_probability;
extern double ff_level_placement_homing;
extern bool carefulHoming;
extern bool lutStickiness;
extern bool tvpack_initial_placement;
extern bool detectOrphanRegisters;

extern bool placementRetiming;
extern double retime_placement_activation;
extern double retime_placement_crit;
extern double retime_placement_probability;

extern bool splitDelayBlocks;
extern double splitDelayBlocks_activation;
extern double splitDelayBlocks_crit;
extern double splitDelayBlocks_probability;

extern bool unzipNets;
extern double unzipNets_activation;
extern double unzipNets_crit;
extern double unzipNets_probability;

extern int shiftRegisterMinPacking;
extern int shiftRegisterMaxPacking;

extern double tdPlacementLambda;
extern double tdPlacementUpdateInterval;
extern double tdPlacementCritExponentFinal;
extern bool noWindowing;
extern bool pipeline_aware_routing;
extern bool lockRegPlacement;
extern bool cut_occ_reg;
extern bool maxCutCost;
extern bool aveCutCost;
extern bool critical_path_headroom;
extern double critical_path_headroom_fac;
extern int pipelined_cost_model;
extern bool vanilla_timing;

extern bool zeroCLBWireCost;

extern bool testInitialPlacementOnly;

int main(int argc,char** argv){
  FILE* fp;char outfile_name [256];
  char placefile_name [256];
  char * tempStrPtrHead;
  char *tempStrPtrTail;
  time_t start_time,end_time;
  int i;

  int switchNum;
  bool input_placement = false;
  bool input_initialPlacement = false;
  bool timing_driven_placement = false;
  bool place_only = false;

  bool timing_driven_routing = false;
  
  char arguments[256];

  //  try{

    /*** command line error check ***/
    if(argc < 7){
      cout<<endl<<"Incorrect number of command line parameters. Usage is";
      cout<<endl<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file> <seed> [options]";
      cout<<endl<<"\t<netlist> : is a RaPiD netlist converted into the VPR format";
      cout<<endl<<"\t<num_clb_rows> : is the number of clb rows in the target array";
      cout<<endl<<"\t<num_clb_cols> : is the number of clb columns in the target array";
      cout<<endl<<"\t<io_rat> : is the number of io blocks per clb row/col at the periphery of the target array";
      cout<<endl<<"\t<rrgraph_file> : is the rr_graph.echo that EXACTLY matches the intended target array. Use VPR to generate this file...";
      cout<<endl<<"\t<seed> : is the seed to initialize the random number generator";

      cout<<endl<<"\tGENERAL OPTIONS LIST:";
      cout << endl << "\t[-num_switches <int> <floats>] : Provide the delay of N kinds of switches";
      cout << endl << "\t\tBy default they are all set to 0.0";
      cout<<endl<<"\t[-timing_driven] : option to perform both timing-driven Pathfinder and timing-driven placement (default = false)";
      cout << endl;
      cout<<endl<<"\tPLACEMENT OPTIONS LIST:";
      cout<<endl<<"\t[-placement filename] : option to input a placement for the netlist on a given architecture";
      cout<<endl<<"\t[-initialPlacement filename] : option to input an initial placement for the netlist on a given architecture";
      cout <<endl<<"\t[-lowTempAnneal] : option to perform additional low-temp anneal after placement has been determined(default = false)";
      cout<<endl<<"\t[-updateSlack] : option to update the slack of a link each time we make a move during placement (default = true)";
      cout<<endl<<"\t[-deltaDelayCost] : option to make timing-driven placement cost = delta delay * (link criticality) (default = false)";
      cout<<endl<<"\t[-zeroCLBWireCost] : option to make the wires inside a CLB zero cost";
      cout<<endl<<"\t[-noIOAnneal] : option to not anneal IO pins";
      cout<<endl<<"\t[-lotsTimingInfo] : option to print out sink criticalities after each temp iteration";

      cout<<endl<<"\t[-timing_driven_placement] : option to perform only timing-driven placement (default = false)";
      cout<<endl<<"\t[-pipeline_aware_placement] : option to perform pipelined routing with QuickRoute or QuickRoute-D (default = false)";
      cout << endl << "\t[-place_only] : option only produces placement and does not perform routing";

      cout << endl << "\t[-lut_level_placement]: This allows LUTs to move independently of one another -";
      cout << endl << "\t\tthis ignores packing (default = false)";
      cout << endl << "\t[-FF_level_placement <float W from [0.0 to 1.0]> <float X from [0.0 to 1.0]> <float Y from [0.0 to 1.0]> <float Z from [0.0 to 1.0]> [true|false]]:";
      cout << endl << "\t\tThis allows theFFs inside a CLB to move seperately from the logic.";
      cout << endl << "\t\tFF-level placement when the placement window is at W*array_size.";
      cout << endl << "\t\tAn activation point of 0.0001 kicks in when the window is size 1 and an ";
      cout << endl << "\t\tactivation point of 0.0 runs a post-placement low-temp annealing.";
      cout << endl << "\t\tWhen this kicks in, we will only apply this to blocks that are";
      cout << endl << "\t\t\tconnected to a signal X% critial or more.";
      cout << endl << "\t\tWhen we find a FF that satisfies the above requirements, and it is currently inside the same CLB as";
      cout << endl << "\t\t\tits parent, we will seperated the two with a probabilty of Y.  Otherwise we will move the whole CLB.";
      cout << endl << "\t\tWhen we find a FF that is not in the same CLB as its parent, regardless of criticality, we will try to";
      cout << endl << "\t\t\treturn it to its parent's CLB with a probabilty of Z.";
      cout << endl << "[-carefulHoming] : option that detects when a FF is trying to take the place of a parent register";
      cout << endl << "\t\tWhen we do register homing. (default = false)";
      cout << endl << "\t[-lutStickiness] : option that forces all FFs in the same CLB to move when a lut is moved when";
      cout << endl << "\t\tFF-level placement is performed.  That is, when we move a lut, the entire CLB is moved.";
      cout << endl << "\t\tto seperate a FF from a given CLB, we have to select the FF specifically. (default = move LUTs as";
      cout << endl << "\t\ta group seperate from the FFs with a probability of Y above)";
      cout << endl << "\t[-tvpack_initial_placement] : option pays special attention to where it puts pipelining registers";
      cout << endl << "\t\tduring the initial placement.  This is important because this is where we determine how the";
      cout << endl << "\t\tinitial CLB-level placement will progress.";
      cout << endl << "\t[-detectOrphanRegisters] : option to look for registers not in the same CLB as their host LUT (default = false)";

      cout << endl << "\t[-retime_placement <float X from (0.0 to 1.0)> <float Y from (0.0 to 1.0)>] <float Z from [0.0 to 1.0]>]: This allows the";
      cout << endl << "\t\tplacer to perform retiming moves during placement.";
      cout << endl << "\t\tWe will begin doing this when the movement window is at X*array_size.";
      cout << endl << "\t\tAn activation point of 0.0001 kicks in when the window is size 1 and an ";
      cout << endl << "\t\tactivation point of 0.0 runs a post-placement low-temp annealing.";
      cout << endl << "\t\tWhen this kicks in, we will only apply this to blocks that are connected to a signal Y% critial or more.";
      cout << endl << "\t\tWhen we find a block that is eligible for retiming, we will attempt it with a probability of Z.";
 
      cout << endl << "\t[-splitDelayBlocks <float X from (0.0 to 1.0)> <float Y from (0.0 to 1.0)> <float Z from [0.0 to 1.0]>] : PipeRoute's";
      cout << endl << "\t\tplacer assumed maximum sharing for all registers and packed them as densely";
      cout << endl << "\t\tas possible into bus connectors.  This option still requires maximum sharing among sinks, but allows";
      cout << endl << "\t\tthe placer to separate out registers less densely to improve timing. (default = false)";
      cout << endl << "\t\tWe will begin splitting/merging registers within a register bank when the placement window is at X*array_size.";
      cout << endl << "\t\tAn activation point of 0.0001 kicks in when the window is size 1 and an ";
      cout << endl << "\t\tactivation point of 0.0 runs a post-placement low-temp annealing.";
      cout << endl << "\t\tWhen this kicks in, we will only consider splitting registers that are part of a net that is Y% critical or more";
      cout << endl << "\t\tWhen we find a register that meets these requirements, we will attempt splitting the registers";
      cout << endl << "\t\twith a probability of Z.";

      cout << endl << "\t[-unzipNets <float X from (0.0 to 1.0)> <float Y from (0.0 to 1.0)> <float Z from [0.0 to 1.0]>] :";
      cout << endl << "\t\tPipeRoute's placer assumed maximum sharing for all registers and packed them as densely";
      cout << endl << "\t\tas possible into bus connectors.  This option allows the placer to duplicate registers and ";
      cout << endl << "\t\tzip or unzip nets at will to improve timing.  Automatically turns on -splitDelayBlocks option";
      cout << endl << "\t\t(default = false).";
      cout << endl << "\t\tWe will begin zipping/unzipping registers when the placement window is at X*array_size.";
      cout << endl << "\t\tAn activation point of 0.0001 kicks in when the window is size 1 and an ";
      cout << endl << "\t\tactivation point of 0.0 runs a post-placement low-temp annealing.";
      cout << endl << "\t\tWhen this kicks in, we will only consider duplicating registers that are part of a net that is Y% critical or more";
      cout << endl << "\t\ttWhen we find a register that meets these requirements, we will attempt duplication with a probability of Z.";

      cout << endl << "\t[-tdpLambda <float from [0.0 to 1.0]>] : option to increase timing emphasis during placement (default = 0.5)";
      cout << endl << "\t[-tdpUpdateInterval <float (0.0 to 1.0]>] : during a temperature iteration, how long do we go before at perform a timing analysis? (default = 1.0)";
      cout << endl << "\t\tat the very least we perform timing analysis after each temperature iteration. Smaller values will perform a STA more often";
      cout << endl << "\tand will result in more accurate critial path estimation. Howeved, this will make annealing run slower";
      cout << endl << "\t[-tdpCritExponentFinal <float>] : How much should the critiality affect the timing cost? (default = 8.0)";
      cout << endl << "\t\tlarger values (along with increasing lambda) will emphasize timing more heavily";
      cout << endl << "\t[-noWindowing] : disable range-limit windowing (default = false) - NOTE should not be used with";
      cout << endl << "\t\tanything but conventional CLB-based placement.";
      cout << endl << "\t[-place_inner_loop <float>] : Adjust placement effort by changing number of moves per temperature iteration (default = 10.0)";
      cout << endl << "\t[-shiftRegisterPacking <int from (1 to 16)> <int from (1 to 16)>] : If there are shift registers in the architecture,";
      cout << endl << "\t\tany time we see between X and Y registers in a row (without fanout), we'll try to put them into an";
      cout << endl << "\t\tavailable shift register location.  If there are none left, we'll just grab any register and";
      cout << endl << "\t\tdistribute them as usual.  This really only applies to the initial placement.";
      cout << endl;
      cout << endl << "\t[-testInitialPlacementOnly] : Just try to perform an initial placement, then exit.";
      cout << endl << "\t\tThis option is a handy and fast way to determine if a given array size is sufficient";
      cout << endl << "\t\tfor a specific netlist and shift register packing.";
      cout << endl<<"\tROUTING OPTIONS LIST:";
      cout<<endl<<"\t[-timing_driven_routing] : option to perform only timing-driven routing (default = false)";
      cout<<endl<<"\t[-pipeline_aware_routing] : option to perform pipelined routing with QuickRoute or QuickRoute-D (default = false)";
      cout << endl << "\t[-lockRegPlacement] : Since we considered registers during placement, bind registers at placement time - that is, do not";
      cout << endl << "\t\tuse Armada to route! (default = false)";
      cout << endl << "\t[-routeBoundingBox <int>] : option to increase or decrease the region we search during routing (default = 3)";
      cout << endl;
      cout << endl << "\tGENERALLY UNUSED OPTIONS:";
      cout << endl << "\t[-cut_occ_reg] : Architecture option to cut the helper node to helper node connection in occupied data regs (default = false)";
      cout << endl << "\t[-maxCutCost] : RaPiD-specific placement option that includes max cut into the placement cost function.";
      cout << endl << "\t\tAs per PipeRoute, if used, this weights max cut to contribute 30% to the placement cost. (default = false)";
      cout << endl << "\t[-aveCutCost] : RaPiD-specific placement option that replaces wirelength with average cut size -";
      cout << endl << "\t\tAUTOMATICALLY ASSUMES -maxCutCost OPTION! (default = false)";
      cout << endl << "\t[-critical_path_headroom 0.X] : pipelined routing option to calculate next iterations with X critical path headroom (default 0.0)";
      cout << endl << "\t[-pipelined_cost_model pathfinder | modified_pathfinder | linear | quadratic | cubic] : pipeline routing cost calculation style (default = modified_pathfinder)";
      cout << endl << "\t[-vanilla_timing] : critical path delay calculation option to assume all logic blocks are registered at source/sink disregarding impact on latency (default = false)";
      cout << endl << "\t\ttiming for pipeline aware routing is only minorly changed, but timing for pipeline unaware routes is";
      cout << endl << "\t\tsource->sink delay divided by theoretical latency";

      cout<<endl;
      exit(1);
    }

    //Print the arguments we are running with
    cout << "Arguments = ";
    for(i = 0; i < argc; i++){
      cout << argv[i] << " ";
    }
    cout << endl;

    //Skip over normal inputs and go to options
    for(i = 7; i < argc; i++){
      //General Options
      if(strcmp(argv[i], "-num_switches") == 0){
	if(i + 1 >= argc){
	  ERROR("-num_switches option requires argument");
	}
	switches.resize(atoi(argv[i+1]), 0.0);
	
	if(i + 1 + switches.size() >= argc){
	  ERROR("-num_switches does not have enough switch delays!");
	}

	for(switchNum = 0; switchNum < switches.size(); switchNum++){
	  switches.at(switchNum) = atof(argv[i + 2 + switchNum]);
	}

	i += 1 + switches.size();
      }
      else if(strcmp(argv[i], "-timing_driven") == 0){
	timing_driven_placement = true;
	timing_driven_routing = true;
      }
      //Placement options
      else if(strcmp(argv[i], "-placement") == 0){
	input_placement = true;
	//See if there is an argument
	i++;
	if(i < argc){
	  strcpy(placefile_name, argv[i]);
	}
	else{
	  ERROR("-placement option requires argument");
	}		      
      }
     else if(strcmp(argv[i], "-initialPlacement") == 0){
	input_initialPlacement = true;
	//See if there is an argument
	i++;
	if(i < argc){
	  strcpy(placefile_name, argv[i]);
	}
	else{
	  ERROR("-placement option requires argument");
	}		      
      }
      else if(strcmp(argv[i], "-lowTempAnneal") == 0){
	lowTempAnneal = true;		      
      }

      else if(strcmp(argv[i], "-updateSlack") == 0){
	updateSlack = true;		      
      }
      else if(strcmp(argv[i], "-deltaDelayCost") == 0){
	deltaDelayCost = true;		      
      }
      else if(strcmp(argv[i], "-zeroCLBWireCost") == 0){
	zeroCLBWireCost = true;		      
      }
      else if(strcmp(argv[i], "-noIOAnneal") == 0){
	noIOAnneal = true;		      
      }
      else if(strcmp(argv[i], "-lotsTimingInfo") == 0){
	lotsTimingInfo = true;		      
      }

      else if(strcmp(argv[i], "-timing_driven_placement") == 0){
	timing_driven_placement = true;
      }
      else if(strcmp(argv[i], "-pipeline_aware_placement") == 0){
	pipeline_aware_placement = true;
      }
      else if(strcmp(argv[i], "-place_only") == 0){
	place_only = true;
      }

      else if(strcmp(argv[i], "-lut_level_placement") == 0){
	ERROR("LUT-level placement incomplete!");
	lut_level_placement = true;
      }
      else if(strcmp(argv[i], "-FF_level_placement") == 0){
	if(i + 4 >= argc){
	  ERROR("-FF_level_placement option requires 4 floating point arguments");
	}
	ff_level_placement_activation = atof(argv[i + 1]);
	if(ff_level_placement_activation < 0.0 || ff_level_placement_activation > 1.0){
	  ERROR("-FF_level_placement activation factor must be [0.0, 1.0]");
	}
	ff_level_placement_crit = atof(argv[i + 2]);
	if(ff_level_placement_crit < 0.0 || ff_level_placement_crit > 1.0){
	  ERROR("-FF_level_placement crit factor must be [0.0, 1.0]");
	}
	ff_level_placement_probability = atof(argv[i + 3]);
	if(ff_level_placement_probability <= 0.0 || ff_level_placement_probability > 1.0){
	  ERROR("-FF_level_placement probability factor must be (0.0, 1.0]");
	}
	ff_level_placement_homing = atof(argv[i + 4]);
	if(ff_level_placement_homing < 0.0 || ff_level_placement_homing > 1.0){
	  ERROR("-FF_level_placement homing factor must be [0.0, 1.0]");
	}
	i+=4;
      }
      else if(strcmp(argv[i],"-carefulHoming") == 0){
	carefulHoming = true;
      }
      else if(strcmp(argv[i], "-lutStickiness") == 0){
	lutStickiness = true;		      
      }
      else if(strcmp(argv[i], "-tvpack_initial_placement") == 0){
	tvpack_initial_placement = true;		      
      }
      else if(strcmp(argv[i], "-detectOrphanRegisters") == 0){
	detectOrphanRegisters = true;		      
      }

      else if(strcmp(argv[i], "-retime_placement") == 0){
	if(i + 3 >= argc){
	  ERROR("-retime_placement option requires 3 floating point arguments");
	}
	retime_placement_activation = atof(argv[i + 1]);
	if(retime_placement_activation < 0.0 || retime_placement_activation > 1.0){
	  ERROR("-retime_placement activation factor must be [0.0, 1.0]");
	}
	retime_placement_crit = atof(argv[i + 2]);
	if(retime_placement_crit < 0.0 || retime_placement_crit > 1.0){
	  ERROR("-retime_placement crit factor must be [0.0, 1.0]");
	}
	retime_placement_probability = atof(argv[i + 3]);
	if(retime_placement_probability <= 0.0 || retime_placement_probability > 1.0){
	  ERROR("-retime_placement probability factor must be (0.0, 1.0]");
	}
	i+=3;
      }

      else if(strcmp(argv[i], "-splitDelayBlocks") == 0){
	if(i + 3 >= argc){
	  ERROR("-splitDelayBlocks placement option requires 3 floating point arguments");
	}
	splitDelayBlocks_activation = atof(argv[i + 1]);
	if(splitDelayBlocks_activation < 0.0 || splitDelayBlocks_activation > 1.0){
	  ERROR("-splitDelayBlocks placement activation factor must be [0.0, 1.0]");
	}
	splitDelayBlocks_crit = atof(argv[i + 2]);
	if(splitDelayBlocks_crit < 0.0 || splitDelayBlocks_crit > 1.0){
	  ERROR("-splitDelayBlocks placement crit factor must be [0.0, 1.0]");
	}
	splitDelayBlocks_probability = atof(argv[i + 3]);
	if(splitDelayBlocks_probability <= 0.0 || splitDelayBlocks_probability > 1.0){
	  ERROR("-unzipNets placement crit factor must be (0.0, 1.0]");
	}
	i+=3;
      }
      else if(strcmp(argv[i], "-unzipNets") == 0){
	//ERROR("Code not complete for duplicating pipelining registers!");
	if(i + 3 >= argc){
	  ERROR("-unzipNets placement option requires 3 floating point arguments");
	}
	unzipNets_activation = atof(argv[i + 1]);
	if(unzipNets_activation < 0.0 || unzipNets_activation > 1.0){
	  ERROR("-unzipNets placement activation factor must be [0.0, 1.0]");
	}
	unzipNets_crit = atof(argv[i + 2]);
	if(unzipNets_crit < 0.0 || unzipNets_crit > 1.0){
	  ERROR("-unzipNets placement crit factor must be [0.0, 1.0]");
	}
	unzipNets_probability = atof(argv[i + 3]);
	if(unzipNets_probability <= 0.0 || unzipNets_probability > 1.0){
	  ERROR("-unzipNets placement probability factor must be (0.0, 1.0]");
	}
	i+=3;
      }
 
      else if(strcmp(argv[i], "-tdpLambda") == 0){
	if(i + 1 >= argc){
	  ERROR("-tdpLambda option requires argument");
	}
	tdPlacementLambda = atof(argv[i + 1]);
	if(tdPlacementLambda < 0.0 || tdPlacementLambda > 1.0){
	  ERROR("-tdpLambda factor must be [0.0, 1.0]");
	}
	i++;
      }
      else if(strcmp(argv[i], "-tdpUpdateInterval") == 0){
	if(i + 1 >= argc){
	  ERROR("-tdpUpdateInterval option requires argument");
	}
	tdPlacementUpdateInterval = atof(argv[i + 1]);
	if(tdPlacementUpdateInterval <= 0.0 || tdPlacementUpdateInterval > 1.0){
	  ERROR("-tdpUpdateInterval factor must be (0.0, 1.0]");
	}
	i++;
      }
      else if(strcmp(argv[i], "-tdpCritExponentFinal") == 0){
	if(i + 1 >= argc){
	  ERROR("-tdpCritExponentFinal option requires argument");
	}
	tdPlacementCritExponentFinal = atof(argv[i + 1]);
	if(tdPlacementCritExponentFinal < 1.0){
	  ERROR("-tdpCritExponent factor must be greater than/equal to 1.0");
	}
	i++;
      }
      else if(strcmp(argv[i], "-noWindowing") == 0){
	noWindowing = true;
      }
      else if(strcmp(argv[i], "-place_inner_loop") == 0){
	if(i + 1 >= argc){
	  ERROR("-place_inner_loop option requires argument");
	}
	placeInnerLoop = atof(argv[i + 1]);
	if(placeInnerLoop < 1.0){
	  ERROR("-place_inner_loop factor should be >= 1.0");
	}

	i++;
      }
      else if(strcmp(argv[i], "-shiftRegisterPacking") == 0){
	if(i + 2 >= argc){
	  ERROR("-shiftRegisterPacking option requires 2 arguments");
	}
	shiftRegisterMinPacking = atoi(argv[i + 1]);
	if(shiftRegisterMinPacking < 1 || shiftRegisterMinPacking > 16){
	  ERROR("shiftRegisterMinPacking should be between 1 and 16");
	}

	shiftRegisterMaxPacking = atoi(argv[i + 2]);
	if(shiftRegisterMaxPacking < shiftRegisterMinPacking || shiftRegisterMaxPacking > 16){
	  ERROR("-shiftRegisterMaxPacking should be between shiftRegisterMinPacking and 16");
	}
	i+=2;
      }
      else if(strcmp(argv[i], "-testInitialPlacementOnly") == 0){
	testInitialPlacementOnly = true;
      }

      //Routing Options
      else if(strcmp(argv[i], "-timing_driven_routing") == 0){
	timing_driven_routing = true;
      }
      else if(strcmp(argv[i], "-pipeline_aware_routing") == 0){
	pipeline_aware_routing = true;
      }
      else if(strcmp(argv[i], "-lockRegPlacement") == 0){
	lockRegPlacement = true;
      }
      else if(strcmp(argv[i], "-routeBoundingBox") == 0){
	if(i + 1 >= argc){
	  ERROR("-routeBoundingBox option requires argument");
	}
	routeBoundingBox = atoi(argv[i + 1]);
	if(routeBoundingBox < 0){
	  ERROR("-routeBoundingBox factor must be zero or more");
	}

	i++;
      }

      //Unused options
      else if(strcmp(argv[i], "-cut_occ_reg") == 0){
	cut_occ_reg = true;
      }
      else if(strcmp(argv[i], "-maxCutCost") == 0){
	maxCutCost = true;
      }
      else if(strcmp(argv[i], "-aveCutCost") == 0){
	maxCutCost = true;
	aveCutCost = true;
      }
      else if(strcmp(argv[i], "-critical_path_headroom") == 0){
	if(i + 1 >= argc){
	  ERROR("-critical_path_headroom option requires argument");
	}
	critical_path_headroom = true;
	critical_path_headroom_fac = atof(argv[i + 1]);
	if(critical_path_headroom_fac <= 0.0 || critical_path_headroom_fac >= 1.0){
	  ERROR("-critical_path_headroom factor must be (0.0, 1.0)");
	}

	i++;
      }
      else if(strcmp(argv[i], "-pipelined_cost_model") == 0){
	if(i + 1 >= argc){
	  ERROR("-cost_model option requires argument");
	}
	if(strcmp(argv[i + 1], "pathfinder") == 0){
	  pipelined_cost_model = 0;
	}
	else if(strcmp(argv[i + 1], "modified_pathfinder") == 0){
	  pipelined_cost_model = 1;
	}
	else if(strcmp(argv[i + 1], "linear") == 0){
	  pipelined_cost_model = 2;
	}
	else if(strcmp(argv[i + 1], "quadratic") == 0){
	  pipelined_cost_model = 3;
	}
	else if(strcmp(argv[i + 1], "cubic") == 0){
	  pipelined_cost_model = 4;
	}
	else{
	  ERROR("Unrecognized cost model");
	}
	i++;
      }
      else if(strcmp(argv[i], "-vanilla_timing") == 0){
	vanilla_timing = true;
      }
      else{
	cout << argv[i] << " ??" << endl;
	ERROR("Unrecognized command line option");
      }
    }

    cout << "Using architecture " << argv[5] << endl;

    if(switches.size() == 0){
      cout << "WARNING: running with no defined switches!" << endl;
    }
    cout << "Switch delays: " << endl;
    for(i = 0; i < switches.size(); i++){
      cout << i << " delay = " << switches.at(i) << endl;
    }

    /* open input file */
    yyin=fopen(argv[1],"r");

    if(!yyin){
      cout<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file>"<<endl;
      ERROR("Could not open netlist file.\n");
    }

    /* read seed from command line */
    randomSeed=atoi(argv[6]);
    srand48(randomSeed);

    /* give the mcnc netlist its name */
    //Take out any preceeding path stuff from the netlist name to get the real name, not relative
    tempStrPtrHead = argv[1];
    tempStrPtrTail = strchr(argv[1], '/');
    while(tempStrPtrTail != 0){
      tempStrPtrHead = tempStrPtrTail + 1;
      tempStrPtrTail = strchr(tempStrPtrHead, '/');
    }

    mcnc_netlist = new mcncNetlist;
    mcnc_netlist->set_name(tempStrPtrHead);

    //Parse netlist input file
    while(!feof(yyin)){
      yyparse();
    }
    
    fclose(yyin);

    // Now that we have pulled in the mcnc netlist, 
    // convert it to our generic interal format
    //mcnc_netlist->print_mcncBlockList();
    netlist = convert2GenericNetlist(mcnc_netlist);

    if(atoi(argv[2]) < 1){
      cout<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file>"<<endl;
      ERROR("<num_clb_rows> needs to be a positive integer > 0.");
    }
    if(atoi(argv[3]) < 1){
      cout<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file>"<<endl;
      ERROR("<num_clb_cols> needs to be a positive integer > 0.");
    }
    if(atoi(argv[4]) < 1){
      cout<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file>"<<endl;
      ERROR("<io_rat> needs to be a positive integer > 0.");
    }

    // create vpr challenge architecture
    FPGA = create_vprFPGA(netlist, atoi(argv[2]),atoi(argv[3]),atoi(argv[4]));

    // parse rr_graph.echo file
    yyin=fopen(argv[5],"r");

    if(!yyin){
      cout<<endl<<"./armada <netlist> <num_clb_rows> <num_clb_cols> <io_rat> <rrgraph_file> [-placement filename]"<<endl;
      ERROR("Could not open rrgraph_file.\n");
    }

    while(!feof(yyin)){
      yyparse();
    }
    fclose(yyin);

    cout << endl << "Armada is running with the following options:" << endl;

    if(input_placement && input_initialPlacement){
      ERROR("Cannot provide both regular placement file and initial placement file!");
    }
    if(input_placement){
      cout << "Given placement from file " << placefile_name << endl;
    }
    if(input_initialPlacement){
      cout << "Given initial LUT placement from file " << placefile_name << endl;
      cout << "\tNOTICE: only LUTs and explicit registers have placements, no pipelining registers have been placed!" << endl;
    }

    if(!input_placement || lowTempAnneal){
      cout << "Generating ";
      if(timing_driven_placement){
	cout << "Timing-Driven ";
      }
      else{
	cout << "Wirelength-Driven ";
      }
      if(lut_level_placement){
	if(clb_level_placement){
	  ERROR("Cannot do lut AND clb level placement");
	}
	else{
	  cout << "LUT-level ";
	}
      }

      if(!clb_level_placement){
	cout << "FF-level ";
      }
      else{
	cout << "CLB-level ";
      }

      if(pipeline_aware_placement){
	cout << "Pipeline Aware ";
      }
      else{
	cout << "Pipeline Unaware ";
      }
      cout << "Placement" << endl;

      if(shiftRegisterMinPacking != 1000 || shiftRegisterMaxPacking != 1000){
	cout << "Packing netlist into shift registers from size " << shiftRegisterMinPacking << " to " <<
	  shiftRegisterMaxPacking <<endl;
      }

      if(ff_level_placement_activation > 0.0){
	if(ff_level_placement_activation != 1.0 && noWindowing){
	  ERROR("Cannot perform late activation ff-level placement without windowing!");
	}

	cout << "\tFF-level placement will be performed when the move window is less than or equal to " << ff_level_placement_activation 
	  * (float) max(FPGA->get_numRows(), FPGA->get_numCols()) << endl;
	cout << "\t\tFF-level placement will be performed on blocks of criticality greater than or equal to " << ff_level_placement_crit << endl;
	cout << "\t\twith a separation probability of " << ff_level_placement_probability << endl;
	cout << "\t\tand a homing probability of " << ff_level_placement_homing << endl;
      }
      else if(ff_level_placement_activation == 0.0){
	cout << "\tPost-processing low-temp anneal with FF-level placement will be performed." <<endl;
	cout << "\t\tFF-level placement will be performed on blocks of criticality greater than or equal to " << ff_level_placement_crit << endl;
	cout << "\t\twith a separation probability of " << ff_level_placement_probability << endl;
	cout << "\t\tand a homing probability of " << ff_level_placement_homing << endl;
      }
      
      if(timing_driven_placement){
	sprintf(arguments, "tdpLambda%1.3f_tdpExp%2.1f_FFPlaceAct%f_RetimingAct%f", tdPlacementLambda, tdPlacementCritExponentFinal,
		ff_level_placement_activation, retime_placement_activation);
	cout << "\tTiming-driven Lambda = " << tdPlacementLambda << endl;
	cout << "\tTiming-driven update interval = " << tdPlacementUpdateInterval << endl;
	cout << "\tTiming-driven Final Criticality Exponent = " << tdPlacementCritExponentFinal << endl;
      }
      if(pipeline_aware_placement){
	if(unzipNets_activation > 0.0){
	  if(unzipNets_activation != 1.0 && noWindowing){
	    ERROR("Cannot perform late activation register zipping/unzipping without windowing!");
	  }
	  
	  cout << "\tRegister zipping/unzipping placement will be performed when the move window is less than or equal to " << unzipNets_activation 
	    * (float) max(FPGA->get_numRows(), FPGA->get_numCols()) << endl;
	  cout << "\t\tRegister zipping/unzipping placement will be performed on blocks of criticality greater than or equal to " << unzipNets_crit << endl;
	  cout << "\t\twith an unzipping probability of " << unzipNets_probability << endl;
	}
	else if(unzipNets_activation == 0.0){
	  cout << "\tPost-processing low-temp anneal with register zipping/unzipping placement will be performed." << endl;
	  cout << "\t\tRegister zipping/unzipping placement will be performed on blocks of criticality greater than or equal to " << unzipNets_crit << endl;
	  cout << "\t\twith an unzipping probability of " << unzipNets_probability << endl;
	}
	else{
	  cout << "\tPipelined nets must maximally share registers during placement" << endl;

	  if(splitDelayBlocks_activation > 0.0){
	    if(splitDelayBlocks_activation != 1.0 && noWindowing){
	      ERROR("Cannot perform late activation register splitting without windowing!");
	    }
	    
	    cout << "\tRegister splitting placement will be performed when the move window is less than or equal to " << splitDelayBlocks_activation 
	      * (float) max(FPGA->get_numRows(), FPGA->get_numCols()) << endl;
	    cout << "\t\tRegister splitting placement will be performed on blocks of criticality greater than or equal to " << splitDelayBlocks_crit << endl;
	    cout << "\t\twith an splitting probability of " << splitDelayBlocks_probability << endl;
	  }
	  else if(splitDelayBlocks_activation == 0.0){
	    cout << "\tPost-processing low-temp anneal with register splitting placement will be performed." << endl;
	    cout << "\t\tRegister splitting placement will be performed on blocks of criticality greater than or equal to " << splitDelayBlocks_crit << endl;
	    cout << "\t\twith an splittingprobability of " << splitDelayBlocks_probability << endl;
	  }
	  else{
	    cout << "\tRegisters cannot be split during placement" << endl;
	  }
	}

	if(retime_placement_activation > 0.0){
	  if(retime_placement_activation != 1.0 && noWindowing){
	    ERROR("Cannot perform late activation retiming placement without windowing!");
	  }
	  
	  cout << "\tRetiming placement will be performed when the move window is less than or equal to " << retime_placement_activation 
	    * (float) max(FPGA->get_numRows(), FPGA->get_numCols()) << endl;
	  cout << "\t\tRetiming placement will be performed on blocks of criticality greater than or equal to " << retime_placement_crit << endl;
	  cout << "\t\twith a probability of " << retime_placement_probability << endl;
	}
	else if(retime_placement_activation == 0.0){
	  cout << "\tPost-processing low-temp anneal with retiming placement will be performed." << endl;
	    cout << "\tRetiming placement will be performed on blocks of criticality greater than or equal to " << retime_placement_crit << endl;
	    cout << "\t\twith a probability of " << retime_placement_probability << endl;
	}
	else{
	  cout << "\tRetiming is NOT performed" << endl;
	}

	if(maxCutCost){
	  cout << "\tTaking into account Max Cut Cost" << endl;
	}
	else{
	  cout << "\tCongestion is being calculated purely on wirelength, not max cut cost" << endl;
	}
	if(aveCutCost){
	  cout << "\tTaking into account Ave Cut Cost - no wirelength" << endl;
	}
	else{
	  cout << "\tCongestion is being calculated purely on wirelength, not ave cut cost" << endl;
	}
      }
    }
    if(place_only){
      cout << "\tRouting will not be performed" << endl;
    }
    else{
      if(timing_driven_routing){
	cout << "Performing Timing-Driven ";
      }
      else{
	cout << "Performing Congestion-Driven ";
      }
      if(pipeline_aware_routing && !lockRegPlacement){
	cout << "Armada ";
      }
      else if(pipeline_aware_routing && lockRegPlacement){
	cout << "Binding Register Placement PathFinder ";
      }
      else{
	cout << "PathFinder ";
      }
      cout << "Routing" << endl;

      cout << "Routing Bounding Box = " << routeBoundingBox << endl;

      if(pipeline_aware_routing){
	if(cut_occ_reg){
	  cout << "\tOccupied GPRs have the link between sink and source cut to facilitate routing" << endl;
	}
	else{
	  cout << "\tOccupied GPRs do not have the internal links cut, router must resolve any sharing" << endl;
	}
	if(critical_path_headroom){
	  cout << "\tWe have discouraged paths to become critical by including a headroom factor of " << critical_path_headroom_fac << endl;
	}
	if(timing_driven_routing){
	  if(pipelined_cost_model == 0){
	    cout << "\tUsing Original PathFinder cost function to balance congestion and delay - (crit * delay) + ((1 - crit) * congestionCost)" << endl;
	  }
	  if(pipelined_cost_model == 1){
	    cout << "\tUsing Armada cost function to balance congestion and delay - (crit/(1-crit) * delay) + congestionCost " << endl;
	  }
	  if(pipelined_cost_model == 2){
	    cout << "\tUsing linear cost function to balance congestion and delay - (crit * delay) + congestionCost" << endl;
	  }
	  if(pipelined_cost_model == 3){
	    cout << "\tUsing quadratic cost function to balance congestion and delay - (crit^2 * delay) + congestionCost" << endl;
	  }
	  if(pipelined_cost_model == 4){
	    cout << "\tUsing cubic cost function to balance congestion and delay - (crit^3 * delay) + congestionCost" << endl;
	  }
	}
      }
    }
    if(vanilla_timing){
      cout << "Calculating critial path timing by only taking into account point-to-point delay" <<endl;
      cout << "\t(all blocks are considered registered items) then simply dividing by the number of" << endl;
      cout << "\tregisters on that link" << endl; 
    }

    //Initialize the routing base costs for the architecture and make sure the
    //architecture is sufficient for the needs of the netlist
    initializeAndCheckArch(netlist, pipeline_aware_placement);

    //Initialize placement data structures
    /* set up a initial placement of the netlist on the FPGA */
    if(input_placement){
      // Placement is sucked in from file.
      yyin=fopen(placefile_name, "r");
      
      if(!yyin){
	ERROR("Could not open placement file.");
      }
      
      while(!feof(yyin)){
	yyparse();
      }
      fclose(yyin);

      //We grabbed the placement from an input file, but we would like to do something to it before routing.
      if(detectOrphanRegisters || lowTempAnneal){
	// start measuring the initialization time
	start_time=time(0);
	FPGA->initPlacementDS(testInitialPlacementOnly);
	end_time=time(0);
	cout<<endl<<endl<<"TOTAL PLACEMENT INITIALIZATION RUNTIME IN SECONDS : "<<end_time-start_time << endl;
	
	//Figure out the cost of this placement & estimate the critical path
	findCostAndCPofPlacement(netlist);

	if(lowTempAnneal){
	  ltAnneal(netlist, pipeline_aware_placement, timing_driven_placement);
	}

	//See if we are supposed to look for registers that are not in the same CLB as their host LUT
	if(detectOrphanRegisters){
	  findOrphanRegisters(netlist, timing_driven_placement);
	}
	
	//We don't have to worry about binding-versus non-binding placement.
	//When we input a placement we ALWAYS do binding placement.
	//Thus, just zero out the slack vectors.
	netlist->zeroSlackVectors();
	
	//Empty the timing table information - we don't need it anymore
	FPGA->emptyTimingTable();  
      }

      //Route this placement
      if(place_only){
	cout << "The '-place_only' option was used, does this make sense with the '-placement' option?" << endl;
      }
      else{
	//If we simply sucked in the placement, the sinks will not be sorted.
	//Depending upon what kind of routing we are doing, we should sort the sinks
	//before routing
	//Pipeline unaware:
	//  If we are doing congestion-driven routing we should sort the sinks
	//    by the normal A* distance based upon the row/column.
	//  If we are doing timing-driven routing we normally sort based upon timing
	//    but since this is the first iteration we don't know anything about
	//    the timing yet so we can just use the normal A*-type sorting
	//Pipeline aware:
	//  If we are doing either type of routing we should first sort sinks
	//    in non-decreasing order - unpipelined links to the highest latency
	//    sinks.  Withing a given latency we can sort based on the normal A* basis.
	if(!pipeline_aware_routing){
	  /* to facilitate A*, sort sinks in increasing order of separation from source */
	  netlist->sort_netSinksDistance();
	}
	else{
	  if(!lockRegPlacement){
	    netlist->sort_netPipelineSinks(false);
	  }
	  else{
	    ERROR("Locked registers are not supported with input placement - placements already describe fixed registers!");
	    //netlist->sort_netPipelineSinks_LockedRegs(timing_driven_routing);
	  }
	}

	route(netlist, pipeline_aware_routing, timing_driven_routing);
      }
    }
    else{
      //Take out any preceeding path stuff from the rrgraph name to get the real name, not relative
      tempStrPtrHead = argv[5];
      tempStrPtrTail = strchr(argv[5], '/');
      while(tempStrPtrTail != 0){
	tempStrPtrHead = tempStrPtrTail + 1;
	tempStrPtrTail = strchr(tempStrPtrHead, '/');
      }
      
      //Open a placement output file
      //if(!testInitialPlacementOnly){
	sprintf(outfile_name,"%s_%s_%s_%ld.out",netlist->get_name().c_str(),arguments, tempStrPtrHead,randomSeed);
	fp=fopen(outfile_name,"w");
	fprintf(fp,"\n\n\nNetlist is %s. Routing graph built from %s",argv[1],argv[5]);
	//}
  
      // start measuring the initialization time
      start_time=time(0);
      FPGA->initPlacementDS(testInitialPlacementOnly);
      end_time=time(0);
      cout<<endl<<endl<<"TOTAL PLACEMENT INITIALIZATION RUNTIME IN SECONDS : "<<end_time-start_time << endl;

      //Perform placement
      if(input_initialPlacement){
	// Placement is sucked in from file.
	yyin=fopen(placefile_name, "r");
	
	if(!yyin){
	  ERROR("Could not open placement file.");
	}
	
	while(!feof(yyin)){
	  yyparse();
	}
	fclose(yyin);
      }	

      place(netlist, pipeline_aware_placement, timing_driven_placement, fp, !input_initialPlacement);
    
      if(testInitialPlacementOnly){
	return 0;
      }
      if(ff_level_placement_activation == 0.0 || retime_placement_activation == 0.0 ||
	 unzipNets_activation == 0.0 || lowTempAnneal){
	ltAnneal(netlist, pipeline_aware_placement, timing_driven_placement);
      }

      if(detectOrphanRegisters){
	findOrphanRegisters(netlist, timing_driven_placement);
      }

      //If we are making the placement register binding, don't clean this information up
      // because we'll need it later on.
      if(!lockRegPlacement){
	//Go through the netlists and clear all of the link delay and slack vectors.
	netlist->clearLinkDelayAndSlackVectors();
	cleanUpPipelinePlacement(netlist);
      }
      //If the placement is binding, then just zero out all of the slack vectors
      else{
        netlist->zeroSlackVectors();
      }
 
      //Empty the timing table information - we don't need it anymore
      FPGA->emptyTimingTable();

      printVPRPlacement(netlist, randomSeed, tempStrPtrHead, arguments, fp);

      //We also have to dump a current netlist - we have created and placed pipeline registers
      // and we may have retimed the netlist.
      netlist->printNewNetlist(randomSeed, tempStrPtrHead, arguments);

      fclose(fp);

      //Since we are most likely not doing Independence placement, the sinks will not be sorted.
      //Depending upon what kind of routing we are doing, we should sort the sinks
      //before routing
      //Pipeline unaware:
      //  If we are doing congestion-driven routing we should sort the sinks
      //    by the normal A* distance based upon the row/column.
      //  If we are doing timing-driven routing we normally sort based upon timing
      //    but since this is the first iteration we don't know anything about
      //    the timing yet so we can just use the normal A*-type sorting
      //Pipeline aware:
      //  If we are doing pipeline-aware routing (timing or congestion driven), we should first sort sinks
      //    in non-decreasing order - unpipelined links to the highest latency
      //    sinks.  Withing a given latency we can sort based on the normal A* basis.
      //  If the placement of the registers is binding, we should do either Manhattan distance or use the
      //    estimated timing from placement to sort the sinks of all nets and pipelined nets.
      //    Even if we did not do timing-driven placement, we have slack numbers from the critical path
      //    timing calculations.
      if(!pipeline_aware_routing){
	/* to facilitate A*, sort sinks in increasing order of separation from source */
	netlist->sort_netSinksDistance();
      }
      else{
	if(!lockRegPlacement){
	  netlist->sort_netPipelineSinks(false);
	}
	else{
	  netlist->sort_netPipelineSinks_LockedRegs(timing_driven_routing);
	}
      }

      //Route this placement
      if(!place_only){
	route(netlist, pipeline_aware_routing, timing_driven_routing);
      }
    }

    return 0;
    //}
    //catch(out_of_range){
    // cerr<<"vector overflow error...\n";
    // }
}


/***
 * Error reporting function for the yacc parser.
 ***/
/*  void yyerror(char* s){ */
/*    fprintf(stderr,"%s AND ERRONEOUS TOKEN IS %s\n",s,yytext); */
/*  } */

//Routines for sucking in the netlist file
/***
 * curr_tok, stringList, curr_clb, and curr_block are declared in mcncParser.y.  
 * If curr_tok is INPUT_TOK, OUTPUT_TOK or CLB_TOK, create a new object
 * of mcncClb type. If curr_tok is GLOBAL_TOK, do
 * nothing for the time being. 
 * If curr_tok is SUBBLOCK_TOK, make a new LUT.  If curr_tok is 
 * SUBBLOCKREG_TOK, make a new FF.
 * Finally, if curr_tok is PINLIST_TOK, use
 * stringList to assign the various fields of curr_clb or curr_block. 
 ***/
void processCircuitInfo(char* annotation_string){

  if(curr_tok==INPUT_TOK)
    process_mcncInputBlock(annotation_string);
  else if(curr_tok==OUTPUT_TOK)
    process_mcncOutputBlock(annotation_string);
  else if(curr_tok==CLB_TOK)
    process_mcncClb(annotation_string);
  else if(curr_tok==TYPE_TOK)
    process_mcncType();
  else if(curr_tok==PINLIST_TOK)
    process_mcncPinlist(annotation_string);

  /* For the time being, do nothing about global signals */
  else if(curr_tok==GLOBAL_TOK){
    ;
  }
  else if(curr_tok==SUBBLOCK_TOK){
    process_mcncLUT();
  }
  else if(curr_tok==SUBBLOCKREG_TOK){
    process_mcncRegister();
  }
  else if(curr_tok==DELAY_TOK){
    process_mcncLatency();
  }
}

/***
 * Create a new input block and set its name using stringList.
 ***/
void process_mcncInputBlock(char* annotation_string){

  /* an input block does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating an input block...");
  }

  /* create a new mcncIO block */
  mcncBlock* input_block = new mcncBlock;

  /* give it a name */
  input_block->set_name(*(stringList.begin()));

  /* make it an input block */
  input_block->set_blockType(MCNC_INPUT);

  /* set curr_clb to point to this block */
  curr_clb = input_block;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist->addto_mcncBlockList(input_block);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING INPUT BLOCK WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name<<endl;
#endif
}

/***
 * Create a new output block and set its name using stringList. 
 ***/
void process_mcncOutputBlock(char* annotation_string){

  /* an output block does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating an output block...");
  }

  /* create a new mcncIO block */
  mcncBlock* output_block = new mcncBlock;

  /* give it a name */
  output_block->set_name(*(stringList.begin()));

  /* make it an output block */
  output_block->set_blockType(MCNC_OUTPUT);

  /* set curr_clb to point to this block */
  curr_clb = output_block;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist->addto_mcncBlockList(output_block);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING OUTPUT BLOCK WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name<<endl;
#endif
}

/***
 * Create a new clb and set its name using stringList. 
 ***/
void process_mcncClb(char* annotation_string){

  /* a CLB does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating a clb...");
  }

  /* create a new clb */
  mcncBlock* clb = new mcncBlock;
  clb->set_blockType(MCNC_CLB);

  /* give it a name */
  clb->set_name(*(stringList.begin()));

  /* set curr_clb to point to this block */
  curr_clb = clb;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist->addto_mcncBlockList(clb);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING CLB WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name;
#endif

}

//This will set the type of curr_clb.  Right now, only clbs should
//have a type since it's obvious what type inputs and outputs are
void process_mcncType(){
  mcncBlock* clb=0;

  if(pinlist_tok != CLB_TOK)
    ERROR("Only clbs can have a type token");

  clb = curr_clb;

  if(!clb)
    ERROR("error in locating block to set type");

  if(strcmp(*(stringList.begin()), "CLB") != 0)
    ERROR("Unknown block type");
}

//We found a ".subblock" token, so we should create a new LUT
void process_mcncLUT(){
  /* create a new LUT */
  mcncBlock* lut = new mcncBlock;
  lut->set_blockType(MCNC_LUT);

  /* give it a name */
  lut->set_name(*(stringList.begin()));

  /* set curr_block to point to this block */
  curr_block = lut;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to the current clb */
  if(!curr_clb){
    ERROR("No CLB to add LUT to!");
  }
  curr_clb->addLUT(lut);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING LUT WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name;
#endif
}


//We found a ".subblockReg" token, so we should create a new
// delay block
void process_mcncRegister(){
  /* create a new reg */
  mcncBlock* reg = new mcncBlock;
  reg->set_blockType(MCNC_REG);

  /* give it a name */
  reg->set_name(*(stringList.begin()));

  /* set curr_block to point to this block */
  curr_block = reg;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to the current clb */
  if(!curr_clb){
    ERROR("No CLB to add register to!");
  }
  curr_clb->addRegister(reg);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING REGISTER WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name;
#endif
}

//This will grab the latency values for each of the inputs of the curr_clb or curr_block
//This must occur after the pinlist has been taken in!
void process_mcncLatency(){
  mcncBlock* iob=0;
  mcncBlock* block=0;
  mcncNet* net;
  int latency;
  list<char*>::iterator s_itr;
  list<mcncNet*>::iterator n_itr;

  /* determine the type of block for which this latency value is being processed */
  if(pinlist_tok != SUBBLOCK_TOK && pinlist_tok != SUBBLOCKREG_TOK && pinlist_tok != OUTPUT_TOK)
    ERROR("Only clbs or outputs can have a latency token");

  if(pinlist_tok==OUTPUT_TOK)
    iob = curr_clb;
  else if(pinlist_tok==SUBBLOCK_TOK || pinlist_tok == SUBBLOCKREG_TOK)
    block = curr_block;
  else
    ERROR("incorrect token type here...");

  if(!iob && !block)
    ERROR("error in locating block whose latency is being processed...");

  if(iob){
    //See if there is something connected to the input
    if(iob->inputList.empty())
      ERROR("Output pin list must be defined before latency");

    //See if there is a latency number
    if(stringList.size() != 1)
      ERROR("Output pin latency list is must have exactly 1 value");

    latency = atoi(*stringList.begin());
    if(latency < 0){
      cout << "Problem with IOB " << iob->get_name() << endl;
      ERROR("Encountered negative latency");
    }
    else{
      if(MAXLATENCY < latency){
	MAXLATENCY = latency;
      }
      iob->pushBackInputLatencyVector(latency);
    }
  }
  else if(block){
    //This is either a LUT or a register
    if(stringList.size() == 0)
      ERROR("Empty latency list");

    if(stringList.size() != block->inputList.size()){
      ERROR("Mismatch between input list and latency list!");
    }

    for(s_itr = stringList.begin(), n_itr = block->inputList.begin();
	s_itr != stringList.end() && n_itr != block->inputList.end();
	s_itr++, n_itr++){

      latency = atoi(*s_itr);

      if(latency < 0)
	ERROR("Encountered negative latency");

      if(MAXLATENCY < latency){
	MAXLATENCY = latency;
      }

      //There is a net connected to this input
      if(*n_itr){
	block->pushBackInputLatencyVector(latency);
      }
      //There is not a net connected to this input
      else{
	if(!latency == 0)
	  ERROR("Unconnected pin cannot have latency");
	block->pushBackInputLatencyVector(latency);
      }
    }

    //Check to make sure that there were the same number of inputs and delays
    if(!(s_itr == stringList.end() && n_itr == block->inputList.end()))
      ERROR("Input pin/latency mismatch");
  }
}

/***
 * Process pinlist for the block that is pointed to by curr_clb or curr_block. 
 *
 * Each net in the pinlist is first searched in mcnc_netlist by name. 
 * If found, the returned net is made an input / output net in case 
 * curr_clb is of type mcncIO, or added to fanin-list if curr_el is of type
 * mcncClb.    
 ***/
void process_mcncPinlist(char* annotation_string){
  mcncBlock* iob=0;
  mcncBlock* block=0;
  mcncNet* net;
  int i;

  /* a pinlist does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating a pinlist...");
  }

  /* determine the type of block for which this pinlist is being processed */
  if(pinlist_tok==INPUT_TOK || pinlist_tok==OUTPUT_TOK)
    iob = curr_clb;
  else if(pinlist_tok==SUBBLOCK_TOK || pinlist_tok == SUBBLOCKREG_TOK)
    block = curr_block;
  else
    ERROR("incorrect token type here...");

  if(!iob && !block)
    ERROR("error in locating block whose pinlist is being processed...");

  /* if we have an mcncIO block, set its input / output net */
  if(iob){
    //If we have an
    // IO block, pinlist should contain exactly 1 net. 
    if(!stringList.size() == 1)
      ERROR("IOB must have exactly one net defined");

    if(strcmp(*stringList.begin(),"open")==0){
      ERROR("IO pin should be connected");
    }
    else{
      /* look for the net */
      net = find_mcncNet(*stringList.begin());
    }
    /* set its input / output net */
    /* update net data */
    if(iob->get_blockType() == MCNC_OUTPUT){
      iob->addto_inputList(net);
      net->addto_sinkList(iob);
    }
    else if(iob->get_blockType() == MCNC_INPUT){
      iob->addto_outputList(net);
      net->set_source(iob);
    }
    else
      ERROR("io block must be either input_only or output_only...");
  }

  /* we have a LUT or a FF */
  else if(block){
    if(pinlist_tok==SUBBLOCK_TOK){
      //We've got a LUT
      //Since we are inputting 4-LUTs, each lut has 4 inputs and 1 output.
      // some or all of which can be marked 'open'.  
      list<char*>::iterator s_itr = stringList.begin();
      for(i=1;s_itr!=stringList.end() && i <= 5;i++,s_itr++){
	//When we find an "open", make net = null
	if(strcmp(*s_itr,"open")==0){
	  net = 0;
	}
	else{
	  /* look for the net */
	  net = find_mcncNet(*s_itr);
	}
	
	/* first 4 pins correspond to lut inputs */
	/* add LUT to net's sinkList */
	if(i<=4){
	  block->addto_inputList(net);
	  if(net)
	    net->addto_sinkList(block);
	}
	/* Pin 5 is a lut output */
	/* make lut the source of each net */
	else if(i == 5){
	  block->addto_outputList(net);
	  if(net)
	    net->set_source(block);
	}
      }
      //Check to make sure that all i/o are defined
      if(!(i == 6 && s_itr == stringList.end())){
	cout << "Problem with lut " << block->get_name() << endl;
	cout << "i = " << i << " stringList = " << *stringList.back() << endl;
	ERROR("I/O mismatch on lut");
      }
    }

    else if(pinlist_tok==SUBBLOCKREG_TOK){
      //We've got a FF
      //Since we are inputting a FF, each one has 1 input and 1 output.
      // some or all of which can be marked 'open'.  
      list<char*>::iterator s_itr = stringList.begin();
      for(i=1;s_itr!=stringList.end() && i <= 2;i++,s_itr++){
	//When we find an "open", make net = null
	if(strcmp(*s_itr,"open")==0){
	  net = 0;
	}
	else{
	  /* look for the net */
	  net = find_mcncNet(*s_itr);
	}
	
	/* first pin corresponds to the FF input */
	if(i<=1){
	  block->addto_inputList(net);
	  if(net)
	    net->addto_sinkList(block);
	}
	/* Pin 2 is a lut output */
	/* make lut the source of each net */
	if(i == 2){
	  block->addto_outputList(net);
	  if(net)
	    net->set_source(block);
	}
      }
      //Check to make sure that all i/o are defined
      if(!(i == 3 && s_itr == stringList.end())){
	cout << "Problem with FF " << block->get_name() << endl;
	cout << "i = " << i << " stringList = " << *stringList.back() << endl;
	ERROR("I/O mismatch on FF");
      }
    }
#if defined(PARSER_DEBUG)
    cout<<"PROCESSING PINLIST:";
    list<char*>::iterator str_itr = stringList.begin();
    for(;str_itr!=stringList.end();str_itr++)
      cout<<endl<<"\t"<<*str_itr;
#endif
  }
}

/***
 * Look for a net with name net_name in mcnc_netlist. If found, return it
 * else create a new net, give it a name, add it to mcnc_netlist and then
 * return it.
 ***/
mcncNet* find_mcncNet(char* net_name){
  mcncNet* return_net;
  if(!net_name)
    ERROR("name of a net cannot be NULL...");
  if(return_net = mcnc_netlist->found_mcncNet(net_name))
    return return_net;
  else{
    return_net = new mcncNet;
    return_net->set_name(net_name);
    mcnc_netlist->addto_mcncNetList(return_net);
    return return_net;
  }
}


//Routines for sucking in the RR graph file
/***
 * Look for a routing node with name node_name. If one doesn't exist, create
 * a new one and return.
 ***/
routingNode* find_rnode(char* rnode_name){
  routingNode* return_node=0;
  string rname = rnode_name;

  if(!rnode_name)
    ERROR("name of a routing node cannot be NULL...");

/*   cout<<endl<<"String object name is "<< rname; */

  if(return_node = rNodeMap[rname]){
    ;
  }
  else{
    return_node = new routingNode(MAXLATENCY);
    return_node->set_name(rnode_name);
    FPGA->addto_rnodeList(return_node);
    rNodeMap[rname]=return_node;
/*     cout<<". String read-back from map is "<<(rNodeMap[rname])->name; */
  }

  return return_node;
}

void process_blockName(char* name){
  G_currNode->set_blockName(name);
}

void process_blockType(char* name){
  if(strcmp(name, "INPUTPAD") == 0){
    //G_currNode->set_blockType(INPUTPAD);
  }
  else if(strcmp(name, "OUTPUTPAD") == 0){
    //G_currNode->set_blockType(OUTPUTPAD);
  }
  else if(strcmp(name, "CLB") == 0){
    //G_currNode->set_blockType(LUT);
    vprLUT* lut = FPGA->getCLBAt(G_currNode->row_low, G_currNode->col_low)->getLUT(G_currPTC);
    lut->set_blockType(LUT);
  }
  else{
    cout << name << endl;
    ERROR("Unknown block type");
  }
}

//void process_wireName(char* name){
//  G_currNode->set_wireName(name);
//}

void process_sourceNode(int col, int row, int ptc){
  int i;

  if(col < 0 || col > (FPGA->numCols + 1) || row < 0 || row > (FPGA->numRows + 1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(SOURCE);
  G_currNode->row_low = G_currNode->row_high = row;
  G_currNode->col_low = G_currNode->col_high = col;
  G_currPTC = ptc;

  if(col == 0 || col == (FPGA->numCols+1) || row == 0 || row == (FPGA->numRows+1)){
    // We have an IOB here
    vprIOB* io_block = FPGA->getOrMakeIOBAt(row, col);

    if(ptc >= io_block->numIOs){
      cout<<endl<< ptc <<"\t";
      ERROR("PTC number beyond I/O ratio!");
    }

    if(io_block->getfrom_sourceArray(ptc)==0){
      io_block->addto_sourceArray(G_currNode,ptc);      
    }
    else{
      ERROR("Source has already been designated!");
    }
  }
  else{
    // We have a LUT here
    vprLUT* lut = FPGA->getOrMakeCLBAt(row, col)->getOrMakeLUT(ptc);
    lut->set_sourceNode(G_currNode);
  }
  //G_currNode->baseCost = 1.0;
}

//Input pins don't need to be associated with a particular location - only sources and sinks
void process_ipinNode(int col, int row, int ptc){
  int i;

  if(col < 0 || col > (FPGA->numCols + 1) || row < 0 || row > (FPGA->numRows + 1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(INPUT_PIN);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;
  G_currPTC = ptc;

  /*if(col==0 || col==(FPGA->numCols+1) || row==0 || row==(FPGA->numRows+1)){
    // We have an IOB here
    vprIOB* io_block=(vprIOB *)(FPGA->get_blockAt(row, col));

    if(ptc >= io_block->numIOs){
      cout<<endl<<i<<"\t";
      ERROR("PTC number beyond I/O ratio!");
    }

    if(io_block->getfrom_inputPinArray(ptc) == 0){
      io_block->addto_inputPinArray(G_currNode, ptc);
    }
  }
  else{
    // We have a CLB here
    vprLUT* lut=(vprLUT *)(FPGA->get_blockAt(row,col));
    lut->addto_inputPinList(G_currNode);
  }*/
  //G_currNode->baseCost = 0.95;
}



void process_sinkNode(int col, int row, int ptc){
  int i;

  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(SINK);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;
  G_currPTC = ptc;

  if(col==0 || col==(FPGA->numCols+1) || row==0 || row==(FPGA->numRows+1)){
    // We have an IOB here
    vprIOB* io_block = FPGA->getOrMakeIOBAt(row,col);

    if(ptc >= io_block->numIOs){
      cout<<endl<<i<<"\t";
      ERROR("PTC number beyond I/O ratio!");
    }

    if(io_block->getfrom_sinkArray(ptc)==0){
      io_block->addto_sinkArray(G_currNode, ptc);
    }
  }
  else{
    // We have a LUT here
    //This is where we will add in subblock # information - eventually...
    vprLUT* lut = FPGA->getOrMakeCLBAt(row,col)->getOrMakeLUT(ptc);
    lut->set_sinkNode(G_currNode);
  }  
  //G_currNode->baseCost = 0.0;
}

//Output pins don't need to be associated with a particular location - only sources and sinks
void process_opinNode(int col, int row, int ptc){
  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(OUTPUT_PIN);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;
  G_currPTC = ptc;
  
  //G_currNode->baseCost = 1.0;
}


void process_chanNode(int col_low,int row_low,int col_high,int row_high){
  if(col_low<0 || col_high<0 || col_low>(FPGA->numCols+1) || col_high>(FPGA->numCols+1))
    ERROR("array out of bounds here...");
  if(row_low<0 || row_high<0 || row_low>(FPGA->numRows+1) || row_high>(FPGA->numRows+1))
    ERROR("array out of bounds here...");

  if(row_low > row_high){
    ERROR("Row low is larger than row high!");
  }

  if(col_low > col_high){
    ERROR("Col low is larger than row high!");
  }

  G_currNode->set_type(BUS_OR_WIRE);

  G_currNode->row_low=row_low;G_currNode->row_high=row_high;
  G_currNode->col_low=col_low;G_currNode->col_high=col_high;
}

//void process_helperNode(int col, int row){
//  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
//    ERROR("array out of bounds here...");
//
//  G_currNode->set_type(HELPER);
//}

void process_Register(int col, int row, int bank){
  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");

  // When we encounter a register node, we should create a matching "delay block" with the correct row and column.
  //delayBlock* db = FPGA->getOrMakeDelayBlockAt(row, col);
  vprCLB* clb = FPGA->getOrMakeCLBAt(row, col);
  
  //Add this routing node to the proper bank in the delay block
  clb->addRoutingNodeToRegisterBank(G_currNode, bank);
  
  G_currNode->set_type(DELAY);
  G_currNode->row_low = row;  G_currNode->row_high = row;
  G_currNode->col_low = col;  G_currNode->col_high = col;
  G_currNode->setBank(bank);
}

void process_RegisterChain(int col, int row, int bank, int regNumber, int lutAssociation){
  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");

  // When we encounter a register node, we should create a matching "delay block" with the correct row and column.
  //delayBlock* db = FPGA->getOrMakeDelayBlockAt(row, col);
  vprCLB* clb = FPGA->getOrMakeCLBAt(row, col);
  
  //Add this routing node to the proper bank in the delay block
  clb->addRoutingNodeToRegisterBank(G_currNode, bank, regNumber, lutAssociation);
  
  G_currNode->set_type(DELAY);
  G_currNode->row_low = row;  G_currNode->row_high = row;
  G_currNode->col_low = col;  G_currNode->col_high = col;
  G_currNode->setBank(bank);
}


void process_Edges(int num_edges){
  int i;
  routingNode* fanout_node;
  routingEdge* r_edge;

  if(num_edges<0)
    ERROR("number of fanout edges cannot be -ve...");
  if(num_edges==0)
    return;

  list<char*>::iterator str_itr = stringList.begin();
  for(i=0;str_itr!=stringList.end();str_itr++,i++){
    fanout_node=find_rnode(*str_itr);
    r_edge=new routingEdge;
    r_edge->set_origNode(G_currNode);
    r_edge->set_destNode(fanout_node);
    G_currNode->addto_fanoutList(r_edge);
    fanout_node->addto_faninList(r_edge);
  }

  if(i!=num_edges)
    ERROR("problems while connecting routing nodes with edges...");

}

void process_switchTypes(){
  routingEdge* r_edge;
  int switchNum;

  if(stringList.size()==0)
    return;

  list<char*>::iterator str_itr = stringList.begin();
  G_currNode->create_fanoutListItr();

  for(;str_itr!=stringList.end();str_itr++){
    r_edge = G_currNode->getnext_fanoutList();
    switchNum = atoi(*str_itr);

    if(switchNum < 0 || switchNum >= switches.size()){
      ERROR("Invalid switch type!");
    }
    else{
      //r_edge->set_type(switchNum);
      r_edge->set_intrinsicDelay(switches.at(switchNum));
    }
  }

  G_currNode->remove_fanoutListItr();

}

//Routines for sucking in the placement file
/* void process_netlistFile(char* filename){ */
/*   cout << "Original Netlist = " << filename << endl; */
/* } */

/* void process_archFile(char* filename){ */
/*   cout << "Arch = " << filename << endl; */
/* } */

void process_placementFile(char* blockName, int col, int row, int ptc){
  //vprLUT* lut;
  vprIOB* iob;
  vprCLB* clb;
  registerBank* rb;
  logicBlock* lb;

  //Find this logic block in the netlist
  netlist->create_lbListItr();
  for(lb=netlist->getnext_lbList();lb!=0;lb=netlist->getnext_lbList()){
    //See if this logic block matches the name given
    if(strcmp(lb->get_name().c_str(), blockName) == 0){
      //cout << "Found lb = " << lb->get_name() << endl;
      netlist->remove_lbListItr();
      break;
    }
  }
  
  //Check to see if we found the right LB or just fell out of the loop
  if(!lb){
    ERROR("Cannot find logic block in netlist!");
  }
  else if(strcmp(blockName, lb->get_name().c_str()) != 0){
    ERROR("Cannot find logic block in netlist!");
  }

  //Make sure that this logic block has not been placed already
  if(!(lb->row == -1 && lb->col == -1)){
    cout << "Problem with logic block " << lb->get_name() << endl;
    cout << "Row = " << lb->row << " Col = " << lb->col << endl;
    ERROR("Logic block has been already placed!");
  }

  //Figure out what kind of block this lb is
  if(lb->get_blockType() == INPUTPAD || lb->get_blockType() == OUTPUTPAD){
    //Make sure that the row and column are on the rim
    if(col != 0 && col != (FPGA->numCols+1) && row != 0 && row != (FPGA->numRows+1)){
      ERROR("IOB must be place in IOB ring");
    }
    iob = FPGA->getIOBAt(row,col);
    placeBlock(lb,(void *) (iob), IOPIN, ptc);
  }
  else if(lb->get_blockType() == REGISTER_BANK){
    //Make sure that the row and column are not on the rim
    if(col <= 0 || col >= (FPGA->numCols+1) || row <= 0 || row >= (FPGA->numRows+1)){
      ERROR("Register must be placed inside FPGA");
    }
    rb = FPGA->getCLBAt(row, col)->getRegBank(ptc);
    if(!rb){
      ERROR("There is no register bank there!");
    }
    rb->placeLogicBlock(lb);
    lb->updateCLBLocation(rb);
  }
  else if(lb->get_blockType() == LUT){
    //Make sure that the row and column are not on the rim
    if(col <= 0 || col >= (FPGA->numCols+1) || row <= 0 || row >= (FPGA->numRows+1)){
      ERROR("Logic block must be placed inside FPGA");
    }
    //lut = FPGA->getCLBAt(row,col)->getLUT(ptc);
    //placeBlock(lb,(void *) (lut), LB, ptc);
    clb = FPGA->getCLBAt(row,col);
    placeBlock(lb,(void *) (clb), LB, ptc);
    lb->updateCLBLocation(clb->getLUT(ptc));
  }
}
