// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"

extern bool tvpack_initial_placement;

// This function converts the mcnc netlist to a generic netlist. This is the
// step that converts benchmark netlists of a particular format to a representation 
// that the Independence algorithm routines will recognize and
// operate upon.
Netlist* convert2GenericNetlist(mcncNetlist *mcnc_netlist){
  Netlist *netlist = new Netlist;
  mcnc_netlist->convert2Generic(netlist);

  delete mcnc_netlist;
  return netlist;
}

// Create a vpr challenge FPGA object.
vprFPGA* create_vprFPGA(Netlist* netlist, int num_clb_rows,int num_clb_cols,int num_ios){
  vprFPGA *FPGA;

  // Check whether target arrays is big enough
  if(netlist->get_numCLBs() > num_clb_rows * num_clb_cols){
    cout << "WARNING!  Target array may not be big enough - you may need to increase num clb rows & cols"<< endl;
    if(tvpack_initial_placement){
      cout << "Netlist has " << netlist->get_numCLBs() << " clbs!" << endl;
      ERROR("Target array is not big enough. Need to increase num clb rows/cols.");
    }
  }
  if(netlist->get_numInputs() + netlist->get_numOutputs() > num_ios * 2 * (num_clb_rows + num_clb_cols)){
    cout << "WARNING!  Target array may not be big enough - you may need to increase num IOBs per clb"<< endl;
    ERROR("Target array is not big enough. Need to increase num iobs per clb");
  }

  FPGA = new vprFPGA(num_clb_rows, num_clb_cols,num_ios);

  return FPGA;
}
