// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "include.h"
extern int MAXLATENCY;
extern vprFPGA *FPGA;

void initializeAndCheckArch(Netlist *netlist, bool pipeline_aware){
  double averageIntrinsicDelay;

  //FPGA->print_info();

  //Calculate the average intrinsic delay of all routing nodes
  //and use this value to scale base cost values
  averageIntrinsicDelay = FPGA->calculateAverageIntrinsicDelay();
  
  cout << "Average Intrinsic Delay of routing nodes = " << averageIntrinsicDelay << endl;
  FPGA->scaleBaseCosts(averageIntrinsicDelay);
  
  //Count number of alu, multipliers, reg and rams in both the netlist and vpr architecture
  //to make sure that we have enough of everything
  cout << "Netlist has " << netlist->countLUT() << " LUTs" << endl;
  cout << "Netlist has " << netlist->countMult() << " Multipliers" << endl;
  cout << "Netlist has " << netlist->countFixedReg() << " Fixed Registers" << endl;
  cout << "Netlist has " << netlist->countRAM() << " RAM Blocks" << endl;
  cout << "Netlist has " << netlist->get_numInputs() << " inputs" << endl;
  cout << "Netlist has " << netlist->get_numOutputs() << " outputs" << endl;

  //Set up a bunch of pipeline-aware data structures
  if(pipeline_aware){
    cout << "Netlist requires at least " << netlist->countDelayReg() << " Delay Registers" << endl;
    fillPathVectorMask();
  }
  
  cout << "Max latency of any sink " << MAXLATENCY << endl;
  
  if(netlist->countLUT() > FPGA->countLUT()){
    cout << "Architecture does not have enough LUTs" << endl;
    ERROR("Architecture does not have enough LUTs");
  }
  else if(netlist->countMult() > FPGA->countMult()){
    cout << "Architecture does not have enough multipliers" << endl;
    ERROR("Architecture does not have enough multipliers");
  }
  else if(netlist->countFixedReg() > FPGA->countRegBanks()){
    cout << "Architecture does not have enough registers" << endl;
    ERROR("Architecture does not have enough registers");
  }
  else if(netlist->countRAM() > FPGA->countRAM()){
    cout << "Architecture does not have enough RAMs" << endl;
    ERROR("Architecture does not have enough RAMs");
  }
  else if(netlist->get_numInputs() + netlist->get_numOutputs() > FPGA->get_numIOs()){
    cout << "Architecture does not have enough input/output pins" << endl;
    ERROR("Architecture does not have enough input/output pins");
  }

  if(pipeline_aware){
    if(netlist->countFixedReg() + netlist->countDelayReg() > FPGA->countDelayReg()){
      cout << "Architecture does not have enough registers to perform pipeline-aware placement/routing!";
      ERROR("Architecture does not have enough registers to perform pipeline-aware placement/routing!");
    }
  }
}
