// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#include "vprArch.h"

class vprFPGA;

void pipeline_placement(Netlist *netlist, bool timing_driven, bool performInitialPlacement);
void pipelineLowTempAnneal(Netlist *netlist, bool timing_driven);
void init_pipeline_placement(Netlist *netlist);
void init_pipeline_placement_OnlyPipeliningRegisters(Netlist *netlist);
//void initialPlacePipelineRegs(Net *net);
void cleanUpPipelinePlacement(Netlist *netlist);

double pipelineDetermineAnnealingTemp(Netlist *netlist, bool timing_driven, double *criticalPathDelay);
int generatePipelineAwareMove(Netlist *netlist, void** block0, void** block1, int *indexOrBank0, int *indexOrBank1, 
			      enum e_timingMoveBlockType *blockType, int *numMove, int rLim);
int selectPipelineAwareMove(Netlist *netlist);
void generatePipelineAwareMoveLogic(int *moveType, Netlist *netlist, void** block0, void** block1, int *index0, int *index1, 
				    enum e_timingMoveBlockType *blockType, int *numMove, int rLim);

void generateFFLevelHomingTest(logicBlock* lb, int *moveType, void** block0, void** block1, int rLim);

void generatePipelineAwareMovePipeReg(int *moveType, Netlist *netlist, void** block0, void** block1, int *numMove, int rLim);
void generatePipelineAwareMovePipeRegSwap(int *moveType, Netlist *netlist, registerBank** block0, registerBank** block1, int *numMove, int rLim);
void generatePipelineAwareMovePipeRegSplitOrSwap(int *moveType, Netlist *netlist, registerBank** block0, registerBank** block1, int *numMove, int rLim);
void generatePipelineAwareMovePipeRegSplitSwapOrUnzip(int *moveType, Netlist *netlist, registerBank** block0, registerBank** block1, int *numMove, int rLim);
void generatePipelineAwareMoveSplitReg(int *moveType, registerBank** block0, registerBank** block1, int *numMove, int rLim);

void generatePipelineAwareMoveMergeReg(int *moveType, Netlist *netlist, registerBank** block0, registerBank** block1, int *numMove);

bool generatePipelineAwareMoveRegRetimeBackward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove);
bool generatePipelineAwareMoveRegRetimeForward(int *moveType, Netlist *netlist, void** block0, enum e_timingMoveBlockType blockType, int *numMove);

bool makePipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, 
			   enum e_timingMoveBlockType blockType, int numMove,
			   double criticalPathDelay,   list <PipeliningNet *> *retimeUndoList); 

void undoPipelineAwareMove(int *moveType, Netlist *netlist, void** block0, void** block1, int index0, int index1, 
			   enum e_timingMoveBlockType blockType, int numMove,
			   double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);

double determineDeltaCostDeltaDelay(Netlist *netlist, int moveType, void** block0, void** block1, int index0, int index1, 
			  enum e_timingMoveBlockType blockType, int numMove,
			  bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);

double determineDeltaCost(Netlist *netlist, int moveType, void** block0, void** block1, int index0, int index1, 
			  enum e_timingMoveBlockType blockType, int numMove,
			  bool timing_driven, double criticalPathDelay, list <PipeliningNet *> *retimeUndoList);

void updateDeletedCutCostForwards(list <PipeliningNet *> *retimeUndoList);
void updateDeletedCutCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb);
double returnTotalWireCostForwards(list <PipeliningNet *> *retimeUndoList);
double returnTotalWireCostBackwards(list <PipeliningNet *> *retimeUndoList,  bool timing_driven, double criticalPathDelay, logicBlock *lb);
double returnTotalTimingCostForwards(list <PipeliningNet *> *retimeUndoList);
double returnTotalTimingCostBackwards(list <PipeliningNet *> *retimeUndoList, logicBlock *lb);

void clearRetimeUndoListForward(list <PipeliningNet *> *retimeUndoList);
void clearRetimeUndoListBackward(list <PipeliningNet *> *retimeUndoList, void *block, enum e_timingMoveBlockType blockType);
void clearNerfedBlocksBackward(void* block0, enum e_timingMoveBlockType blockType);
void clearNerfedBlocksForward(void* block0, enum e_timingMoveBlockType blockType);

void pipelineAnneal(Netlist *netlist, bool timing_driven, double initialTemperature, float startRLim);

double updateTemp(double currTemp, int movesAccepted, int movesPerTemp);
float updateRLimAndCritExp(float rLim, int movesAccepted, int movesPerTemp);

bool exitCondition(double currTemp, double cost, int numNets);

void findOrphanRegisters(Netlist *netlist, bool pipeline_aware);

void attemptRetiming(Netlist *netlist, double currTemp, double criticalPathDelay, bool timing_driven);
