// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/* define an entry of the priority queue */
class pqNode{
 public:
  routingNode* rNode;  //the routing node that this priority queue entry represents
  double pathCost;  //the cost of the path from the source to rNode
  routingNode* predNode;  //the fanin routing node that was used to explore rNode
  double heurEstimate;

  int magicNum;

  //Stuff I've added in
  double netDelay;  //The delay of the net up to this node
  PipeliningNet* sourcePNet;

  int assumed_crit_index;
  int latency;

  //Stuff needed for timing-driven pipelined routing
  double netCongestion;
  double externalDelay;
  double upstreamCost;

  pExplore pred;
  vector <unsigned int> *pathVector;

  pqNode(){
   rNode=0;
    pathCost=heurEstimate=DBL_MAX;
    predNode=0;
    magicNum=33111;
  }
  pqNode(routingNode* r,double p,double h,routingNode* pnode){
    if(r==0 || p<0. || h<0.)
      ERROR("error in creating a new PQ node...");
    rNode=r;
    pathCost=p;heurEstimate=h;
    predNode=pnode;
    magicNum=33111;
  }

  pqNode(routingNode* r,double p,double h,routingNode* pnode, double delay){
#ifdef ROUTER_DEBUG
    if(r==0 || p<0.0 || h<0.0 || delay < 0.0){
      if(r == 0){
	ERROR("error in creating a new PQ node - r is null...");
      }
      else if(p < 0.0){
	ERROR("error in creating a new PQ node - p < 0");
      }
      else if(h < 0.0){
	ERROR("error in creating a new PQ node - h < 0");
      }
      else if(delay < 0.0){
	ERROR("error in creating a new PQ node - delay < 0");
      }
      else{
	ERROR("unknown error!");
      }
    }
#endif
    rNode=r;
    pathCost=p;heurEstimate=h;
    netDelay = delay;
    predNode=pnode;
    magicNum=33111;
  }

  pqNode(PipeliningNet* pNet, routingNode* node, double cost, double delay, int acIndex, int lat, pExplore predecessorExplore, vector <unsigned int> *pVector){
    if(pNet == 0 || node == 0 || lat < 0 || cost < 0.0 || delay < 0.0 ||
       predecessorExplore.assumed_crit_index < 0 || predecessorExplore.assumed_crit_index >= NUMASSUMEDCRIT ||
       predecessorExplore.latency < 0 || predecessorExplore.exploreIndex < 0 || predecessorExplore.exploreIndex >= 2)
      ERROR("error in creating a new PQ node...");
    sourcePNet = pNet;
    rNode = node;
    pathCost = cost;
    netDelay = delay;
    assumed_crit_index = acIndex;
    latency = lat;
    pred = predecessorExplore;
    magicNum=33111;
    heurEstimate = 0.0;
    pathVector = pVector;
  }

  pqNode(PipeliningNet* pNet, routingNode* node, double cost, double delay, int acIndex, int lat, pExplore predecessorExplore, 
	 vector <unsigned int> *pVector, double congestion, double exDelay, double upCost){
    if(pNet == 0 || node == 0 || lat < 0 || cost < 0.0 || delay < 0.0 ||
       predecessorExplore.assumed_crit_index < 0 || predecessorExplore.assumed_crit_index >= NUMASSUMEDCRIT ||
       predecessorExplore.latency < 0 || predecessorExplore.exploreIndex < 0){
       //|| predecessorExplore.exploreIndex >= 2){
      if(pNet == 0){
	ERROR("error in creating a new PQ node - pNet is null...");
      }
      if(node == 0){
	ERROR("error in creating a new PQ node - node is null...");
      }
      else if(lat < 0.0){
	ERROR("error in creating a new PQ node - lat < 0");
      }
      else if(cost < 0.0){
	ERROR("error in creating a new PQ node - cost < 0");
      }
      else if(delay < 0.0){
	ERROR("error in creating a new PQ node - delay < 0");
      }
      else if(predecessorExplore.assumed_crit_index < 0){
	ERROR("error in creating a new PQ node - ACIndex < 0");
      }
      else if(predecessorExplore.assumed_crit_index >= NUMASSUMEDCRIT){
	ERROR("error in creating a new PQ node - ACIndex >= NUMASSUMEDCRIT");
      }
      else if(predecessorExplore.latency < 0){
	ERROR("error in creating a new PQ node - pred latency < 0");
      }
      else if(predecessorExplore.exploreIndex < 0){
	ERROR("error in creating a new PQ node - pred explore index < 0");
      }
/*       else if(predecessorExplore.exploreIndex >= 2){ */
/* 	ERROR("error in creating a new PQ node - pred explore index >= 2"); */
/*       } */
      else{
	ERROR("unknown error!");
      }
    }

    sourcePNet = pNet;
    rNode = node;
    pathCost = cost;
    netDelay = delay;
    assumed_crit_index = acIndex;
    latency = lat;
    pred = predecessorExplore;
    magicNum=33111;
    heurEstimate = 0.0;
    pathVector = pVector;
    netCongestion = congestion;
    externalDelay = exDelay;
    upstreamCost = upCost;
  }

  ~pqNode(){
    if(magicNum!=33111)
      ERROR("trying to double delete a node...");
    else{
      magicNum=0;
      //delete pathVector;
    }
  }

  bool operator< (const pqNode& x) const { 
    return (pathCost+heurEstimate) > (x.pathCost+x.heurEstimate);
    //return (pathCost) > (x.pathCost);
  }

};
