// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

#ifndef ROUTE_TREE_H
#define ROUTE_TREE_H

#include "routingNode.h"

class treeNode{
 public:
  treeNode* parent;  //parent of this tree node
  list<treeNode*> childList;  //list of children of this tree node
  list<treeNode*>::iterator* child_itr;  //pointer to childList iterator 
  routingNode* rNode;  //the routing node that this tree node corresponds to
  
  double linkDelay; //This is the delay from the root of the tree to this node
  int magicNum;

  treeNode(){
    parent=0;
    rNode=0;
    child_itr=0;
    magicNum=31111;
  }
  
  treeNode(routingNode* r){
    parent=0;
    rNode=r;
    child_itr=0;
    magicNum=31111;
  }
  
  treeNode(routingNode* r,double d){
    parent=0;
    rNode=r;
    linkDelay = d;
    child_itr=0;
    magicNum=31111;
  }

  ~treeNode(){
    if(magicNum!=31111)
      ERROR("trying to double delete this node...");
    else
      magicNum=0;
  }

  void set_parent(treeNode* t);  //set the parent node of this tree node
  void set_rNode(routingNode* r);  //set the routing node that this tree node represents
  //void set_pathCost(double p);  //set the cost of the path from the root to this tree node
  void addto_childList(treeNode* t);  //add a tree node to childList
  void create_childListItr();  //create a new childList iterator
  void remove_childListItr();  //remove childList iterator

  treeNode* get_parent();  //get parent of this tree node
  routingNode* get_rNode();  //get the routing node that this tree node represents
  //double get_pathCost();  //get the cost of the path from the root to this tree node
  treeNode* getnext_childList();  //get the next element in childList
  int get_numChildren();  //get the number of children of this tree node

  //get the iterator that points to the node in childList that holds child_node  
  list<treeNode*>::iterator get_childItr(treeNode* child_node);
  //erase the child node pointed to by ch_itr
  void eraseChild(list<treeNode*>::iterator* ch_itr);

  void removeChild(treeNode* child);

};

class routeTree{
 public:
  treeNode* root;  //the root node of this route tree
  int magicNum;

  routeTree(){
    root=0;magicNum=32111;
  }
  routeTree(treeNode* rt){
    root=rt;magicNum=32111;
  }
  ~routeTree(){
    if(magicNum!=32111)
      ERROR("trying to double delete tree here...");
    else{
      clear();
      magicNum=0;
    }
  }

  void set_root(treeNode* rt);  //set the root of this tree
  treeNode* get_root();  //get the root of this tree
  void print_info();  //print this tree and info relating to it 
  void preorderTraverse(treeNode* r,int do_what);  //traverse this tree preorder
  void postorderTraverse(treeNode* r,int do_what);  //traverse this tree postorder
  void clear();  //clean up this routing tree
  treeNode* findNode(treeNode* r,routingNode* rnode);  //find rnode in this tree and return relevant tree node
  //make this route tree a copy of the tree rooted at tree_root
  void populateWith(treeNode* tree_root);
  void populateWith(treeNode* tree_root,routingNode* sink_rnode);
  treeNode* makeCopy(treeNode* r);

  //get the root of the subtree that sinks at sink_rnode
  treeNode* get_subTree(routingNode* sink_rnode); 

  //increment the occupancies of the routing nodes that constitute this tree
  void incrementOccupancies(int how);

  //find tree node that points to routing node rnode
  treeNode* getNode(routingNode* rnode);

  // Clean the routing nodes in this tree. This is a post-BFS reinit operation.
  void clean_rNodes(treeNode* r);

  routeTree* copyPortion(treeNode* r);

  void reCalculateOccupancy(treeNode *r);

  void clearInPRTree(treeNode *r);

  void printPath(Terminal *sink);
  
  treeNode *returnTreeNode(treeNode* tNode, routingNode *rNode);
};

#endif
