// Copyright (C) 2006 Ken Eguro, Scott Hauck, Akshay Sharma
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
//
// You can contact the authors electronically at eguro at ee dot washington dot edu
// or by regular mail at 
// Box 352500, 
// University of Washington
// Seattle, WA 98195

/****************************************************************
Class definition of the edge that connects two routing nodes.
An edge between two routing nodes essentially represents a switch
between the two nodes. Each switch can be buffered or unbuffered.
For a bidirectional switch between two routing nodes, two opposite
pointing routing edges are created.

We use the Elmore delay model here. Hence, each edge has a 
resistance and input / output capacitances. Further, an edge
can also have an intrinsic delay in case the switch that it
represents delays signals passing thru it.
****************************************************************/    
#ifndef ROUTING_EDGE_H
#define ROUTING_EDGE_H

class routingEdge{
 public:
  enum e_routingEdgeType type;  //type of routing edge based on rr_graph.echo and arch.echo
  //bool buffered;
  routingNode* origNode;  //the routing node that originates this edge
  routingNode* destNode;  //the routing node that sinks this edge
  //double R,Cin,Cout;  //edges will have a resistance and input / output capacitances.
  //                    //this is because an edge basically represents a switch / buffer.
  double intrinsicDelay;  //the intrinsic delay of this edge
  //double intrinsicCost;
  int magicNum;

  routingEdge(){
    type=RE_NO_TYPE;
    origNode=destNode=0;
    //R=Cin=Cout=0.;
    intrinsicDelay=0.0;
    //intrinsicCost=0.0;
    //buffered=false;
    magicNum=20111;
  }
/*   routingEdge(enum e_routingEdgeType t,routingNode* orig,routingNode* dest,double r,double cout, */
/*               double cin, double intD, double intC){ */
/*     if(r<0. || cin<0. || cout<0. || intD<0.) */
/*       ERROR("resistance / capacitance / intrinsic delay cannot be -ve..."); */
/*     if(orig==0 || dest==0) */
/*       ERROR("source / destination routing node cannot be NULL..."); */
/*     type=t; */
/*     origNode=orig;destNode=dest; */
/*     R=r;Cin=cin;Cout=cout; */
/*     intrinsicDelay=intD; */
/*     intrinsicCost = intC; */
/*     magicNum=20111; */
/*     //buffered=false; */
/*   } */
  ~routingEdge(){
    if(magicNum!=20111)
      ERROR("trying to double delete node here...");
    else
      magicNum=0;
  }

  void set_type(enum e_routingEdgeType t);  //set type of routing edge i.e. buffered or unbuffered
  void set_origNode(routingNode* n);  //set the routing node that originates this edge
  void set_destNode(routingNode* n);  //set the routing node that sinks this edge
  //void set_R(double r);  //set the resistance of this edge
  //void set_Cin(double c);  //set the input capacitance of this edge
  //void set_Cout(double c);  //set the output capacitance of this edge
  void set_intrinsicDelay(double d);  //set the intrinsice delay of this edge

  enum e_routingEdgeType get_type();  //get type of this routing edge
  routingNode* get_origNode();  //get the routing node that originates this edge
  routingNode* get_destNode();  //get the routing node that sinks this edge
  //double get_R();  //get the resistance of this edge
  //double get_Cin();  //get the input capacitance of this edge
  //double get_Cout();  //get the output capacitance of this edge
  //double get_intrinsicDelay();  //get the intrinsic delay of this edge

  void print_info();  //print this routing edge's info 

};

#endif


