/* Ken Eguro */
/* 6/1/07 */

using namespace std;

//Classes
class c_block;
class c_wire;

enum e_blockType {INPUTPIN, OUTPUTPIN, CLB, SUBBLOCK, HOST};

class c_block{
 public:
  enum e_blockType type;
  string name;
  vector <c_wire *> inputs;
  vector <int> inputDelays;
  vector <c_wire *> outputs;
  vector <c_block *> subblocks;
  bool removed;
  bool isOnlyFF;
  bool registered;
  bool forceExplicitRegister;
  bool shadowFeedback;
  bool visited;
  bool shadowed;
  float delay;
  float deltaV;
  int RV;
  
  c_block* currSubBlock;

  c_block(enum e_blockType typeNum, string str, int numInputs, int numOutputs, float del){
    type = typeNum;
    name = str;
    inputs.resize(numInputs, 0);
    inputDelays.resize(numInputs, 0);
    outputs.resize(numOutputs, 0);
    delay = del;
    removed = false;
    isOnlyFF = false;
    registered = false;
    forceExplicitRegister = false;
    shadowFeedback = false;
    shadowed = false;
    delay = del;
    deltaV = -1.0;
    int RV = 0;
  }

  void printInfo();
  void addInput(c_wire *wirePtr, int inputNum);
  void addOutput(c_wire *wirePtr, int outputNum);

  c_block* findSubBlock(string str);
  c_block* addSubBlock(string str, float del);

  void addSubBlockInput(int inputNum, string input);
  void addSubBlockOutput(string output);

  void addSubBlockReg(string str);
};

class c_wire{
 public:
  string name;
  int nodeNum;
  c_block* sourceBlock;
  int sourceOutput;
  vector<c_block*> sinkBlockVector;
  vector<int> sinkInputVector;
  bool removed;
  float delay;
  bool hostWire;

  c_wire(string str){
    name = str;
    sourceBlock = NULL;
    sourceOutput = -1;
    removed = false;
    hostWire = false;
  }

  bool hasSource();
  bool hasSink();
  void addSink(c_block* blockPtr, int inputNum);
  void addSource(c_block* blockPtr, int outputNum);
  bool hasSink(c_block* blockPtr, int inputNum);
  void removeSink(c_block* blockPtr, int inputNum);
  void removeSource(c_block* blockPtr, int outputNum);
};

//General Types
class c_moduleDef{
 public:
  string name;
  vector <c_block *> blocks;
  vector <c_wire *> wires;

  c_block* addBlock(enum e_blockType typeNum, string str, int numInputs, int numOutputs, float delay);

  c_wire* addWire(string str);
  c_wire* findWire(string str);

  c_block* findBlock(string str);

  void initialize();

  void process(bool clbReg);

  int countOutputPins();
  int returnNumAllBlocks();

  void addHostVertex();

  void outputBlif(string outFile);
  
  void outputNetlist(char *outFile, bool clbReg);

  void printInfo();

  void countNetlist();
};

