%{
  #include "include.h"
  extern FILE *yyin;
    
  //Global Variables
  long lineCount = 1; //Lexer Line Counter
  vector <string> textVector;
  
  //Error messaging
  void yyerror(const char *str){
    printf("\n");
    fprintf(stderr,"*** Error at or near line %d: %s\n", lineCount + 1, str);
    exit(0);
  }
 
  int yywrap(){
    return 1;
  }
  
  //Need to declare as global variable since bison is not C++ aware
  c_moduleDef mDef;
  int numCLBInputs;
  int numCLBOutputs;
  int numSBInputs;
  c_block *currBlock;

  int main(int argc, char *argv[]){
    int i;
    bool clbReg = false;
    bool retime = false;
    int expectedRetimingLogicDepth = 0;
    bool pipeline = false;
    int pipelineAmt = 0;
    bool cslow = false;
    int cslowAmt = 0;
    bool postCSlow = false;
    int postCSlowAmt = 0;
     //bool detectCycles = false;
    bool detectZeroLinks = false;
    bool pipelineInputs = false;
    bool countOnly = false;
    bool blifOutput = false;
    string blifName;
    //bool shiftRegisterPacking = false;
    //int shiftRegisterPackingAmt = 1000;

    if(argc < 6)
      errorMessage("Usage: convertNetlist inputVPRNetlist outputArmadaNetlist numInputsPerCLB numOutputsPerCLB numInputsPerSubblock [+CLBREG][+retime expectedLogicDepth][+pipeline X][+CSLOW X][+BLIFOUT X][+POSTCSLOW X]");    
    for(i = 6; i < argc; i++){
      if(strcmp(argv[i], "+CLBREG") == 0){
	clbReg = true;
	cout << "Outputting explicit registers" << endl;
      }
      else if(strcmp(argv[i], "+retime") == 0){
	if(argc <= i + 1){
	  errorMessage("Retime option requires a numerical argument");
	}
	retime = true;
	cout << "Retiming netlist" << endl;

	expectedRetimingLogicDepth = atoi(argv[i + 1]);
	i ++;
	cout << "Expected post-retiming logic depth = " << expectedRetimingLogicDepth <<endl;
      }
      else if(strcmp(argv[i], "+pipeline") == 0){
	if(argc <= i + 1){
	  errorMessage("Pipeline option requires a numerical argument");
	}
	pipeline = true;
        pipelineAmt = atoi(argv[i + 1]);
	i ++;
	
	if(pipelineAmt < 0 || pipelineAmt > 40){
	  errorMessage("Pipeline amount should be between 1 and 40");
	}

	cout << "Pipelining by " << pipelineAmt << endl;
      }

      else if(strcmp(argv[i], "+pipelineInputs") == 0){
	cout << "Pipelining inputs once" << endl;
      	pipelineInputs = true;
      }

      else if(strcmp(argv[i], "+CSLOW") == 0){
	if(argc <= i + 1){
	  errorMessage("C-Slow option requires a numerical argument");
	}
	cslow = true;
	cslowAmt = atoi(argv[i + 1]);
	i ++;

	if(cslowAmt < 1 || cslowAmt > 20){
	  errorMessage("C-slow amount should be between 2 and 20");
	}

	cout << "C-slowing by " << cslowAmt << endl;
      }

      else if(strcmp(argv[i], "+POSTCSLOW") == 0){
	if(argc <= i + 1){
	  errorMessage("C-Slow option requires a numerical argument");
	}
	postCSlow = true;
	postCSlowAmt = atoi(argv[i + 1]);
	i ++;

	if(postCSlowAmt < 1 || postCSlowAmt > 20){
	  errorMessage("C-slow amount should be between 2 and 20");
	}

	cout << "Post retiming C-slowing by " << postCSlowAmt << endl;
      }

      //else if(strcmp(argv[i], "+detectCycles") == 0){
      //	detectCycles = true;
      //}

      else if(strcmp(argv[i], "+detectZeroLinks") == 0){
      	detectZeroLinks = true;
      }

      else if(strcmp(argv[i], "+countOnly") == 0){
	countOnly = true;
      }

      else if(strcmp(argv[i], "+BLIFOUT") == 0){
	if(argc <= i + 1){
	  errorMessage("+BLIFOUT option requires an output filename");
	}
	blifOutput = true;
	blifName.assign(argv[i + 1]);
	i ++;
      }
/*       else if(strcmp(argv[i], "+ShiftRegPack") == 0){ */
/* 	if(argc <= i + 1){ */
/* 	  errorMessage("+ShiftRegPack option requires an integer argument"); */
/* 	} */
/* 	shiftRegisterPackingAmt = atoi(argv[i + 1]); */
/* 	i ++; */

/* 	if(shiftRegisterPackingAmt < 0 || shiftRegisterPackingAmt > 20){ */
/* 	  errorMessage("Shift register packing amount should be between 1 and 20"); */
/* 	} */
	
/* 	shiftRegisterPacking = true; */
/* 	cout << "Packing " << regPackingAmt << " or more registers into shift register chains" << endl; */
/*       } */

      else{
	errorMessage("Usage: convertNetlist inputVPRNetlist outputArmadaNetlist numInputsPerCLB numOutputsPerCLB numInputsPerSubblock [+CLBREG][+retime][+pipeline int][+CSLOW int][+BLIFOUT filename][+POSTCSLOW X]");   
      }
    }

    //If we are doing retiming, pipelining, or c-slowing we should not have the +CLBREG option on too.
    //if((retime || pipeline || cslow || shiftRegisterPacking) && clbReg){
    if((retime || pipeline || cslow || postCSlow) && clbReg){
      errorMessage("Retiming, pipelining and c-slowing cannot also have the +CLBREG option.\nInstead, output to .blif, repack and then use +CLBREG option.");
    }

    numCLBInputs = atoi(argv[3]);
    numCLBOutputs = atoi(argv[4]);
    numSBInputs = atoi(argv[5]);

    printf("Architecture has %d inputs and %d outputs per CLB, %d inputs per subblock\n", numCLBInputs, numCLBOutputs, numSBInputs);
    
    yyin = fopen(argv[1], "r");
    cout << "Input netlist " << argv[1] << endl;

    if(!yyin){
      errorMessage("Unable to open file");
    }
    
    yyparse();
    
    mDef.process(clbReg);

    //mDef.printInfo();
    findShadowedBlocks();

    if(cslow){
      cslowNetlist(cslowAmt);
      //mDef.printInfo();
    }
    
    if(pipeline){
      pipelineNetlist(pipelineAmt);
      //mDef.printInfo();
    }

    if(retime){
      retimeNetlist(expectedRetimingLogicDepth);
      //mDef.printInfo();
    }
    
    if(pipelineInputs){
      pipelineInputsOnce();
    }

    if(postCSlow){
      postCslowNetlist(postCSlowAmt);
    }

    if(detectZeroLinks){
      detectZeroLatencyLinks();
    }

    if(countOnly){
	mDef.countNetlist();
    }
    else{
      if(blifOutput){
	mDef.outputBlif(blifName);
      }
      
      mDef.outputNetlist(argv[2], clbReg);
    }

    cout << endl << endl;
      
    return 0;
   }
  
%}

%union {
  char str[MAXCHARSTARSIZE];
}

//Tokens
%token TOKGLOBAL
%token TOKINPUT
%token TOKOUTPUT
%token TOKCLB
%token TOKPINLIST
%token TOKSUBBLOCK
%token TOKENDL
%token <str> TOKTEXT

%%
commands: /* empty */
	  | commands command;

command:
global | input_pin | output_pin | clb | pinlist | subblock | empty;

texts: /* empty */
| texts TOKTEXT {
  textVector.push_back($2);
};

global:
TOKGLOBAL texts TOKENDL{
  if(textVector.size() == 1){
    if(!mDef.addWire(textVector[0]))
      exit(-1);
    textVector.clear();
  }
  else{
    cout << "Problem with wire at line "  << lineCount << endl;
    exit(-1);
  }
};

input_pin:
TOKINPUT texts TOKENDL{
  if(textVector.size() == 1){
    //We create input pins with an input in case we are doing retiming
    currBlock = mDef.addBlock(INPUTPIN, textVector[0], 1, 1, INPUTPINDELAY);
    if(!currBlock)
      exit(-1);
    textVector.clear();
  }
  else{
    cout << "Problem with input pin at line "  << lineCount << endl;
    exit(-1);
  }
};

output_pin:
TOKOUTPUT texts TOKENDL{
  if(textVector.size() == 1){
    //We create output pins with an output in case we are doing retiming
    currBlock = mDef.addBlock(OUTPUTPIN, textVector[0], 1, 1, OUTPUTPINDELAY);
    if(!currBlock)
      exit(-1);
    textVector.clear();
  }
  else{
    cout << "Problem with output pin at line "  << lineCount << endl;
    exit(-1);
  }
};

clb:
TOKCLB texts TOKENDL{
  if(textVector.size() == 1){
    currBlock = mDef.addBlock(CLB, textVector[0], numCLBInputs + 1, numCLBOutputs, 0.0);
    if(!currBlock)
      exit(-1);
    textVector.clear();
  }
  else{
    cout << "Problem with CLB at line "  << lineCount << endl;
    exit(-1);
  }
};

pinlist:
TOKPINLIST texts TOKENDL{
  int i;
  c_wire *wirePtr;
  
  if(currBlock->type == INPUTPIN){
    if(textVector.size() == 1){
      wirePtr = mDef.addWire(textVector[0]);
      currBlock->addOutput(wirePtr, 0);
    }
    else{
      cout << "Problem with input pin list at line "  << lineCount << endl;
      exit(-1);
    }
  }
  
  else if(currBlock->type == OUTPUTPIN){
    if(textVector.size() == 1){
      wirePtr = mDef.addWire(textVector[0]);
      currBlock->addInput(wirePtr, 0);
    }
    else{
      cout << "Problem with output pin list at line "  << lineCount << endl;
      exit(-1);
    }
  }
  
  else if(currBlock->type == CLB){
    if(textVector.size() == numCLBInputs + numCLBOutputs + 1){
      for(i = 0; i < numCLBInputs; i++){
	wirePtr = mDef.addWire(textVector[i]);
	currBlock->addInput(wirePtr, i);
      } 
      for(i = 0; i < numCLBOutputs; i++){
	wirePtr = mDef.addWire(textVector[i + numCLBInputs]);
	currBlock->addOutput(wirePtr, i);
      }
      wirePtr = mDef.addWire(textVector[numCLBInputs + numCLBOutputs]);
      currBlock->addInput(wirePtr, numCLBInputs);
    }
    else{
      cout << "Problem with CLB pin list at line "  << lineCount << endl;
      exit(-1);
    }
  }
  
  else{
    cout << "Unknown block type at line "  << lineCount << endl;
    exit(-1);
  }
  
  textVector.clear();
};

subblock:
TOKSUBBLOCK texts TOKENDL{
  c_block *subBlockPtr;
  int i;
  
  if(currBlock->type == CLB){
    if(textVector.size() == numSBInputs + 3){
      currBlock->currSubBlock = currBlock->addSubBlock(textVector[0], LUTDELAY);

      for(i = 0; i < numSBInputs; i++){	
	currBlock->addSubBlockInput(i, textVector[i + 1]);
      } 
      currBlock->addSubBlockOutput(textVector[numSBInputs + 1]);
      currBlock->addSubBlockInput(numSBInputs, textVector[numSBInputs + 2]);
    }
    else{
      cout << "Problem with subblock pinlist at line "  << lineCount << endl;
      exit(-1);
    }
  }
  else{
    cout << "Block does not have subblock at line "  << lineCount << endl;
    exit(-1);
  }

  textVector.clear();
};

empty:
TOKENDL{
  
};










