#include "include.h"
extern c_moduleDef mDef;

//To c-slow the netlist, just multiply the latency on all links by N
void cslowNetlist(int cslowAmt){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  vector <c_wire *>::iterator inputIter;
  vector<int>::iterator inputLatencyIter;
  
  cout << "C-slowing netlist by " << cslowAmt << endl;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }
    
    if((*blockIter)->type == OUTPUTPIN){
      if((*blockIter)->shadowed){
	cout << "WARNING::Output pin " << (*blockIter)->name << " is shadowed, so we will C-slowed it, but BE CAREFUL!" << endl;
      }
      if((*blockIter)->inputs.size() != 1){
	cout << "Error::Output pin " << (*blockIter)->name << " should only have one input!" <<endl;
	exit(-1);
      }
      (*blockIter)->inputDelays.at(0) *= cslowAmt;
    }

    else if((*blockIter)->type == INPUTPIN){
      if((*blockIter)->inputs.size() != 1){
	cout << "Error: Input pin " << (*blockIter)->name << " should only have one input!" <<endl;
	exit(-1);
      }
      if((*blockIter)->inputDelays.at(0) != 0){
	cout << "Error: Input pin " << (*blockIter)->name << " should not have any latency on the input!" <<endl;
	exit(-1);
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if((*subBlockIter)->shadowed){
	  cout << "WARNING::Subblock " << (*subBlockIter)->name << " is shadowed, so we will C-slowed it, but BE CAREFUL!" << endl;
	}
	for(inputIter = (*subBlockIter)->inputs.begin(), inputLatencyIter = (*subBlockIter)->inputDelays.begin(); 
	    inputIter != (*subBlockIter)->inputs.end() && inputLatencyIter != (*subBlockIter)->inputDelays.end(); 
	    inputIter++, inputLatencyIter++){
	  if(!*inputIter){
	    //The latency better be zero
	    if(*inputLatencyIter != 0){
	      cout << "Error: Subblock " << (*subBlockIter)->name << " should not have latency on an open pin!" << endl;
	      exit(-1);
	    }
	  }
	  else{
	    (*inputLatencyIter) *= cslowAmt;
	  }
	}
      }
    }
  }
}

//To c-slow the netlist, just multiply the latency on all links by N
// However, since we have already retimed the netlist (potentially), don't
// add and more delay to input pins.
void postCslowNetlist(int cslowAmt){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  vector <c_wire *>::iterator inputIter;
  vector<int>::iterator inputLatencyIter;
  
  cout << "C-slowing netlist by " << cslowAmt << endl;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }
    
    if((*blockIter)->type == OUTPUTPIN){
      if((*blockIter)->shadowed){
	cout << "WARNING::Output pin " << (*blockIter)->name << " is shadowed, so we will post-retiming C-slowed it, but BE CAREFUL!" << endl;
      }
      if((*blockIter)->inputs.size() != 1){
	cout << "Error::Output pin " << (*blockIter)->name << " should only have one input!" <<endl;
	exit(-1);
      }
      (*blockIter)->inputDelays.at(0) *= cslowAmt;
    }

    else if((*blockIter)->type == INPUTPIN){
      if((*blockIter)->inputs.size() != 1){
	cout << "Error: Input pin " << (*blockIter)->name << " should only have one input!" <<endl;
	exit(-1);
      }
      //if((*blockIter)->inputDelays.at(0) != 0){
      //cout << "Error: Input pin " << (*blockIter)->name << " should not have any latency on the input!" <<endl;
      //	exit(-1);
      //}
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if((*subBlockIter)->shadowed){
	  cout << "WARNING::Subblock " << (*subBlockIter)->name << " is shadowed, so we will post-retiming C-slowed it, but BE CAREFUL!" << endl;
	}
	for(inputIter = (*subBlockIter)->inputs.begin(), inputLatencyIter = (*subBlockIter)->inputDelays.begin(); 
	    inputIter != (*subBlockIter)->inputs.end() && inputLatencyIter != (*subBlockIter)->inputDelays.end(); 
	    inputIter++, inputLatencyIter++){
	  if(!*inputIter){
	    //The latency better be zero
	    if(*inputLatencyIter != 0){
	      cout << "Error: Subblock " << (*subBlockIter)->name << " should not have latency on an open pin!" << endl;
	    }
	  }
	  else{
	    (*inputLatencyIter) *= cslowAmt;
	  }
	}
      }
    }
  }
}

//To pipeline the netlist, just add latency to the output pins
void pipelineNetlist(int pipelineAmt){
  vector <c_block *>::iterator blockIter;
  c_block *blockPtr;

  //See if there is "pclk" input pin.
  if(pipelineAmt != 0){
    blockPtr = mDef.findBlock("pclk");
    if(!blockPtr){
      //Make one, since we are pipelining a previously non-sequential circuit
      mDef.addBlock(INPUTPIN, "pclk", 1, 1, INPUTPINDELAY);
    }
  }

  cout << "Pipelining netlist with " << pipelineAmt << " registers on output pins" << endl;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }
    
    if((*blockIter)->type == OUTPUTPIN){
      if((*blockIter)->shadowed){
	cout << "WARNING::Output pin " << (*blockIter)->name << " is shadowed, so we will pipeline it, but BE CAREFUL!" << endl;
      }

      if((*blockIter)->inputs.size() != 1){
	cout << "Error::Output pin " << (*blockIter)->name << " should only have one input!" <<endl;
	exit(-1);
      }
      (*blockIter)->inputDelays.at(0) += pipelineAmt;
    }
  }
}

//Add one latency to all primary inputs
void pipelineInputsOnce(){
  vector <c_block *>::iterator blockIter;
  c_block *blockPtr;
  vector <c_block *>::iterator sinkIter;
  vector<int>::iterator sinkIndexIter;

  //See if there is "pclk" input pin.
  blockPtr = mDef.findBlock("pclk");
  if(!blockPtr){
    //Make one, since we are pipelining a previously non-sequential circuit
    mDef.addBlock(INPUTPIN, "pclk", 1, 1, INPUTPINDELAY);
  }

  cout << "Pipelining netlist with 1 register on each input pin" << endl;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }
    
    if((*blockIter)->type == INPUTPIN){
      for(vector <c_wire *>::iterator outputWireIter = (*blockIter)->outputs.begin();
	  outputWireIter != (*blockIter)->outputs.end(); outputWireIter++){
	if(*outputWireIter){
	  for(sinkIter = (*outputWireIter)->sinkBlockVector.begin(), sinkIndexIter = (*outputWireIter)->sinkInputVector.begin(); 
	      sinkIter != (*outputWireIter)->sinkBlockVector.end() && sinkIndexIter != (*outputWireIter)->sinkInputVector.end(); 
	      sinkIter++, sinkIndexIter++){
	    if((*sinkIter)->inputs.at(*sinkIndexIter) != *outputWireIter){
	      cout << "Error: Sink is not connected to correct wire!" << endl;
	      exit(-1);
	    }
	    
	    (*sinkIter)->inputDelays.at(*sinkIndexIter)++;
	  }
	}
      }
    }
  }
}

//Check to see if any inputs have zero latency
void detectZeroLatencyLinks(){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  c_block *blockPtr;
  vector <c_block *>::iterator sinkIter;
  vector<int>::iterator sinkIndexIter;
  vector <c_wire *>::iterator inputIter;
  vector<int>::iterator inputLatencyIter;

  //See if there is "pclk" input pin.
  blockPtr = mDef.findBlock("pclk");
  if(!blockPtr){
    cout << "Error: there is no clock signal!" << endl;
    exit (-1);
  }

  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }
    
    if((*blockIter)->type != CLB){
      for(inputIter = (*blockIter)->inputs.begin(), inputLatencyIter = (*blockIter)->inputDelays.begin(); 
	  inputIter != (*blockIter)->inputs.end() && inputLatencyIter != (*blockIter)->inputDelays.end(); 
	  inputIter++, inputLatencyIter++){
	if(!*inputIter){
	  //The latency better be zero
	  if(*inputLatencyIter != 0){
	    cout << "Error: Block " << (*blockIter)->name << " should not have latency on an open pin!" << endl;
	    exit(-1);
	  }
	}
	else{
	  if(*inputLatencyIter == 0){
	    cout << "Block " << (*blockIter)->name << " has a zero latency input" << endl;
	  }
	}
      }
    }
    else{
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	for(inputIter = (*subBlockIter)->inputs.begin(), inputLatencyIter = (*subBlockIter)->inputDelays.begin(); 
	    inputIter != (*subBlockIter)->inputs.end() && inputLatencyIter != (*subBlockIter)->inputDelays.end(); 
	    inputIter++, inputLatencyIter++){
	  if(!*inputIter){
	    //The latency better be zero
	    if(*inputLatencyIter != 0){
	      cout << "Error: Block " << (*subBlockIter)->name << " should not have latency on an open pin!" << endl;
	      exit(-1);
	    }
	  }
	  else{
	    if(*inputLatencyIter == 0){
	      cout << "Block " << (*subBlockIter)->name << " has a zero latency input" << endl;
	    }
	  }
	}
      }
    }
  }
}

//Take the netlist and retime it.
//We only need to be careful about forced explicit FFs - a block that connects to itself but does not
// have any latency on the connection because it is implicit in the block itself.
void retimeNetlist(int expectedRetimingLogicDepth){
  float bestCPD;
  float failedCPD = 0.0;
  float attemptCPD;
  bool success = true;

  //Since we did not use the +CLBREG option, the existing netlist should be almost Leiserson/Saxe graph.
  //Add a host vertex, with latency (weight) 1 edges to and from it. Don't add a link to the clock if there is one.
  mDef.addHostVertex();

  //Mark possible troublemakers
  //The first we found was blocks that are in a cycle, but are not connected to 
  // the input pins.
  //First, find all shadowed blocks
  //findShadowedBlocks();

  //Determine the current critical path delay
  bestCPD = calculateCPD();
  cout << "Initial CPD = " << bestCPD << endl;
  //mDef.printInfo();

  attemptCPD = expectedRetimingLogicDepth;

  //Go until the we find the minimum achievable critical path delay
  //Try to retime the netlist to the attempted CPD
  cout << "Attempt to retime to " << attemptCPD << endl;
  success = retime(attemptCPD);
  //mDef.printInfo();
  
  if(success){
    cout << "Succeeded!" << endl;
  }
  else{
    cout << "Failed!" << endl;
  }

  //See if we current graph is actually the result of retiming the netlist at the current best CPD.
  //If not retime the netlist once more
  if(!success){
    cout << "Error: We were not able to retime to that logic depth!" << endl;
    exit(-1);
    //mDef.printInfo();
  }

  //See if the host node needs to be cleaned up (we need exactly one register to/from the host vertex like
  // we started with.
  cleanupHost();
  //cout << "After host cleanup" << endl;
  //mDef.printInfo();

  //Forced explicit registers may have had registers pile up on their output.
  //Since these are constants, this is unnecessary - remove them.
  //Nix this, we don't add delay onto forced explict registers in the first place
  //cleanupForcedExplicitRegisters();
}

//Shadowed blocks are those not connected upstream to the input pins.
void findShadowedBlocks(){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  vector <c_wire *>::iterator wireIter;
  vector<c_block*>::iterator sinkIter;
  list <c_block *> queue;
  c_block *currBlock;
  c_block *neighborBlock;  


  //**********************THIS IS GOING FROM OUTPUTS TO INPUTS!************************
 //Clear all visited variables and put all output pins into queue
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN){
      (*blockIter)->visited = false;
    }
    else if((*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      (*blockIter)->visited = false;
      queue.push_back(*blockIter);
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	(*subBlockIter)->visited = false;
      }
    }
  }  

  //Do until we run out of blocks
  while(!queue.empty()){
    //Pull off a block &  mark it visited
    currBlock = queue.front();
    currBlock->visited = true;
    queue.pop_front();

    //Go through all the input pins of this block
    for(wireIter = currBlock->inputs.begin(); wireIter != currBlock->inputs.end(); wireIter++){
      if(*wireIter){
	//Go through all of the sources of this input pin
	if(!(*wireIter)->sourceBlock->visited){
	  queue.push_back((*wireIter)->sourceBlock);
	}
      }
    }
  }
  
  //Go through all of the blocks and warn the user about all unvisited blocks
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN || (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      if(!(*blockIter)->visited){
	cout << "WARNING: block " << (*blockIter)->name << " is not connected to the output pins!" << endl;
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if(!(*subBlockIter)->visited){
	  cout << "WARNING: block " << (*subBlockIter)->name << " is not connected to the output pins!" << endl;
	}
      }
    }
  }   
 
  //********Proceed with the normal shadow finding code

  //Clear all visited variables and put all input pins into queue
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN){
      (*blockIter)->visited = false;
      queue.push_back(*blockIter);
    }
    else if((*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      (*blockIter)->visited = false;
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	(*subBlockIter)->visited = false;
      }
    }
  }  

  //Do until we run out of blocks
  while(!queue.empty()){
    //Pull off a block &  mark it visited
    currBlock = queue.front();
    currBlock->visited = true;
    queue.pop_front();

    //Go through all the output pins of this block
    for(wireIter = currBlock->outputs.begin(); wireIter != currBlock->outputs.end(); wireIter++){
      if(*wireIter){
	//Go through all of the sinks of this output pin
	for(sinkIter = (*wireIter)->sinkBlockVector.begin(); sinkIter != (*wireIter)->sinkBlockVector.end(); sinkIter++){
	  //If this sink has not been visited, put it into the queue
	  if(!(*sinkIter)->visited){
	    queue.push_back(*sinkIter);
	  }
	}
      }
    }
  }
  
  //Go through all of the blocks and warn the user about all unvisited blocks
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN || (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      if(!(*blockIter)->visited){
	(*blockIter)->shadowed = true;
	cout << "WARNING: block " << (*blockIter)->name << " is not connected to the input pins! BE CAREFUL!" <<endl;
	//and will not be retimed or pipelined!" << endl;
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if(!(*subBlockIter)->visited){
	  (*subBlockIter)->shadowed = true;
	  cout << "WARNING: block " << (*subBlockIter)->name << " is not connected to the input pins! BE CAREFUL!" <<endl;
	  //and will not be retimed!" << endl;

	  //Check to see if any of the inputs are connected to the subblock's output
	  //Notice that this does not find ALL problems we might have with shadowed loops, only
	  // those that have a cycle of 1 LUT.
	  //If this becomes a problem, we will have to figure out a way to detect
	  // cycles of greater length
	  //Go through all the output pins of this block
	  for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
	    if(*wireIter){
	      //Go through all of the sinks of this output pin
	      for(sinkIter = (*wireIter)->sinkBlockVector.begin(); sinkIter != (*wireIter)->sinkBlockVector.end(); sinkIter++){
		//If this sink is the same as the block that originated it, mark the block
		if(*sinkIter == *subBlockIter){
		  (*subBlockIter)->shadowFeedback = true;
		  cout << "WARNING: block is also part of a shadowed feedback cycle - this may lead to MORE trouble" << endl;
		}
	      }
	    }
	  }	  
	}
      }
    }
  }   
  
}

//Determine the deltaV for all nodes and return the maximum
float calculateCPD(){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  list <c_block *> queue;
  c_block *currBlock;
  c_block *neighborBlock;
  vector <c_wire *>::iterator wireIter;
  vector<c_block*>::iterator sinkIter;
  vector<int>::iterator sinkInputIter;
  float currPathDelay;
  float maxPathDelay;
  bool allRegisteredInputs;

  //Set the path delay for all nodes to it's own self delay and put into the queue;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->removed){
      cout << "Error: Block " << (*blockIter)->name << " should have been removed!" << endl;
      exit(-1);
    }

    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      (*blockIter)->deltaV = (*blockIter)->delay;
      //cout << "Block " << (*blockIter)->name << " inherant delay = " <<  (*blockIter)->deltaV << endl;
      queue.push_back(*blockIter);
      //cout << "Pushed block " << (*blockIter)->name <<endl;
    }

    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){

/* 	if((*subBlockIter)->name.compare("n_n13895") == 0){ */
/* 	  cout << "We found n_n13895!" << endl; */
/* 	} */
	if((*subBlockIter)->removed){
	  cout << "Error: Block " << (*subBlockIter)->name << " should have been removed!" << endl;
	  exit(-1);
	}
	
	(*subBlockIter)->deltaV = (*subBlockIter)->delay;
	//(*subBlockIter)->deltaV = 1.0;

	//cout << "Sub block " << (*subBlockIter)->name << " inherant delay = " <<  (*subBlockIter)->deltaV << endl;
	
	//Don't add this subblock in if it is a forced explict register
	if(!(*subBlockIter)->forceExplicitRegister){
	  queue.push_back(*subBlockIter);
	  //cout << "Pushed subblock " << (*subBlockIter)->name <<endl;
	}
	else{
	  
	}
      }
    }
  }

  while(!queue.empty()){
    currBlock = queue.front();
    currPathDelay = currBlock->deltaV;

    if(currBlock->removed){
      cout << "Error: Encountered removed block!" << endl;
      exit(-1);
    }

    queue.pop_front();

    //cout << "Popped block " << currBlock->name << endl;
    
    //Go through all the output pins of this block
    for(wireIter = currBlock->outputs.begin(); wireIter != currBlock->outputs.end(); wireIter++){
      if(*wireIter){
	//Go through all of the sinks of this output pin
	for(sinkIter = (*wireIter)->sinkBlockVector.begin(), sinkInputIter = (*wireIter)->sinkInputVector.begin();
	    sinkIter != (*wireIter)->sinkBlockVector.end() && sinkInputIter != (*wireIter)->sinkInputVector.end();
	    sinkIter++, sinkInputIter++){
	  //Make sure this sink is actually connected to the wire we think it is
	  if((*sinkIter)->inputs.at(*sinkInputIter) != *wireIter){
	    cout << "Error: Sink is not connected to correct wire!" << endl;
	    exit(-1);
	  }
	 
/* 	  if((*sinkIter)->name.compare("n_n13895") == 0){ */
/* 	    cout << "We found n_n13895! deltaV = " << (*sinkIter)->deltaV << endl; */
/* 	  } */
/* 	  if((*sinkIter)->name.compare("[33256]") == 0){ */
/* 	    cout << "We found [33256]! deltaV = " << (*sinkIter)->deltaV << endl; */
/* 	  } */
/* 	  if((*sinkIter)->name.compare("ni31") == 0){ */
/* 	    cout << "We found ni31! deltaV = " << (*sinkIter)->deltaV << endl; */
/* 	  } */

	  //See if this sink is connected with a unregistered wire (latency == 0)
	  if((*sinkIter)->inputDelays.at(*sinkInputIter) == 0){
	    //See if the current path delay + neighbor's delay is greater than the neighbor's path delay
	    if(currPathDelay + (*sinkIter)->delay > (*sinkIter)->deltaV){
	      //If it is, set a new path delay for the neighbor and put it on the queue
	      (*sinkIter)->deltaV = currPathDelay + (*sinkIter)->delay;
	      queue.push_back(*sinkIter);
	      
	      //cout << "Pushed block " << (*sinkIter)->name << endl;
	    }
	  }
	}
      }
    }
  }

  //Find the maximum deltaV
  maxPathDelay = 0.0;
  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      //cout << "Block " << (*blockIter)->name << " deltaV = " << (*blockIter)->deltaV <<endl;
/*       if((*blockIter)->deltaV == 0.0 ){ */
/* 	cout << "Why is this block " << (*blockIter)->name << " still at delay 0.0?" << endl; */
/*       } */

      if((*blockIter)->deltaV > maxPathDelay){
	maxPathDelay = (*blockIter)->deltaV;
	//cout << "MaxPathDelay set by block " << (*blockIter)->name << " at delay " << (*blockIter)->deltaV <<endl;
      }
      else if((*blockIter)->deltaV == maxPathDelay && maxPathDelay != 0.0){
	//cout << "MaxPathDelay also at block " << (*blockIter)->name << " at delay " << (*blockIter)->deltaV <<endl;
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	//cout << "Sub block " << (*subBlockIter)->name << " deltaV = " << (*subBlockIter)->deltaV <<endl;
	if((*subBlockIter)->deltaV > maxPathDelay){
	  maxPathDelay = (*subBlockIter)->deltaV;
	  //cout << "MaxPathDelay set by block " << (*subBlockIter)->name << " at delay " << (*subBlockIter)->deltaV <<endl;
	}
	else if((*subBlockIter)->deltaV == maxPathDelay && maxPathDelay != 0.0){
	  //cout << "MaxPathDelay also at block " << (*subBlockIter)->name << " at delay " << (*subBlockIter)->deltaV <<endl;
	}
      }
    }
  }
  return maxPathDelay;
}

//Attempt to retime this netlist to the given CPD
bool retime(float attemptCPD){
  c_block *currBlock;
  int iterNum = 0;
  int numBlocks = mDef.returnNumAllBlocks();
  bool success = false;
  bool movedRegisters;

  //Set the delay of the host vertex to the attempted critical path delay
  currBlock = mDef.findBlock("HOST");
  currBlock->delay = attemptCPD;

  //Repeat this # nodes times (we'll go 2x that for safety)
  while(iterNum < numBlocks + 5 && !success){
    //cout << "Iteration #" << iterNum << endl;
    //A) Determine the delta V for all nodes
    calculateCPD();    
    
    //B) Set the R(v) for all nodes
    movedRegisters = setRV(attemptCPD);
    
    //C) Check to see if there are any R(v) values greater than 0
    if(movedRegisters){
      //D) If there are some, update the graph to reflect the moved registers
      updateEdges();
    }
    else{
      //E) If there are not, we acheived the attempted CPD
      success = true;
    }
    iterNum++;
  }

  if(!success){
    //printOverCriticalBlocks(attemptCPD);
  }

  return success;
}

//Go through all of the blocks and set their R(v) value to either 0 or 1
//If the deltaV is larger than the attemptCPD, set it to 1, if not, set it to 0
bool setRV(float attemptCPD){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  bool movedRegisters = false;

  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      if((*blockIter)->deltaV > attemptCPD){
	(*blockIter)->RV = 1;
	movedRegisters = true;
      }
      else{
	(*blockIter)->RV = 0;
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if((*subBlockIter)->deltaV > attemptCPD){
	  (*subBlockIter)->RV = 1;
	  movedRegisters = true;
	}
	else{
	  (*subBlockIter)->RV = 0;
	}
      }
    } 
  }
  return movedRegisters;
}

//Based on the R(v) values, update the edge weights for all of the nodes
void updateEdges(){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  vector <c_wire *>::iterator wireIter;
  vector<c_block*>::iterator sinkIter;
  vector<int>::iterator sinkInputIter;

  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      //If RV is 1, then we have to remove one register from each of the outputs and put it on the inputs
      if((*blockIter)->RV == 1){
	//Take off a register on the output
	//cout << "Pulling registers off of output of block " << (*blockIter)->name << endl;
	for(wireIter = (*blockIter)->outputs.begin(); wireIter != (*blockIter)->outputs.end(); wireIter++){
	  if(*wireIter){
	    for(sinkIter = (*wireIter)->sinkBlockVector.begin(), sinkInputIter = (*wireIter)->sinkInputVector.begin();
		sinkIter != (*wireIter)->sinkBlockVector.end() && sinkInputIter != (*wireIter)->sinkInputVector.end();
		sinkIter++, sinkInputIter++){
	      (*sinkIter)->inputDelays.at(*sinkInputIter)--;
	    }
	  }
	}
	//Put a register on the input
	for(wireIter = (*blockIter)->inputs.begin(), sinkInputIter = (*blockIter)->inputDelays.begin(); 
	    wireIter != (*blockIter)->inputs.end() && sinkInputIter != (*blockIter)->inputDelays.end(); 
	    wireIter++, sinkInputIter++){
	  if(*wireIter){
	    //If the input is a forced explicit register or 
	    // this block itself is the beginning of a shadowed cycle, don't add a register.
/* 	    if(!(*wireIter)->sourceBlock->forceExplicitRegister && !(*wireIter)->sourceBlock->shadowFeedback){ */
/* 	      cout << "Putting register on output of block " << (*wireIter)->sourceBlock->name << " to input of block "  */
/* 	      << (*blockIter)->name << endl; */
/* 	     (*sinkInputIter)++; */
/* 	    } */
	    if(!(*wireIter)->sourceBlock->forceExplicitRegister && !(*blockIter)->shadowFeedback){
	      //cout << "Putting register on output of block " << (*wireIter)->sourceBlock->name << " to input of block " 
	      //<< (*blockIter)->name << endl;
	     (*sinkInputIter)++;
	    }
	    else{
	      //cout << "Warning: Block " << (*wireIter)->sourceBlock->name << " is a shadowed feedback register or a forced explicit register so we can't retime this" << endl;
	    }
	  }
	}
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	//If RV is 1, then we have to remove one register from each of the outputs and put it on the inputs
	if((*subBlockIter)->RV == 1){
	  //Take off a register on the output
	  //cout << "Pulling registers off of output of subblock " << (*subBlockIter)->name << endl;
	  for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
	    if(*wireIter){
	      for(sinkIter = (*wireIter)->sinkBlockVector.begin(), sinkInputIter = (*wireIter)->sinkInputVector.begin();
		  sinkIter != (*wireIter)->sinkBlockVector.end() && sinkInputIter != (*wireIter)->sinkInputVector.end();
		  sinkIter++, sinkInputIter++){
		(*sinkIter)->inputDelays.at(*sinkInputIter)--;
	      }
	    }
	  }
	  
	  //Put a register on the input
	  //cout << "Putting register on input of block " << (*subBlockIter)->name << endl;
	  for(wireIter = (*subBlockIter)->inputs.begin(), sinkInputIter = (*subBlockIter)->inputDelays.begin(); 
	      wireIter != (*subBlockIter)->inputs.end() && sinkInputIter != (*subBlockIter)->inputDelays.end(); 
	      wireIter++, sinkInputIter++){
	    if(*wireIter){
	      //If the input is a forced explicit register or the end of a shadowed cycle, don't add a register.
/* 	      if(!(*wireIter)->sourceBlock->forceExplicitRegister && !(*wireIter)->sourceBlock->shadowFeedback){ */
/* 		//cout << "Putting register on output of block " << (*wireIter)->sourceBlock->name << " to input of block "  */
/* 		//<< (*subBlockIter)->name << endl; */
/* 		(*sinkInputIter)++;	       */
/* 	      } */
	      if(!(*wireIter)->sourceBlock->forceExplicitRegister && !(*blockIter)->shadowFeedback){
		//cout << "Putting register on output of block " << (*wireIter)->sourceBlock->name << " to input of block " 
		//<< (*subBlockIter)->name << endl;
		(*sinkInputIter)++;	      
	      }
	      else{
		//cout << "Warning: Block " << (*wireIter)->sourceBlock->name << " is a shadowed feedback register or a forced explict register so we can't retime this" << endl;
	      }

	      if(*sinkInputIter > 100){
		cout << "Probable shadowed cycle through wire " <<(*wireIter)->name << " - handle by hand!" << endl;
		exit(-1);
	      }
	    }
	  }
	}
      }
    }
  }

  //Go through all of the blocks and figure out if there are any negative edge weights.
  //If there are, we have a unpipelined cycle in this netlist
    for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      for(wireIter = (*blockIter)->inputs.begin(), sinkInputIter = (*blockIter)->inputDelays.begin();
	  wireIter != (*blockIter)->inputs.end() && sinkInputIter != (*blockIter)->inputDelays.end();
	  wireIter++, sinkInputIter++){
	if(*wireIter){
	  if(*sinkIter < 0){
	    cout << "Error: I think we have a unpipelined cycle along block " << (*blockIter)->name << endl;
	    exit(-1);
	  }
	}
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	for(wireIter = (*subBlockIter)->inputs.begin(), sinkInputIter = (*subBlockIter)->inputDelays.begin(); 
	    wireIter != (*subBlockIter)->inputs.end() && sinkInputIter != (*subBlockIter)->inputDelays.end(); 
	    wireIter++, sinkInputIter++){
	  if(*wireIter){
	    if(*sinkIter < 0){
	      cout << "Error: I think we have a unpipelined cycle along block " << (*subBlockIter)->name << endl;
	      exit(-1);
	    }
	  }
	}
      }
    }
  }
}

//See if the host node needs to be cleaned up (we need exactly one register to/from the host vertex like
// we started with.
void cleanupHost(){
  c_block *hostBlock;
  c_wire *wirePtr;
  vector <c_wire *>::iterator wireIter;
  vector<c_block*>::iterator sinkIter;
  vector<int>::iterator sinkInputIter;
  vector<c_block*>::iterator sinkIter2;
  vector<int>::iterator sinkInputIter2;

  hostBlock = mDef.findBlock("HOST");

  //Go through each of the inputs to the host (the output pins)
  // and make sure there is exactly one latency on each edge.
  for(wireIter = hostBlock->inputs.begin(), sinkInputIter = hostBlock->inputDelays.begin(); 
      wireIter != hostBlock->inputs.end() && sinkInputIter != hostBlock->inputDelays.end(); 
      wireIter++, sinkInputIter++){
    if(!*wireIter){
      cout << "Error: Host block is missing a connection!" << endl;
      exit(-1);
    }
    else{
      if(*sinkInputIter > 1){
	//We have more registers than we want on this input,
	// so push them upstream
	while(*sinkInputIter > 1){
	  //The should be an output pin, so it should only have 1 input
	  if((*wireIter)->sourceBlock->inputDelays.size() > 1){
	    cout << "Error: Host driver " << (*wireIter)->sourceBlock->name << " has more than one input!" << endl;
	    exit(-1);
	  }
	  (*wireIter)->sourceBlock->inputDelays.at(0)++;
	  (*sinkInputIter)--;
	}
      }
      else if(*sinkInputIter == 0){
	//We have to pull a register from the input of the output pin.
	if((*wireIter)->sourceBlock->inputDelays.size() > 1){
	  cout << "Error: Host driver has more than one input!" << endl;
	  exit(-1);
	}
	if((*wireIter)->sourceBlock->inputDelays.at(0) < 1){
	  cout << "Error: Host driver " << (*wireIter)->sourceBlock->name << " does not have enough registers to clean up!" << endl;
	  exit(-1);
	}
	(*wireIter)->sourceBlock->inputDelays.at(0)--;
	(*sinkInputIter)++;
      }
      else if(*sinkInputIter < 0){
	cout << "Error: Host has negative latency!" << endl;
	exit(-1);
      }
    }
  }

  //Go through each of the outputs of the host (the input pins)
  // and make sure there is exactly one latency on each edge.
  wirePtr = hostBlock->outputs.at(0);
  for(sinkIter = wirePtr->sinkBlockVector.begin(), sinkInputIter = wirePtr->sinkInputVector.begin();
      sinkIter != wirePtr->sinkBlockVector.end() && sinkInputIter != wirePtr->sinkInputVector.end();
      sinkIter++, sinkInputIter++){
    if((*sinkIter)->inputDelays.at(*sinkInputIter) > 1){
      //We have more registers than we want on this output, so push them downstream
      while((*sinkIter)->inputDelays.at(*sinkInputIter) > 1){
	//The should be an input pin, so it should only have 1 output
	if((*sinkIter)->outputs.size() > 1){
	  cout << "Error: Host sink " << (*sinkIter)->name << " has more than one output!" << endl;
	  exit(-1);
	}

	//Decrease the delay on input pin itself
	(*sinkIter)->inputDelays.at(*sinkInputIter)--;

	//Increase the delay on the sinks of this input pin
	for(sinkIter2 = (*sinkIter)->outputs.at(0)->sinkBlockVector.begin(), sinkInputIter2 = (*sinkIter)->outputs.at(0)->sinkInputVector.begin();
	    sinkIter2 != (*sinkIter)->outputs.at(0)->sinkBlockVector.end() && sinkInputIter2 != (*sinkIter)->outputs.at(0)->sinkInputVector.end();
	    sinkIter2++, sinkInputIter2++){
	  (*sinkIter2)->inputDelays.at(*sinkInputIter2)++;
	}
      }
    }

    else if((*sinkIter)->inputDelays.at(*sinkInputIter) == 0){
      //We have to pull a register from the output of the input pin
      //Increase the delay on the input pin itself;
      (*sinkIter)->inputDelays.at(*sinkInputIter)++;

      //Go through all of the sinks of the input pin and decrease their latency
      for(sinkIter2 = (*sinkIter)->outputs.at(0)->sinkBlockVector.begin(), sinkInputIter2 = (*sinkIter)->outputs.at(0)->sinkInputVector.begin();
	  sinkIter2 != (*sinkIter)->outputs.at(0)->sinkBlockVector.end() && sinkInputIter2 != (*sinkIter)->outputs.at(0)->sinkInputVector.end();
	  sinkIter2++, sinkInputIter2++){
	if((*sinkIter2)->inputDelays.at(*sinkInputIter2) < 1){
	  cout << "Error: Host sink " << (*sinkIter2)->inputs.at(*sinkInputIter2)->name << " does not have enough registers to clean up!" << endl;
	  exit(-1);
	}
	(*sinkIter2)->inputDelays.at(*sinkInputIter2)--;
      }
    }

    else if((*sinkIter)->inputDelays.at(*sinkInputIter) < 0){
      cout << "Error: Host has negative latency!" << endl;
      exit(-1);
    }
  }  
}

//Forced explicit registers may have had registers put on their output.
//Since these are constants, this is unnecessary - remove them.
void cleanupForcedExplicitRegisters(){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;
  vector<c_block*>::iterator sinkIter;
  vector<int>::iterator sinkInputIter;

  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if((*subBlockIter)->forceExplicitRegister){
	  //Go through the sinks of this block and make sure that they are are all latency 0
	  if((*subBlockIter)->outputs.size() != 1){
	    cout << "Error: A forced explicit register should have exactly 1 output!" << endl;
	    exit(-1);
	  }
	  for(sinkIter = (*subBlockIter)->outputs.at(0)->sinkBlockVector.begin(), 
		sinkInputIter = (*subBlockIter)->outputs.at(0)->sinkInputVector.begin();
	      sinkIter != (*subBlockIter)->outputs.at(0)->sinkBlockVector.end() &&
		sinkInputIter != (*subBlockIter)->outputs.at(0)->sinkInputVector.end();
	      sinkIter++, sinkInputIter++){
	    if((*sinkIter)->inputDelays.at(*sinkInputIter) > 0){
	      (*sinkIter)->inputDelays.at(*sinkInputIter) = 0;
	    }
	  }
	}
      }
    }
  }
}

//Go through all blockas and show which ones have a delay too high to achieve the
// currently attempted CPD
void printOverCriticalBlocks(float attemptCPD){
  vector <c_block *>::iterator blockIter;
  vector <c_block *>::iterator subBlockIter;

  for(blockIter = mDef.blocks.begin(); blockIter != mDef.blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN ||  (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      if((*blockIter)->deltaV > attemptCPD){
	cout << "Block " << (*blockIter)->name << " has delay " << (*blockIter)->deltaV << endl;
      }
      else{
	cout << "Block " << (*blockIter)->name << " has OKdelay " << (*blockIter)->deltaV << endl;
      }
    }
    else if((*blockIter)->type == CLB){
      //Go through the subblocks of this CLB and do this
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)-> subblocks.end(); subBlockIter++){
	if((*subBlockIter)->deltaV > attemptCPD){
	  cout << "Block " << (*subBlockIter)->name << " has delay " << (*subBlockIter)->deltaV << endl;
	}
	else{
	  cout << "Block " << (*subBlockIter)->name << " has OKdelay " << (*subBlockIter)->deltaV << endl;
	}
      }
    } 
  }
}
