#include "include.h"
extern c_moduleDef mDef;
extern int numCLBInputs;
extern int numCLBOutputs;
extern int numSBInputs;

c_block* c_moduleDef::addBlock(enum e_blockType typeNum, string str, int numInputs, int numOutputs, float delay){
  c_block *blockPtr;

  blockPtr = findBlock(str);
  if(blockPtr){
    cout << "Error: Block " << str << " already exists!" << endl;
    exit(-1);
  }
  
  //cout << "Adding block " << str << endl;
  blockPtr = new c_block(typeNum, str, numInputs, numOutputs, delay);

  blocks.push_back(blockPtr);
  return blockPtr;
}

c_wire* c_moduleDef::addWire(string str){
  vector<c_wire *>::iterator wireIter;
  c_wire *tempWire;

  // See if we are trying to connect the wire to nothing
  if(str.compare("open") == 0){
    return NULL;
  }

  //Search the wire vector to see if this wire already exists
  tempWire = findWire(str);
  if(tempWire){
    return tempWire;
  }

  //If it doesn't exist already, add it to the list
  tempWire = new c_wire(str);

  //cout << "Adding wire " << str << endl;
  wires.push_back(tempWire);

  return tempWire;
}

c_wire* c_moduleDef::findWire(string str){
  vector<c_wire *>::iterator wireIter;
  c_wire *tempWire;

  //Search the wire vector to see if this wire already exists
  for(wireIter = wires.begin(); wireIter != wires.end(); wireIter++){
    if((*wireIter)->name.compare(str) == 0){
      return *wireIter;
    }
  }

  return NULL;
}

void c_moduleDef::process(bool clbReg){
  vector<c_block *>::iterator blockIter;
  vector<c_block *>::iterator subBlockIter;
  vector<c_wire *>::iterator wireIter;
  vector<c_block *>::iterator blockIter2;
  vector<int>::iterator inputIter;
  c_wire *inputWire;
  c_wire *outputWire;

  string::size_type index;
  int subblockNum;
  int inputNum;
  int i, j;
  bool done;

  //We are trying to find subblocks in which the input is connected to a "ble_X" input.
  //If that "ble_X" wire does not have a source connected to it, that means that
  // the logic block output pin is connected to an external wire and we should 
  // use that wire instead.
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
      for(wireIter = (*subBlockIter)->inputs.begin(), inputNum = 0; wireIter != (*subBlockIter)->inputs.end(); wireIter++, inputNum++){
	//See if the name of this wire is "ble_X"
	if(*wireIter){
	  index = (*wireIter)->name.find_last_of(":ble_", (*wireIter)->name.size() - 1);
	  if(index != string::npos ){
	    //See if the source of this wire is unassigned
	    if(!(*wireIter)->sourceBlock){
	      //Connect this block's input to the correct subblock's output
	      (*wireIter)->removeSink(*subBlockIter, inputNum);
	      subblockNum = atoi((*wireIter)->name.substr(index + 1).c_str());
	      (*wireIter) = (*blockIter)->subblocks.at(subblockNum)->outputs.at(0);
	      (*wireIter)->addSink(*subBlockIter, inputNum);
	    }
	    //If the source of this wire is defined, then we must have no external wire to connect to.
	    //Keep this wire how it is.
	  }
	}
      }
    }
  }

  //Go through the subblocks and figure out who's outputs are registered.
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
      //Check to see if the subblock is registered (last input pin is occupied)
      if((*subBlockIter)->inputs.at(numSBInputs)){ 
	//If it is registered, mark it
	(*subBlockIter)->registered = true;
	//cout << "Block " << (*blockIter)->name << " subblock " << (*subBlockIter)->name << " output is registered" << endl;
	
	//Now we don't need the connection to the clock anymore.
	//Remove this block from the clock wire's sink list
	inputWire = (*subBlockIter)->inputs.at(numSBInputs);
	if(inputWire){
	  inputWire->removeSink(*subBlockIter, numSBInputs);
	}
	else{
	  cout << "Error: Something is odd with the clock input on block " << (*subBlockIter)->name << endl;
	  exit(-1);
	}

	//Disconnect this block from the clock wire
	(*subBlockIter)->inputs.at(numSBInputs) = NULL;

	//See if it is a lone FF (there is less than 2 inputs that are not open)
	inputNum = 0;
	for(wireIter = (*subBlockIter)->inputs.begin(); wireIter != (*subBlockIter)->inputs.end(); wireIter++){
	  if(*wireIter){
	    inputNum ++;
	  }
	}

	if(inputNum < 2){
	  (*subBlockIter)->isOnlyFF = true; 
	}
	if(inputNum < 1){
	  cout << "Error: Block " << (*subBlockIter)->name << " does not have an input!" << endl;
	  exit(-1);
	}

	//If we are explicitly dumping registers, everyone is at a delay of 0 since we don't want any MORE registers.
	if(!clbReg){
	  //If not, we have to go through and determine the latency on all inputs.

	  //Make sure that this not a *forced* explicit register - FF that connects to itself
	  //If it is, we don't want to increase the latency on this wire because it is already
	  // understood.
	  if((*subBlockIter)->isOnlyFF && (*subBlockIter)->inputs.at(0) == (*subBlockIter)->outputs.at(0)){
	    (*subBlockIter)->forceExplicitRegister = true;
	    continue;
	  }

	  //Go through the sinks of this wire and increase their delay by 1
	  for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
	    if(*wireIter){
	      for(blockIter2 = (*wireIter)->sinkBlockVector.begin(), inputIter = (*wireIter)->sinkInputVector.begin(); 
		  blockIter2 != (*wireIter)->sinkBlockVector.end(), inputIter != (*wireIter)->sinkInputVector.end();
		  blockIter2++, inputIter++){
		(*blockIter2)->inputDelays.at(*inputIter)++;
	      }
	    }
	    else{
	      cout << "Error: Subblock " << (*subBlockIter)->name << " has no output!" << endl;
	      exit(-1);
	    }
	  }
	}
      }
      //This block is not registered, but we should check the number of inputs - if it only has one input
      // we could lose this block in the retiming/re-packing process
      else{
	inputNum = 0;
	for(wireIter = (*subBlockIter)->inputs.begin(); wireIter != (*subBlockIter)->inputs.end(); wireIter++){
	  if(*wireIter){
	    inputNum ++;
	  }
	}

	if(inputNum < 2){
	  cout << "WARNING: block " << (*subBlockIter)->name;
	  cout << " is an inverter and we could lose it in the retiming/repacking process!" << endl; 
	}
      }
    }
  }

  //If we are not dumping explicit registers we should remove all of the subblocks that only use their FF
  if(!clbReg){
    done = false;
    
    while(!done){
      done = true;
      for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
	for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	  if(!(*subBlockIter)->removed && (*subBlockIter)->isOnlyFF){
	    //If it is only a FF, find all other blocks that take the outputs as inputs and increase their delay by one
	    // then remove the subblock

	    inputWire = (*subBlockIter)->inputs.at(0);
	    outputWire = (*subBlockIter)->outputs.at(0);

	    //Find all nets that are connected to the output of this subblock and re-connect them to the non-open input
	    // of this block (should be input 0) and increase their delay by one.
	    //Then, remove this subblock
	    if(!inputWire){
	      cout << "Error: FF " << (*subBlockIter)->name << " needs to be WORKED BY HAND!" << endl;
	      exit (-1);
	    }
	    if(!outputWire){
	      cout << "Error: Output of FF " << (*subBlockIter)->name << " is not connected to anything!" << endl;
	      exit (-1);
	    }

	    //If the input wire and the output wire are the same, then we can't remove this FF.
	    //Otherwise there will be no source.
	    //If this is the case, we will have to leave this as a lone fixed FF, even if
	    // we are not dumping explict registers, this will be an explicit register.
	    if((*subBlockIter)->forceExplicitRegister){
	      cout << "FF " << (*subBlockIter)->name << " must be explicitly defined, even though we are not dumping explict registers!" << endl;
	      continue;
	    }

	    done = false;

	    //Go through the blocks that are attached to the output wire and connect them to the input wire
	    for(blockIter2 = outputWire->sinkBlockVector.begin(), inputIter = outputWire->sinkInputVector.begin();
		blockIter2 != outputWire->sinkBlockVector.end(), inputIter != outputWire->sinkInputVector.end();
		blockIter2++, inputIter++){
	      //Connect this block to the input wire
	      (*blockIter2)->inputs.at(*inputIter) = inputWire;
	      (*blockIter2)->inputDelays.at(*inputIter) += (*subBlockIter)->inputDelays.at(0);

	      //Make this block a sink of the input wire
	      inputWire->addSink(*blockIter2, *inputIter);
	    }
	    
	    //Remove everything from the output wire's sink list
	    outputWire->sinkBlockVector.clear();
	    outputWire->sinkInputVector.clear();

	    //Remove the source from the output wire's source list
	    outputWire->removeSource(*subBlockIter, 0);

	    //Remove this block from the input wire's sink list
	    inputWire->removeSink(*subBlockIter, 0);

	    //Now, remove sub-block and wire
	    (*subBlockIter)->removed = true;
	    outputWire->removed = true;

	    cout << "Removing FF " << (*subBlockIter)->name << endl;
	    //cout << "Removing Wire " << outputWire->name << endl;
	  }
	}
      }    
    }
  }
  else{
    //We are dumping explict registers.  Thus, we should look at all of the subblocks to see if they
    // need an explict register.  If a subblock is registered, but is not a FF only then we will
    // need to disconnect the output and insert an explicit register
    for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
      for(subblockNum = 0; subblockNum < (*blockIter)->subblocks.size(); subblockNum++){
	if((*blockIter)->subblocks.at(subblockNum)->registered && !(*blockIter)->subblocks.at(subblockNum)->isOnlyFF){
	  //Make the block unregistered
	  (*blockIter)->subblocks.at(subblockNum)->registered = false;
	  
	  //Create a new subblock that is a register
	  (*blockIter)->addSubBlockReg((*blockIter)->name + ":RegBlock" + itoa(subblockNum));

	  //Connect the register output to the old subblock's output
	  outputWire = (*blockIter)->subblocks.at(subblockNum)->outputs.at(0);

	  if(!outputWire){
	    cout << "Error: Subblock " << (*blockIter)->subblocks.at(subblockNum)->name << " has no output!" << endl;
	    exit(-1);
	  }

	  outputWire->removeSource((*blockIter)->subblocks.at(subblockNum), 0);
	  (*blockIter)->currSubBlock->outputs.at(0) = outputWire;
	  outputWire->addSource((*blockIter)->currSubBlock, 0);

	  //Create a new wire 
	  inputWire = mDef.addWire((*blockIter)->name + ":RegWire" + itoa(subblockNum));

	  //Add that wire to both the new subblock input and the old subblock output
	  (*blockIter)->subblocks.at(subblockNum)->outputs.at(0) = inputWire;
	  inputWire->addSource((*blockIter)->subblocks.at(subblockNum), 0);

	  (*blockIter)->currSubBlock->inputs.at(0) = inputWire;
	  inputWire->addSink((*blockIter)->currSubBlock, 0);
	}

      }
    }
  }

  //Perform some sanity checks.
  //1) Check that any wire that has a valid source has valid sinks and vice-versa
  for(wireIter = wires.begin(); wireIter != wires.end(); wireIter++){
    //The clock wire should have a source but no sinks.
    if((*wireIter)->name.compare("pclk") == 0){
       if(!(*wireIter)->hasSource() || (*wireIter)->hasSink()){
	 cout << "Error: Clock wire, if present, should have a source but no sinks by this point!" << endl;
	 exit(-1);
       }
    }
    
    //All the other wires should be checked as normal
    else{
      if((*wireIter)->hasSource() && !(*wireIter)->hasSink()){
	cout << "Error: Wire " << (*wireIter)->name << " has a source but no sinks!" << endl;
	exit(-1);
      }
      else if (!(*wireIter)->hasSource() && (*wireIter)->hasSink()){
	cout << "Error: Wire " << (*wireIter)->name << " has sinks but no source!" << endl;
	exit(-1);
      }
      else if (!(*wireIter)->hasSource() && !(*wireIter)->hasSink()){
	//We can remove this wire entirely from the system since it isn't connected to anything
	if(!(*wireIter)->removed){
	  //cout << "Removing wire " << (*wireIter)->name << endl;
	  (*wireIter)->removed = true;
	}
      }
    }
  }
  //2) Go through all of the valid blocks and make sure that all of the wires it is connected to are valid
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if(!(*blockIter)->removed){
      if((*blockIter)->type != CLB){
	for(wireIter = (*blockIter)->inputs.begin(); wireIter != (*blockIter)->inputs.end(); wireIter++){
	  if(*wireIter){
	    if((*wireIter)->removed){
	      cout << "Error: Block " << (*blockIter)->name << " is attached to removed wire!" << endl;
	      exit(-1);
	    }
	  }
	}
	for(wireIter = (*blockIter)->outputs.begin(); wireIter != (*blockIter)->outputs.end(); wireIter++){
	  if(*wireIter){
	    if((*wireIter)->removed){
	      cout << "Error: Block " << (*blockIter)->name << " is attached to removed wire!" << endl;
	      exit(-1);
	    }
	  }
	}
	//If it the block is an input pin, make sure that is is actually attached to something
	if((*blockIter)->type == INPUTPIN){
	  if(!(*blockIter)->outputs.at(0)){
	    cout << "Error: Input pin " << (*blockIter)->name << " is not connected to anything!" << endl;
	    exit(-1);
	  }
	}
	
	//If the block is an output pin, make sure that it is actually attached to something
	else if((*blockIter)->type == OUTPUTPIN){
	  if(!(*blockIter)->inputs.at(0)){
	    cout << "Error: Output pin " << (*blockIter)->name << " is not connected to anything!" << endl;
	    exit(-1);
	  }
	}
      }
      else{
	//If the block is a CLB, make sure that there are no invalid wires - we already checked the 
	// *number* of inputs and outputs, but not their validity.
	for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	  inputNum = 0;
	  if(!(*subBlockIter)->removed){
	    for(wireIter = (*subBlockIter)->inputs.begin(); wireIter != (*subBlockIter)->inputs.end(); wireIter++){
	      if(*wireIter){
		inputNum++;
		if((*wireIter)->removed){
		  cout << "Error: Subblock " << (*subBlockIter)->name << " is attached to removed wire!" << endl;
		  exit(-1);
		}
	      }
	    }
	    if(inputNum == 0){
	      cout << "Error: Subblock " << (*subBlockIter)->name << " has no input!" << endl;
	      exit(-1);
	    }
	    for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
	      if(*wireIter){
		if((*wireIter)->removed){
		  cout << "Error: Subblock " << (*subBlockIter)->name << " is attached to removed wire!" << endl;
		  exit(-1);
		}
	      }
	      else{
		cout << "Error: Subblock " << (*subBlockIter)->name << " has no output!" << endl;
		exit(-1);
	      }
	    }	  
	  }
	}
      }
    }
  }

  //Actually remove all wires marked for removal and all blocks marked for removal
  for(i = 0; i < wires.size(); i++){
    if(wires.at(i)->removed){
      wires.erase(wires.begin() + i);
      i--;
    }
  }
  for(i = 0; i < blocks.size(); i++){
    if(blocks.at(i)->removed){
      blocks.erase(blocks.begin() + i);
      i--;
    }
    else{
      for(j = 0; j < blocks.at(i)->subblocks.size(); j++){
	if(blocks.at(i)->subblocks.at(j)->removed){
	  blocks.at(i)->subblocks.erase(blocks.at(i)->subblocks.begin() + j);
	  j--;
	}
      }
    }
  }
  
}

//Go through the netlist and figure out how many output pins we have
int c_moduleDef::countOutputPins(){
  int numOutPins = 0;
  vector<c_block *>::iterator blockIter;

  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == OUTPUTPIN){
      numOutPins ++;
    }
  }
  return numOutPins;
}

int c_moduleDef::returnNumAllBlocks(){
  int numBlocks = 0;
  vector<c_block *>::iterator blockIter;

  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN || (*blockIter)->type == OUTPUTPIN || (*blockIter)->type == HOST){
      numBlocks++;
    }
    else if((*blockIter)->type == CLB){
      numBlocks += (*blockIter)->subblocks.size();
    }
  }
  return numBlocks;
}

//Take the existing netlist and add a host vertex.
//This means that all input pins and output pins are
// attached to a new node named "HOST" with a single latency.
void c_moduleDef::addHostVertex(){
  int numOutputPins;
  c_block *host;
  c_wire *hostInWire;
  c_wire *hostOutWire;
  int i;
  vector<c_block *>::iterator blockIter;

  //Figure out how many output pins we have
  numOutputPins = countOutputPins();

  //Create a host vertex
  host = addBlock(HOST, "HOST", numOutputPins, 1, 0.0);
  hostOutWire = addWire("hostOutWire");
  host->outputs.at(0) = hostOutWire;
  hostOutWire->addSource(host, 0);
  hostOutWire->hostWire = true;

  //Connect the host to all output pins
  for(blockIter = blocks.begin(), i = 0; blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == OUTPUTPIN){
      hostInWire = addWire((*blockIter)->name + ":ToHost" + itoa(i));
      host->inputs.at(i) = hostInWire;
      host->inputDelays.at(i) = 1;
      hostInWire->addSink(host, i);
      hostInWire->hostWire = true;
      i++;

     (*blockIter)->outputs.at(0) = hostInWire;
      hostInWire->addSource((*blockIter), 0);      
    }
    else if((*blockIter)->type == INPUTPIN){
      //Don't add a link to the clock pin
      if((*blockIter)->name.compare("pclk") != 0){
	(*blockIter)->inputs.at(0) = hostOutWire;
	(*blockIter)->inputDelays.at(0) = 1;
	hostOutWire->addSink((*blockIter), 0);
      }
    }    
  }
}

//Output a .blif file
void c_moduleDef::outputBlif(string outFile){
  vector<c_block *>::iterator blockIter;
  vector<c_block *>::iterator subBlockIter;
  vector<c_wire *>::iterator wireIter;
  vector<c_block *>::iterator sinkIter;
  vector <int>::iterator sinkInputIter;
  int i, maxLatency;
  int charCount;

  cout << "Outputting .blif file to " << outFile << endl;

  ofstream fout(outFile.c_str(), ios::out);

  if(!fout){
    cout << "Error: Cannot open output file " << outFile << "!" << endl;
    exit(-1);
  }

  fout << ".model top" << endl;

  //Output all of the input pins
  fout << ".inputs ";
  charCount = 0;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN){
      if(charCount + (*blockIter)->name.size() + 2 > 100){
	fout << "\\" << endl;
	charCount = 0;
      }
      if((*blockIter)->name.compare("pclk") != 0){
	fout << (*blockIter)->name << "_0 ";
	charCount += (*blockIter)->name.size() + 2;
      }
      else{
	fout << (*blockIter)->name << " ";
	charCount += (*blockIter)->name.size() + 1;
      }
    }
  }  
  fout << endl;

  //Output all of the output pins
  fout << ".outputs ";
  charCount = 0;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == OUTPUTPIN){
      if((*blockIter)->inputs.size() > 1){
	cout << "Error: Output pin should not have more than 1 inputs!" << endl;
	exit(-1);
      }
      
      if(charCount + (*blockIter)->inputs.at(0)->name.size() + 2 > 100){
	fout << "\\" << endl;
	charCount = 0;
      }

      //The name of the pin will be wirename_latency amount
      fout << (*blockIter)->inputs.at(0)->name << "_" << (*blockIter)->inputDelays.at(0) << " ";
      charCount += (*blockIter)->inputs.at(0)->name.size() + 2;
    }
  }  
  fout << endl;

  //Output all of the required FF, assuming maximum sharing
  //First, go through all of the wires and figure out the maximim latency of all sinks.
  for(wireIter = wires.begin(); wireIter != wires.end(); wireIter++){
    //Make sure that the wire is not a host->input pin or output pin->host wire
    if(!(*wireIter)->hostWire){
      maxLatency = 0;
      for(sinkIter = (*wireIter)->sinkBlockVector.begin(), sinkInputIter = (*wireIter)->sinkInputVector.begin(); 
	  sinkIter != (*wireIter)->sinkBlockVector.end() && sinkInputIter != (*wireIter)->sinkInputVector.end();	
	  sinkIter++, sinkInputIter++){
	if((*sinkIter)->inputDelays.at(*sinkInputIter) > maxLatency){
	  maxLatency = (*sinkIter)->inputDelays.at(*sinkInputIter);

	  if(maxLatency > 100){
	    cout << "Error: Unusually high latency on wire " << (*wireIter)->name << " to sink " << (*sinkIter)->name << endl;
	    cout << "\tsomething might be wrong!" << endl;
	    //exit(-1);
	  }
	}
      }

      //Output N FF's
      for(i = 1; i <= maxLatency; i++){
	fout << ".latch " << (*wireIter)->name << "_" << i - 1 << " " << (*wireIter)->name << "_" << i << " re pclk 2" << endl;
      }
    }
  }

  //Go through all of the blocks and look for forced explicit registers
  //Output all of the LUTs
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == CLB){
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	if((*subBlockIter)->forceExplicitRegister){
	  fout << ".latch " << (*subBlockIter)->outputs.at(0)->name << "_0 " << (*subBlockIter)->outputs.at(0)->name << "_0" << " re pclk 2" << endl;
	}
      }
    }
  }

  //Output all of the LUTs
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == CLB){
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	if((*subBlockIter)->forceExplicitRegister){
	  continue;
	}

	fout << ".names ";
	for(wireIter = (*subBlockIter)->inputs.begin(), sinkInputIter = (*subBlockIter)->inputDelays.begin(); 
	    wireIter != (*subBlockIter)->inputs.end() && sinkInputIter != (*subBlockIter)->inputDelays.end(); 
	    wireIter++, sinkInputIter++){
	  if(*wireIter){
	    fout << (*wireIter)->name << "_" << *sinkInputIter << " ";
	  }
	}
	if((*subBlockIter)->outputs.size() != 1){
	  cout << "Error: Subblock " << (*subBlockIter)->name << " should have one output!"<< endl;
	  exit(-1);
	}
	fout << (*subBlockIter)->outputs.at(0)->name << "_0" << endl;

	//If we wanted to fill in the blif with the correct logic values, we should do it here.
	//However, all of that information is stuck in the original .blif file - we would need to
	// suck that file in and match up the correct blocks.
	fout << endl;
      }
    }
  }  
  fout << ".end" << endl;  

}

//Output a standard Armada netlist
void c_moduleDef::outputNetlist(char *outFile, bool clbReg){
  ofstream fout(outFile, ios::out);
  int numSubBlocks;
 
  vector<c_block *>::iterator blockIter;
  vector<c_block *>::iterator subBlockIter;
  vector<c_wire *>::iterator wireIter;

  vector <int>::iterator delayIter;
  int pinNum;

  cout << "Outputting .ArmadaNet file to " << outFile << endl;

  if(!fout){
    cout << "Error: Cannot open output file " << outFile << "!" << endl;
    exit(-1);
  }
  
  //Dump all of the input pins
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN && (*blockIter)->name.compare("pclk") != 0){
      fout << ".input " << (*blockIter)->name << endl;
      fout << "pinlist: " << (*blockIter)->outputs[0]->name << endl << endl;
    }
  }  

  //Dump all of the output pins
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == OUTPUTPIN){
      fout << ".output " << (*blockIter)->name << endl;
      fout << "pinlist: ";
      for(wireIter = (*blockIter)->inputs.begin(); wireIter != (*blockIter)->inputs.end(); wireIter++){
	fout << (*wireIter)->name << " ";
      }
      fout << endl;

      fout << "delay: ";
      for(delayIter = (*blockIter)->inputDelays.begin(); delayIter != (*blockIter)->inputDelays.end(); delayIter++){
	fout << *delayIter << " ";
      }
      fout << endl <<endl;
    }
  }

  //Dump all of the blocks
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == CLB){
      numSubBlocks = 0;
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	if(!(*subBlockIter)->removed){
	  numSubBlocks++;
	}
      }

      if(numSubBlocks > 0){
	fout << ".clb " << (*blockIter)->name << endl;
	fout << ".type CLB" << endl;

	//We are not dumping explict registers - all that information is included in the delay line	  
	for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	  if(!clbReg){
	    //We are not dumping explict registers
	    if(!(*subBlockIter)->removed){
	      //We might have a forced explicit register
	      if((*subBlockIter)->isOnlyFF){
		if(!(*subBlockIter)->forceExplicitRegister){
		  cout << "Error: We should not have unforced FF in the netlist!" << endl;
		  exit(-1);
		}
		//This is a FF so output as such
		fout << ".subblockREG " << (*subBlockIter)->name << endl;
		
		if((*subBlockIter)->inputs.at(0) && (*subBlockIter)->outputs.at(0)){  
		  fout << "pinlist: " << (*subBlockIter)->inputs.at(0)->name << " " << (*subBlockIter)->outputs.at(0)->name << endl;
		  
		  fout << "delay: " << (*subBlockIter)->inputDelays.at(0) << endl;
		}
		else{
		  cout << "Error: Register " << (*subBlockIter)->name << " has empty input or output!" << endl;
		  exit(-1);
		}
	      }
	      else{
		fout << ".subblock " << (*subBlockIter)->name << endl;
		
		fout << "pinlist: ";
		for(wireIter = (*subBlockIter)->inputs.begin(), pinNum = 0; 
		    wireIter != (*subBlockIter)->inputs.end(), pinNum != numSBInputs; wireIter++, pinNum++){
		  if(*wireIter){
		    fout << (*wireIter)->name << " ";
		  }
		  else{
		    fout << "open ";
		  }
		}
		
		for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
		if(*wireIter){
		  fout << (*wireIter)->name << " ";
		}
		else{
		  cout << "Error: Subblock " << (*subBlockIter)->name << " has no output!" << endl;
		  exit(-1);
		}
		}	  
		
		fout << endl;
		
		fout << "delay: ";
		for(delayIter = (*subBlockIter)->inputDelays.begin(), pinNum = 0; 
		    delayIter != (*subBlockIter)->inputDelays.end(), pinNum != numSBInputs; delayIter++, pinNum++){
		  fout << *delayIter << " ";
		}
		
		fout << endl; 
	      }
	    }
	  }
	  else{
	    //We are dumping explicit registers
	    //Check if it is a FF only subblock
	    if(!(*subBlockIter)->isOnlyFF){
	      //This is not just a FF, so output normally
	      fout << ".subblock " << (*subBlockIter)->name << endl;
	      
	      fout << "pinlist: ";
	      for(wireIter = (*subBlockIter)->inputs.begin(), pinNum = 0; 
		  wireIter != (*subBlockIter)->inputs.end(), pinNum != numSBInputs; wireIter++, pinNum++){
		if(*wireIter){
		  fout << (*wireIter)->name << " ";
		}
		else{
		  fout << "open ";
		}
	      }
	      
	      for(wireIter = (*subBlockIter)->outputs.begin(); wireIter != (*subBlockIter)->outputs.end(); wireIter++){
		if(*wireIter){
		  fout << (*wireIter)->name << " ";
		}
		else{
		  cout << "Error: Subblock " << (*subBlockIter)->name << " has no output!" << endl;
		  exit(-1);
		}
	      }	  
	      
	      fout << endl;
	      
	      fout << "delay: ";
	      for(delayIter = (*subBlockIter)->inputDelays.begin(), pinNum = 0; 
		  delayIter != (*subBlockIter)->inputDelays.end(), pinNum != numSBInputs; delayIter++, pinNum++){
		fout << *delayIter << " ";
	      }
	      
	      fout << endl; 
	    }
	    else{
	      //This is a FF so output as such
	      fout << ".subblockREG " << (*subBlockIter)->name << endl;
	      
	      if((*subBlockIter)->inputs.at(0) && (*subBlockIter)->outputs.at(0)){  
		fout << "pinlist: " << (*subBlockIter)->inputs.at(0)->name << " " << (*subBlockIter)->outputs.at(0)->name << endl;

		fout << "delay: 0" << endl;
	      }
	      else{
		cout << "Error: Register " << (*subBlockIter)->name << " has empty input or output!" << endl;
		exit(-1);
	      }
	    }	    
	  }
	}
	fout << endl;
      }
    }
  }
}

//Find the block with the given name and return a ptr
c_block* c_moduleDef::findBlock(string blockName){
  vector<c_block *>::iterator blockIter;
   
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->name.compare(blockName) == 0){
      //cout << "Found block " << blockName << endl;
      return *blockIter;
    }
  }
  return NULL;
}


void c_moduleDef::printInfo(){
  vector<c_block *>::iterator blockIter;
  int i;

  cout << "*******PRINTINFO*** " << endl;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type != CLB){
      (*blockIter)->printInfo();
    }
    else{
      for(i = 0; i < (*blockIter)->subblocks.size(); i++){
	(*blockIter)->subblocks.at(i)->printInfo();
      }
    }
  }
  cout << "***********END PRINT INFO*** " << endl;
}

void c_moduleDef::countNetlist(){
  vector<c_block *>::iterator blockIter;
  vector<c_block *>::iterator subBlockIter;
  vector<c_wire *>::iterator wireIter;
  vector<int>::iterator delayIter;
  vector<int> latencyCounts(50, 0);
  int i, count, FFcount, unregLUTcount, regLUTcount;

  count = 0;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == INPUTPIN){
      count ++;
    }
  }
  cout << count << " input pins" << endl;

  count = 0;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == OUTPUTPIN){
      count ++;
    }
  }
  cout << count << " output pins" << endl;

  count = 0;
  FFcount = 0;
  unregLUTcount = 0;
  regLUTcount = 0;
  for(blockIter = blocks.begin(); blockIter != blocks.end(); blockIter++){
    if((*blockIter)->type == CLB && !(*blockIter)->removed){
      count ++;
      for(subBlockIter = (*blockIter)->subblocks.begin(); subBlockIter != (*blockIter)->subblocks.end(); subBlockIter++){
	if(!(*subBlockIter)->removed){
	  if((*subBlockIter)->isOnlyFF){
	    FFcount++;
	  }
	  else if((*subBlockIter)->registered){
	    regLUTcount++;
	  }
	  else{
	    unregLUTcount++;
	  }
	  for(delayIter = (*subBlockIter)->inputDelays.begin(), wireIter = (*subBlockIter)->inputs.begin(); 
	      delayIter != (*subBlockIter)->inputDelays.end(); delayIter++, wireIter++){
	    if(*wireIter){
	      if((*wireIter)->name.compare("pclk") != 0){
		latencyCounts.at(*delayIter)++;
	      }
	    }
	  }
	}
      }
    }
  }
  cout << count << " CLB" << endl;
  cout << unregLUTcount << " unregistered LUTs" << endl;
  cout << regLUTcount << " registered LUTs" << endl;
  cout << FFcount << " FFs" << endl;
  for(delayIter = latencyCounts.begin(), i = 0; delayIter != latencyCounts.end(); delayIter++, i++){
    cout << *delayIter << " signals at latency " << i << endl;
  }
}

void c_block::printInfo(){
  int i;

  cout << "Name = " << name << endl;
  cout << "\tInputs" << endl;
  for(i = 0; i < inputs.size(); i ++){
    if(inputs[i]){
      cout << "\t" << inputs[i]->name;
    }
    else{
      cout << "\topen";
    }
  }
  cout << endl;
  
  cout << "\tDelays" << endl;
  for(i = 0; i < inputDelays.size(); i ++){
    cout << "\t" << inputDelays[i];
  }
  cout << endl;
  
  cout << "\tOutputs" << endl;
  for(i = 0; i < outputs.size(); i++){
    if(outputs[i]){
      cout << "\t" << outputs[i]->name;
    }
    else{
      cout << "\topen";
    }
  }
  cout << endl;
}

void c_block::addInput(c_wire *wirePtr, int inputNum){
  if(inputs.size() <= inputNum){
    cout << "Error: Block " << name << "does not have an input #" << inputNum << endl;
    exit(-1);
  } 

  inputs[inputNum] = wirePtr;
   
  //If we are trying to set the input to "open", don't try to add a sink to it.
  if(wirePtr){
    if(type != CLB){
      //The real sink of a CLB is attached to the subblock - when we figure that out
      // later we'll set the sink of the wire
      wirePtr->addSink(this, inputNum);
    }
  }
}

void c_block::addOutput(c_wire *wirePtr, int outputNum){
  if(outputs.size() <= outputNum){
    cout << "Error: Block " << name << "does not have an output #" << outputNum << endl;
    exit(-1);
  } 

  outputs[outputNum] = wirePtr;

  //If we are trying to set the output to "open", don't try to add a source to it.
  if(wirePtr){
    if(type != CLB){
      //The real source of a CLB is attached to the subblock - when we figure that out
      // later we'll set the source of the wire
      wirePtr->addSource(this, outputNum);
    }
  }
}

c_block* c_block::findSubBlock(string str){
  vector<c_block *>::iterator blockIter;
  
  for(blockIter = subblocks.begin(); blockIter != subblocks.end(); blockIter++){
    if(((*blockIter)->type != SUBBLOCK)){
      cout << "Error: Block " << str << " has a problem!" << endl;
      exit(-1);
    }
    else if((*blockIter)->name.compare(str) == 0){
      return *blockIter;
    }
  }
  return NULL;
}

c_block* c_block::addSubBlock(string str, float del){
  c_block *subBlockPtr;
  
  //Search the block's subblock vector to see if this subblock already exists
  if(findSubBlock(str)){
    cout << "Error: Sub block " << str << " already exists!" << endl;
    exit(-1);
  }
  
  subBlockPtr = new c_block(SUBBLOCK, str, numSBInputs + 1, 1, del);
  
  subblocks.push_back(subBlockPtr);
  
  return subBlockPtr;
}

void c_block::addSubBlockInput(int inputNum, string input){
  c_wire *wirePtr = NULL;

  //Try to find input in the pinlist
  //It could be a number (0 through num inputs per CLB), open, or could be
  //"ble_#" indicating the output of a particular BLE
  if(input.compare("open") == 0){
    //cout << "Subblock " << currSubBlock->name << " input " << inputNum << " connected to open" << endl;
    currSubBlock->inputs.at(inputNum) = wirePtr;
    return;
  }
  else if(input.compare(0, 4, "ble_") == 0){
    //cout << "Subblock " << currSubBlock->name << " input " << inputNum << " connected to other subblock" << endl;
    wirePtr = mDef.addWire(name + ":" + input);
  }
  else if(atoi(input.c_str()) >= 0 && atoi(input.c_str()) < numCLBInputs){
    //cout << "Subblock " << currSubBlock->name << " input " << inputNum << " connected to CLB input " << atoi(input.c_str()) << endl;
    wirePtr = inputs.at(atoi(input.c_str()));
  }
  else if(atoi(input.c_str()) == numCLBInputs + numCLBOutputs){
    //This better be the clock input
    //cout << "Subblock " << currSubBlock->name << " input " << inputNum << " connected to CLB clock input " << atoi(input.c_str()) - numCLBOutputs << endl;
    wirePtr = inputs.at(atoi(input.c_str()) - numCLBOutputs);
  }
  else{
    cout << "Error: Something wrong with subblock input!" << endl;
    exit(-1);
  }

  currSubBlock->inputs.at(inputNum) = wirePtr;
  
  //Add this subblock as a sink of this wire - notice it should never be connected to open at this point
  wirePtr->addSink(currSubBlock, inputNum);
}

void c_block::addSubBlockOutput(string output){
  c_wire *wirePtr = NULL;
 
  //Try to find output in the pinlist
  //It could be a number (0 through num outputs per CLB) or open
  if(output.compare("open") == 0){
    cout << "Error: Subblock output should not be defined as OPEN" << endl;
    exit(-1);
  }
  else if(output.compare(0, 4, "ble_") == 0){
    cout << "Error: Subblock output should not be defined as BLE_" << endl;
    exit(-1);
  }
  else if(atoi(output.c_str()) - numCLBInputs >= 0 && atoi(output.c_str()) - numCLBInputs < numCLBOutputs){
    //cout << "Subblock " << currSubBlock->name << " output connected to CLB output " << atoi(output.c_str()) - numCLBInputs << endl;
    wirePtr = outputs.at(atoi(output.c_str()) - numCLBInputs);
  }

  //If this is connected to an open CLB pin, we should try to find wire "ble_X"
  if(!wirePtr){
    wirePtr = mDef.addWire(name + ":" + "ble_" + itoa(subblocks.size() - 1));
  }

  //cout << "Block " << name << "Subblock " << subblocks.back()->name << " output is connected to wire " << wirePtr->name << endl;

  currSubBlock->outputs.at(0) = wirePtr;
  
  //Add this subblock as the source of this wire
  wirePtr->addSource(currSubBlock, 0);
}

void c_block::addSubBlockReg(string str){
  vector<c_block *>::iterator blockIter;
  c_wire *wirePtr;
  int i;
  
  //Search the block's subblock vector to see if this subblock already exists
  currSubBlock = addSubBlock(str, 0.0);

  currSubBlock->registered = true;
  currSubBlock->isOnlyFF = true;

  addSubBlockInput(numSBInputs, "pclk");
}

bool c_wire::hasSource(){
  if(sourceBlock && sourceOutput != -1){
    if(!sourceBlock->removed){
      return true;
    }
    else{
      return false;
    }
  }
  else{
    return false;
  }
}

bool c_wire::hasSink(){
  vector<c_block *>::iterator blockIter;

  if(sinkBlockVector.size() != 0 && sinkInputVector.size() != 0){
    for(blockIter = sinkBlockVector.begin(); blockIter != sinkBlockVector.end(); blockIter++){
      if((*blockIter)->removed){
	return false;
      }
    }

    return true;
  }
  else{
    return false;
  }
}

void c_wire::addSink(c_block* blockPtr, int inputNum){
  sinkBlockVector.push_back(blockPtr);
  sinkInputVector.push_back(inputNum);
}

void c_wire::addSource(c_block* blockPtr, int outputNum){
  if(sourceBlock){
    cout << "Error: Source of wire already set!" << endl;
    exit(-1);
  }
  sourceBlock = blockPtr;
  sourceOutput = outputNum;
}

bool c_wire::hasSink(c_block* blockPtr, int inputNum){
  vector<c_block*>::iterator blockIter;
  vector<int>::iterator sinkNumIter; 
  //Go through the list and find mention of this block and this input pin
  
  for(blockIter = sinkBlockVector.begin(), sinkNumIter = sinkInputVector.begin();
      blockIter != sinkBlockVector.end(), sinkNumIter != sinkInputVector.end(); blockIter++, sinkNumIter++){
    if(*blockIter == blockPtr && *sinkNumIter == inputNum){
      return true;
    }
  }
  return false;
}

void c_wire::removeSink(c_block* blockPtr, int inputNum){
  vector<c_block*>::iterator blockIter;
  vector<int>::iterator sinkNumIter; 
  //Go through the list and find mention of this block and this input pin
  
  for(blockIter = sinkBlockVector.begin(), sinkNumIter = sinkInputVector.begin();
      blockIter != sinkBlockVector.end(), sinkNumIter != sinkInputVector.end(); blockIter++, sinkNumIter++){
    if(*blockIter == blockPtr && *sinkNumIter == inputNum){
      sinkBlockVector.erase(blockIter);
      sinkInputVector.erase(sinkNumIter);
      return;
    }
  }
  cout << "Error: sink not found!" << endl;
  exit(-1);
}

void c_wire::removeSource(c_block* blockPtr, int outputNum){
  if(sourceBlock == blockPtr){
    sourceBlock = NULL;
  }
  else{
    cout << "Error: Want to remove incorrect source!" << endl;
    exit(-1);
  }

  if(sourceOutput == outputNum){
    sourceOutput = -1;
  }
  else{
    cout << "Error: Want to remove incorrect source!" << endl;
    exit(-1);
  }
}
